Type: Package
Package: MetaScope
Title: Tools and functions for preprocessing 16S and metagenomic
        sequencing microbiome data
Version: 1.99.10
Authors@R: c(
    person("Sean", "Lu", , "seanlu96@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0009-0007-8005-6125")),
    person("Aubrey", "Odom", , "aodom@bu.edu", role = c("aut"),
           comment = c(ORCID = "0000-0001-7113-7598")),
    person("Rahul", "Varki", , "rvarki@bu.edu", role = "aut",
           comment = c(ORCID = "0009-0003-5721-9484")),
    person("W. Evan", "Johnson", , "wej@bu.edu", role = "aut",
           comment = c(ORCID = "0000-0002-6247-6595"))
  )
Description: This package contains tools and methods for preprocessing
    microbiome data. Functionality includes library generation,
    demultiplexing, alignment, and microbe identification.  It is in part
    an R translation of the PathoScope 2.0 pipeline.
License: GPL (>= 3)
URL: https://github.com/wejlab/metascope
        https://wejlab.github.io/metascope-docs/
BugReports: https://github.com/wejlab/MetaScope/issues
Depends: R (>= 4.2.0)
Imports: BiocFileCache, Biostrings, data.table (>= 1.16.2), dplyr,
        ggplot2, magrittr, Matrix, MultiAssayExperiment, purrr,
        Rbowtie2, readr, rlang, Rsamtools, S4Vectors, stringr,
        SummarizedExperiment, taxonomizr, tibble, tidyr, tools
Suggests: animalcules, BiocStyle, biomformat, GenomicRanges, IRanges,
        knitr, lintr, plyr, R.utils, RCurl, rmarkdown, Rsubread,
        spelling, sys, testthat, usethis
Enhances: BiocParallel
VignetteBuilder: knitr
BiocType: Software
biocViews: MicrobiomeData, ReproducibleResearch, SequencingData
Encoding: UTF-8
Language: en-US
LazyData: FALSE
RoxygenNote: 7.3.3
git_url: https://git.bioconductor.org/packages/MetaScope
git_branch: devel
git_last_commit: d49e047
git_last_commit_date: 2025-11-07
Repository: Bioconductor 3.23
Date/Publication: 2025-11-11
NeedsCompilation: no
Packaged: 2025-11-11 16:38:56 UTC; biocbuild
Author: Sean Lu [aut, cre] (ORCID: <https://orcid.org/0009-0007-8005-6125>),
  Aubrey Odom [aut] (ORCID: <https://orcid.org/0000-0001-7113-7598>),
  Rahul Varki [aut] (ORCID: <https://orcid.org/0009-0003-5721-9484>),
  W. Evan Johnson [aut] (ORCID: <https://orcid.org/0000-0002-6247-6595>)
Maintainer: Sean Lu <seanlu96@gmail.com>

# Generated by roxygen2: do not edit by hand

export(align_target)
export(align_target_bowtie)
export(blast_reassignment)
export(convert_animalcules)
export(convert_animalcules_patho)
export(convert_animalcules_silva)
export(count_matches)
export(download_accessions)
export(download_refseq)
export(download_refseq_16S)
export(extract_reads)
export(filter_host)
export(filter_host_bowtie)
export(get_children)
export(meta_demultiplex)
export(metascope_blast)
export(metascope_id)
export(metascope_inspect)
export(metascope_inspect_usage)
export(mk_bowtie_index)
export(mk_subread_index)
export(run_metascope)
importFrom(magrittr,"%>%")
importFrom(rlang,.data)

# MetaScope [![](https://img.shields.io/badge/bioconductor-3.18-blue)](https://www.bioconductor.org/packages/release/data/experiment/html/MetaScope.html)

## What is MetaScope?

MetaScope is a complete R-based 16S, metagenomic, and metatranscriptomic profiling package that can accurately identify the composition of microbes at a strain-level resolution within a sample. MetaScope can be considered as an updated and expanded R translation of [PathoScope 2.0](https://microbiomejournal.biomedcentral.com/articles/10.1186/2049-2618-2-33), a Python-based metagenomic profiling package also created by the Johnson lab. 

A few improvements made in MetaScope over PathoScope include using the BAM file format instead of the SAM file format for significantly less disk space usage, removing all dependencies to NCBI’s now defunct GI sequence annotations, and properly filtering reads that align to filter reference genomes. Functions to analyze host microbiome data are also planned to be added in future updates to the package.

## Documentation
Documentation and tutorials for MetaScope are available at our [website](https://wejlab.github.io/metascope-docs/).

## Installation

MetaScope requires R Version 4.3.

Install the development version of the package from Github:

```
if (!requireNamespace("devtools", quietly=TRUE))
  install.packages("devtools")
devtools::install_github("wejlab/MetaScope")
```

Install the release version of the package from Bioconductor:
```
if (!requireNamespace("BiocManager", quietly = TRUE))
  install.packages("BiocManager")
BiocManager::install("MetaScope")
```

## Docker

A prebuilt Docker image of MetaScope is available, which comes with the 16S reference database and accession mappings preinstalled. You can pull it directly from Docker Hub:
```
docker pull seanlu96/metascope_16s:latest
```
This container provides a ready-to-use environment for running MetaScope without additional setup.

# Helper function to identify parent kingdom, rank
id_kingdom_rank <- function(classification_table, taxon, rank_input, quiet) {
  # Get the parent taxon in super kingdom rank
  parent_kingdom <- classification_table$name[
    classification_table$rank == "superkingdom"]
  parent_rank <- "Superkingdom"
  # The Eukaryota superkingdom isn't a folder on the NCBI FTP site
  # Instead, grab the corresponding kingdom
  # because some eukaryota kingdoms are folders on the FTP site
  if (identical(parent_kingdom, "Eukaryota")) {
    parent_kingdom <- classification_table$name[
      classification_table$rank == "kingdom"]
    parent_rank <- "Kingdom"
    # There are only two eukaryota kingdoms in the taxonomy table that
    # have a corresponding folder on the FTP site. If not one of these
    # two kingdoms then forced to search all the eukaryota folders on the
    # site
    if (length(parent_kingdom) != 0) {
      if (!(parent_kingdom %in% c("Viridiplantae", "Fungi"))) {
        parent_kingdom <- classification_table$name[
          classification_table$rank == "superkingdom"]
        parent_rank <- "Superkingdom"
      }
    } else {
      parent_kingdom <- classification_table$name[
        classification_table$rank == "superkingdom"]
      parent_rank <- "Superkingdom"
    }
  }
  if (!quiet) message(taxon, " is a ", rank_input, " under the ",
                      parent_kingdom, " ", parent_rank)
  parent_kingdom <- tolower(parent_kingdom)
  # Some folders on the NCBI FTP site don't match the parent kingdom taken
  # from the classification table. We'll apply some modifications.
  if (identical(parent_kingdom, "viruses")) parent_kingdom <- "viral"
  if (identical(parent_kingdom, "viridiplantae")) parent_kingdom <- "plant"
  return(parent_kingdom)
}

# Helper function to download the refseq table (for a given group)
get_table <- function(what_tax) {
  refseq_link <- paste("https://ftp.ncbi.nlm.nih.gov/genomes/refseq/",
                       what_tax, "/assembly_summary.txt", sep = "")
  refseq_table <- utils::read.table(refseq_link, header = TRUE,
                                    sep = "\t", comment.char = "",
                                    quote = "", skip = 1, fill = TRUE)
  return(refseq_table)
}

# Helper function to obtain refseq table
download_parentkingdom <- function(parent_kingdom, quiet) {
  # Download assembly summary refseq table from NCBI
  # (includes genome download link)
  if (!quiet) message("Loading the refseq table for ", parent_kingdom)
  to_grab <- c("archaea", "bacteria", "fungi", "plant", "viral")
  if (parent_kingdom %in% to_grab) {
    refseq_table <- get_table(parent_kingdom)
    return(refseq_table)
  } else if (parent_kingdom == "eukaryota") {
    refseq_table <- lapply(c("vertebrate_mammalian", "vertebrate_other",
                             "invertebrate", "protozoa"), get_table) |>
      data.table::rbindlist() |> as.data.frame()
    return(refseq_table)
  } else message("Parent kingdom table could not be retrieved from",
                 "NCBI database. Try a different taxon.")
}

# Helper function to get species table
get_speciestab <- function(children_list, refseq_table, taxon,
                           representative, reference, quiet) {
  if (!quiet) message("Creating table of relevant taxa")
  # Filter table, keep lines with species or strains of input
  # If the taxon specified is a strain or species with no strain,
  # filter table by the taxon
  # Otherwise, filter table by the taxon in the children list
  if (rlang::is_empty(children_list)) {
    species_table <- refseq_table[which(tolower(
      refseq_table$organism_name) %in% tolower(taxon)), ]
  } else {
    species_table <- refseq_table[which(tolower(
      refseq_table$organism_name) %in% tolower(children_list)), ]
  }
  # Reduce table size based on reference or representative
  if (representative) reference <- TRUE
  if (representative && reference) {
    species_table <- species_table[
      species_table$refseq_category %in% c("reference genome",
                                           "representative genome"), ]
  } else if (!representative && reference) {
    species_table <- species_table[
      species_table$refseq_category == "reference genome", ]
  }
  return(species_table)
}

# Helper function to get cache for BiocFileCache
.get_cache <- function() {
  cache <- tools::R_user_dir("MetaScope", which = "cache")
  BiocFileCache::BiocFileCache(cache)
}


# Another helper function to download genomes
download_genomes <- function(species_table, taxon, patho_out, compress,
                             out_dir, caching, quiet) {
  total_genomes <- nrow(species_table)
  if (total_genomes == 0) {
    stop("No available genome for ", taxon,
         " - try setting both `representative` and `reference`",
         " to FALSE to obtain genome from the nucleotide database.")
  } else {
    if (!quiet) message("Downloading ", total_genomes, " ", taxon,
                        " genome(s) from NCBI")
    # Delete existing genome files and combined fasta
    taxon <- taxon %>%
      stringr::str_replace_all("/", "_") %>%
      stringr::str_replace_all(" ", "_")
    download_dir <- file.path(out_dir, paste(taxon, "refseq_download",
                                             sep = "_"))
    # Remove any existing files/directories
    if (file.exists(download_dir)) {
      unlink(file.path(download_dir, "*")) %>% message()
      unlink(download_dir, recursive = TRUE, force = TRUE) %>% message()
    }
    if (compress) {
      combined_fasta <- file.path(out_dir, paste(taxon, "fasta.gz", sep = "."))
      combined_fasta_patho <- file.path(out_dir,
                                        paste(taxon, "pathoscope.fasta.gz",
                                              sep = "."))
    } else {
      combined_fasta <- file.path(out_dir, paste(taxon, "fasta", sep = "."))
      combined_fasta_patho <- file.path(
        out_dir, paste(taxon, "pathoscope.fasta", sep = "."))
    }
    # Start with a new combined file
    if (file.exists(combined_fasta)) unlink(combined_fasta, force = TRUE)
    if (file.exists(combined_fasta_patho)) unlink(combined_fasta_patho,
                                                  force = TRUE)
  }
  # Download the genome
  for (i in seq_len(nrow(species_table))) {
    tryCatch({
      genome_file <- paste(basename(as.character(
        species_table[i, ]$ftp_path)), "genomic.fna.gz", sep = "_")
      location <- paste(species_table[i, ]$ftp_path, genome_file,
                        sep = "/")
      if (!caching) {
        if (!dir.exists(download_dir)) dir.create(download_dir)
        destination <- paste(download_dir, genome_file, sep = "/")
        utils::download.file(location, destination)
      } else if (caching) {
        bfc <- .get_cache()
        rid <- BiocFileCache::bfcquery(bfc, genome_file, "rname")$rid
        if (!length(rid)) {
          if (i %% 10 == 0 && !quiet) {
            message("Number of Genomes Downloaded: ", i, "/",
                    total_genomes, " (",
                    round(100 * i / total_genomes, 2), "%)")
          }
          rid <- names(BiocFileCache::bfcadd(bfc, genome_file, location))
        }
        if (!isFALSE(BiocFileCache::bfcneedsupdate(bfc, rid))) {
          BiocFileCache::bfcdownload(bfc, rid)
          BiocFileCache::bfcrpath(bfc, rids = rid)
        } else {
          if (!quiet) message("Caching is set to TRUE, ",
                              "and it appears that this file is already downloaded ",
                              "in the cache. It will not be downloaded again.")
        }
        destination <- BiocFileCache::bfcrpath(bfc, rids = rid)
      }
      # Write genome to concatenated file
      in_con <- file(destination, open = "rb")
      ref <- gzcon(in_con) %>% base::readLines()
      close(in_con)
      out_con <- gzfile(combined_fasta, open = "a")  # "a" for append
      writeLines(ref, out_con)
      close(out_con)
      # Format for pathoscope and write to file
      if (patho_out) {
        # Identify accessions
        ind <- ref %>% stringr::str_starts(">")
        accession <- ref[ind] %>% stringr::str_split(pattern = " ") %>%
          vapply(S4Vectors::head, n = 1, FUN.VALUE = character(1)) %>%
          stringr::str_remove_all(">")
        ref[ind] <- paste(">ti|", species_table[i, ]$taxid, "|org|",
                          gsub(" ", "_", species_table[i, ]$organism_name),
                          "|accession|", accession, sep = "")
        out_con <- gzfile(combined_fasta_patho, open = "a")  # "a" for append
        writeLines(ref, out_con)
        close(out_con)
      }
    }, error = function(e) message(conditionMessage(e)))
  }
  unlink(file.path(download_dir, "*"), force = TRUE)
  unlink(file.path(download_dir), recursive = TRUE, force = TRUE)
  if (!quiet) message("DONE! ", i, " genome(s) saved to file ",
                      file.path(combined_fasta))
  return(combined_fasta)
}

find_strains <- function(intable) {
  all_rank <- which(intable$rank == "no rank")
  if (all(is.na(all_rank))) return(intable)
  for (t_n in all_rank[all_rank > 1]) {
    if (intable$rank[t_n] == "no rank" && intable$rank[t_n - 1] == "species") {
      intable$rank[t_n] <- "strain"
    }
  }
  return(intable)
}

#' Download RefSeq genome libraries
#'
#' This function will automatically download RefSeq genome libraries in a fasta
#' format from the specified taxon. The function will first download the
#' summary report at:
#' \code{ftp://ftp.ncbi.nlm.nih.gov/genomes/refseq/**kingdom**/assembly_summary.txt},
#' and then use this file to download the genome(s) and combine them in a
#' single compressed or uncompressed .fasta file.
#'
#' When selecting the \code{taxon} to be downloaded, if you receive an error
#' saying \code{Your input is not a valid taxon}, please take a look at the
#' \code{taxonomy_table} object, which can be accessed with the command
#' \code{MetaScope:::taxonomy_table)}. Only taxa with exact spelling as they appear
#' at any level of the table will be acknowledged.
#'
#' @param taxon Name of single taxon to download. The taxon name should be a
#'   recognized NCBI scientific or common name, with no grammatical or
#'   capitalization inconsistencies. All available taxonomies are visible by
#'   accessing the \code{MetaScope:::taxonomy_table} object included in the package.
#' @param reference Download only RefSeq reference genomes? Defaults to
#'   \code{TRUE}. Automatically set to \code{TRUE} if \code{representative =
#'   TRUE}.
#' @param representative Download RefSeq representative and reference genomes?
#' Defaults to \code{FALSE}. If \code{TRUE}, reference is automatically set at
#'   \code{TRUE}.
#' @param compress Compress the output .fasta file? Defaults to \code{TRUE}.
#' @param patho_out Create duplicate outpute files compatible with PathoScope?
#'   Defaults to \code{FALSE}.
#' @param out_dir Character string giving the name of the directory to which
#'   libraries should be output. Defaults to creation of a new temporary
#'   directory.
#' @param caching Whether to use BiocFileCache when downloading genomes.
#'   Default is \code{FALSE}.
#' @param accession_path (character) Filepath to NCBI accessions SQL
#'   database. See \code{taxonomzr::prepareDatabase()}.
#' @param quiet Turns off most messages. Default is \code{TRUE}.
#'
#' @return Returns a .fasta or .fasta.gz file of the desired RefSeq genomes.
#' This file is named after the kingdom selected and saved to the current
#' directory (e.g. 'bacteria.fasta.gz'). This function also has the option
#' to return a .fasta file formatted for PathoScope as well
#' (e.g. bacteria.pathoscope.fasta.gz') if \code{path_out = TRUE}.
#'
#' @export
#' @examples
#' #### Download RefSeq genomes
#'
#' ## Create temporary taxonomizr accession
#' tmp_accession <- system.file("extdata", "example_accessions.sql", package = "MetaScope")
#'
#' ## Download all RefSeq reference Bovismacovirus genus genomes
#' download_refseq('Bovismacovirus', reference = FALSE, representative = FALSE,
#'                 out_dir = NULL, compress = TRUE, patho_out = FALSE,
#'                 caching = TRUE, accession_path = tmp_accession)

download_refseq <- function(taxon, reference = TRUE, representative = FALSE,
                            compress = TRUE, patho_out = FALSE,
                            out_dir = NULL, caching = FALSE,
                            quiet = TRUE, accession_path = NULL) {
  if (is.null(out_dir)) {
    out_dir <- tempfile()
    dir.create(out_dir)
  }
  if (is.null(accession_path)) {
    stop("Process halted. Requires an taxonomizr accession_path")
  }
  taxon_id <- taxonomizr::getId(taxon, sqlFile = accession_path)
  classification_table <- taxonomizr::getRawTaxonomy(taxon_id, accession_path)
  classification_table <- classification_table[[1]] |> as.data.frame() |>
    tibble::rownames_to_column(var = "rank")
  colnames(classification_table) <- c("rank", "name")
  rank_input <- classification_table$rank[1]
  parent_kingdom <- id_kingdom_rank(classification_table, taxon, rank_input, quiet)
  refseq_table <- download_parentkingdom(parent_kingdom, quiet)
  # Get NCBI scientific names of children species or strains
  taxonomy_table <- get0("taxonomy_table", envir = asNamespace("MetaScope"))
  if (rank_input == "no rank") base::stop("No rank detected")
  children_list <- get_children(taxon, rank_input, tax_dat = taxonomy_table)
  species_table <- get_speciestab(children_list, refseq_table, taxon,
                                  representative, reference, quiet)
  combined_fasta <- download_genomes(species_table, taxon, patho_out,
                                     compress, out_dir, caching, quiet)
  return(combined_fasta)
}

# Helper function to create a taxonomy table
mk_table <- function(intable, taxon_ranks = c("superkingdom", "kingdom", "phylum", "class", "order",
                                              "family", "genus", "species", "strain")) {
  this_t <- as.data.frame(intable)
  if (!identical(this_t$rank, character(0))) {
    t_n <- nrow(this_t)
    if (this_t$rank[t_n] == "no rank" && this_t$rank[t_n - 1] == "species") {
      this_t$rank[t_n] <- "strain"
    }
    final_tab <- this_t %>%
      dplyr::filter(.data$rank %in% taxon_ranks) %>%
      dplyr::right_join(dplyr::tibble(rank = taxon_ranks),
                        by = "rank") %>%
      dplyr::arrange(factor(.data$rank, levels = taxon_ranks)) %>%
      dplyr::select(.data$name)
    final_tab[seq_along(taxon_ranks), ] %>%
      magrittr::set_names(taxon_ranks) %>% return()
  }
}

#' Helper function for demultiplexing
#'
#' Helper function for demultiplexing sequencing reads, designed in a way to
#' allow for parallelization across barcodes (parallel extraction of reads by
#' barcode). This function takes a specific barcode (numeric index) from lists
#' of sample names/barcodes, a \code{Biostrings::DNAStringSet} of barcodes by
#' sequence header, and a \code{Biostrings::QualityScaledXStringSet} of reads
#' corresponding to the barcodes. Based on the barcode index given, it extracts
#' all reads for the indexed barcode and writes all the reads from that barcode
#' to a separate .fastq file.
#'
#' @param barcodeIndex Which barcode (integer number or index) in the barcodes
#'   or sample name to use for read extraction.
#' @param barcodes A list of all barcodes in the sequencing dataset. Correlates
#'   and in same order as \code{sampleNames}.
#' @param sampleNames A list of sample names or identifiers associated with each
#'   barcode in the barcodes list.
#' @param index A \code{Biostrings::DNAStringSet} that contains the read headers
#'   and barcode sequence for each header in the sequence slot.
#' @param reads A \code{Biostrings::QualityScaledXStringSet} that has the same
#'   headers and order as the index file, but contains the read sequences and
#'   their quality scores.
#' @param location A directory location to store the demultiplexed read files.
#'   Defaults to generate a new subdirectory at './demultiplex_fastq'
#' @param rcBarcodes Should the barcode indices in the barcodes list be reverse
#'   complemented to match the sequences in the index DNAStringSet? Defaults to
#'   \code{TRUE}.
#' @param hDist Uses a Hamming Distance or number of base differences to allow
#'   for inexact matches for the barcodes/indexes. Defaults to 0. Warning: if
#'   the Hamming Distance is >=1 and this leads to inexact index matches to more
#'   than one barcode, that read will be written to more than one demultiplexed
#'   read files.
#' @param quiet Turns off most messages. Default is \code{TRUE}.
#'
#' @return Writes a single .fastq file that contains all reads whose index
#'   matches the barcode specified. This file will be written to the location
#'   directory, and will be named based on the specified sampleName and barcode,
#'   e.g. './demultiplex_fastq/SampleName1_GGAATTATCGGT.fastq.gz'
#' @export
#' @examples
#'
#' ## Create temporary directory
#' ref_temp <- tempfile()
#' dir.create(ref_temp)
#'
#' ## Load example barcode, index, and read data into R session
#' barcodePath <- system.file("extdata", "barcodes.txt", package = "MetaScope")
#' bcFile <- read.table(barcodePath, sep = "\t", header = TRUE)
#'
#' indexPath <- system.file("extdata", "virus_example_index.fastq",
#' package = "MetaScope")
#' inds <- Biostrings::readDNAStringSet(indexPath, format = "fastq")
#'
#' readPath <- system.file("extdata", "virus_example.fastq",
#'                         package = "MetaScope")
#' reads <- Biostrings::readQualityScaledDNAStringSet(readPath)
#'
#' ## Extract reads from the first barcode
#' results <- extract_reads(1, bcFile[, 2], bcFile[, 1], inds, reads,
#'                         rcBarcodes = FALSE, location = ref_temp)
#'
#' ## Extract reads from multiple barcodes
#' more_results <- lapply(1:6, extract_reads, bcFile[, 2], bcFile[, 1], inds,
#'                        reads, rcBarcodes = FALSE, location = ref_temp)
#'
#' ## Remove temporary directory
#' unlink(ref_temp, recursive = TRUE)
#'

extract_reads <- function(barcodeIndex, barcodes, sampleNames, index, reads,
                          location = "./demultiplex_fastq", rcBarcodes = TRUE,
                          hDist = 0, quiet = TRUE) {
  barcode <- barcodes[barcodeIndex]
  sampleName <- sampleNames[barcodeIndex]
  if (!quiet) message("Finding reads for barcode: ", barcode)
  if (rcBarcodes) {
    rci <- as.character(Biostrings::reverseComplement(
      Biostrings::DNAString(barcode)))
  } else {
    rci <- barcode
  }
  ind_match <- utils::adist(as.character(index), rci) <= hDist
  numReads <- sum(ind_match)
  outFileName <- paste(location, "/", sampleName, "_", barcode,
                       ".fastq.gz", sep = "")
  if (numReads == 0 && !quiet) {
    message("\tFound 0 reads for this barcode; no file will be written.")
  } else {
    if (!quiet) message("\tFound ", sum(ind_match),
                        " reads, writing reads to: ", outFileName)
    Biostrings::writeQualityScaledXStringSet(
      reads[c(ind_match)], outFileName, compress = TRUE)
  }
  return(list(output_file = outFileName, numberOfReads = numReads,
              matchedIndexes = ind_match))
}

errmessages <- function(barcodes, samNames, numReads) {
  warning("Did not find any reads for the following barcodes: ",
          paste(barcodes[numReads == 0], collapse = " "))
  warning("Did not find any reads for the following samples: ",
          paste(samNames[numReads == 0], collapse = " "))
  write(paste("Did not find any reads for the following barcodes:",
              paste(barcodes[numReads == 0], collapse = " "), "\n",
              "Did not find any reads for the following samples: ",
              paste(samNames[numReads == 0], collapse = " ")),
        file = "demultiplex_fastq/unmapped_barcodes_samples.txt")
}

#' Demultiplexing sequencing reads
#'
#' Function for demultiplexing sequencing reads arranged in a common format
#' provided by sequencers (such as Illumina) generally for 16S data. This
#' function takes a matrix of sample names/barcodes, a .fastq file of barcodes
#' by sequence header, and a .fastq file of reads corresponding to the barcodes.
#' Based on the barcodes given, the function extracts all reads for the indexed
#' barcode and writes all the reads from that barcode to separate .fastq files.
#' @param barcodeFile Path to a file containing a .tsv matrix with a
#'   header row, and then sample names (column 1) and barcodes (column 2).
#' @param indexFile Path to a .fastq file that contains the barcodes for
#'   each read. The headers should be the same (and in the same order) as
#'   \code{readFile}, and the sequence in the \code{indexFile} should be the
#'   corresponding barcode for each read. Quality scores are not considered.
#' @param readFile Path to the sequencing read .fastq file that corresponds
#'   to the \code{indexFile}.
#' @param rcBarcodes Should the barcode indexes in the barcodeFile be reverse
#'   complemented to match the sequences in the \code{indexFile}? Defaults to
#'   \code{TRUE}.
#' @param location A directory location to store the demultiplexed read files.
#'   Defaults to generate a new temporary directory.
#' @param threads The number of threads to use for parallelization
#'   (BiocParallel). This function will parallelize over the barcodes and
#'   extract reads for each barcode separately and write them to separate
#'   demultiplexed files.
#' @param hammingDist Uses a Hamming Distance or number of base differences to
#'   allow for inexact matches for the barcodes/indexes. Defaults to \code{0}.
#'   Warning: if the Hamming Distance is \code{>=1} and this leads to inexact
#'   index matches to more than one barcode, that read will be written to more
#'   than one demultiplexed read files.
#' @param quiet Turns off most messages. Default is \code{TRUE}.
#'
#' @return Returns multiple .fastq files that contain all reads whose index
#'   matches the barcodes given. These files will be written to the location
#'   directory, and will be named based on the given sampleNames and barcodes,
#'   e.g. './demultiplex_fastq/SampleName1_GGAATTATCGGT.fastq.gz'
#'
#' @export
#'
#' @examples
#'
#' ## Get barcode, index, and read data locations
#' barcodePath <- system.file("extdata", "barcodes.txt", package = "MetaScope")
#' indexPath <- system.file("extdata", "virus_example_index.fastq",
#'                          package = "MetaScope")
#' readPath <- system.file("extdata", "virus_example.fastq",
#'                          package = "MetaScope")
#'
#' ## Demultiplex
#' demult <- meta_demultiplex(barcodePath, indexPath, readPath, rcBarcodes = FALSE,
#'                       hammingDist = 2)
#' demult
#'

meta_demultiplex <- function(barcodeFile, indexFile, readFile, rcBarcodes = TRUE,
                             location = NULL, threads = 1,
                             hammingDist = 0, quiet = TRUE) {
  if (is.null(location)) {
    location <- tempfile()
    dir.create(location)
  }
  if (!quiet) message("Reading Sample Names and Barcodes from: ",
                      barcodeFile)
  bcFile <- utils::read.table(barcodeFile, sep = "\t", header = TRUE)
  barcodes <- bcFile[, 2]
  samNames <- bcFile[, 1]
  if (!quiet) message("\tFound information for ", length(barcodes),
                      " samples/barcodes")
  if (!quiet) message("Reading Index File: ", indexFile)
  inds <- Biostrings::readDNAStringSet(indexFile, format = "fastq")
  if (!quiet) message("\tFound indexes for ", length(inds), " reads")
  if (!quiet) message("Reading Sequence File: ", readFile)
  reads <- Biostrings::readQualityScaledDNAStringSet(readFile)
  if (!quiet) message("\tFound ", length(reads), " reads")
  ## make output directory if necessary
  if (!dir.exists(location)) dir.create(location)
  # Loop over barcodes
  numReads <- NULL
  ind_no_match <- numeric(length(reads))
  for (i in seq_along(barcodes)) {
    extracted <- extract_reads(
      i, barcodes, samNames, inds, reads, rcBarcodes = rcBarcodes,
      location = location, hDist = hammingDist, quiet = quiet)
    numReads <- c(numReads, extracted$numberOfReads)
    ind_no_match <- ind_no_match + extracted$matchedIndexes
  }
  if (!quiet) message(sum(ind_no_match > 1))
  ind_no_match <- (ind_no_match == 0)
  # number of reads for each barcode
  if (any(numReads == 0)) {
    errmessages(barcodes, samNames, numReads)
  }
  # Track reads without matches, and write them to an 'orphan' file
  if (!quiet) message("Found ", sum(ind_no_match),
                      " reads without a matching barcode (",
                      100 * round(mean(ind_no_match), 4), "%), writing reads to: ",
                      location, "/orphans.fastq.gz")
  Biostrings::writeQualityScaledXStringSet(
    reads[c(ind_no_match)], paste(location, "/orphans.fastq.gz",
                                  sep = ""), compress = TRUE)
  summaryMat <- cbind(bcFile[seq_along(barcodes), ],
                      NumberOfReads = numReads)
  utils::write.table(summaryMat,
                     file = paste(location, "/summary.txt", sep = ""),
                     col.names = FALSE, row.names = TRUE, quote = FALSE)
  return(summaryMat)
}

#' Get child nodes from NCBI taxonomy
#'
#' This function will utilize an organism classification table to obtain all
#' children species and/or strains with available NCBI reference sequences given
#' a parent taxon and its rank.
#'
#' @param input_taxon The parent taxon.
#' @param input_rank The taxonomic rank of the input taxon.
#' @param tax_dat A dataframe of organism classification information.
#' At minimum, should have a column indicating "strain", and and all others
#' should be taxonomic ranks. Each row should be a taxonomic relationship.
#' This defaults to \code{NULL}, which calls the `taxonomy_table` object.
#'
#' @return Returns a vector of all the child species and/or strains of the
#' input taxon.
#'
#' @export
#'
#' @examples
#' ## Get all child species and strains in bacteria superkingdom
#' get_children('Bacteria','superkingdom')
#'
#' ## Get all child species and strains in fungi kingdom
#' get_children('Fungi', 'kingdom')
#'
#' ## Get all child species in primate order
#' get_children('Primates', 'order')
#'

get_children <- function(input_taxon, input_rank, tax_dat = NULL) {
  if (is.null(tax_dat)) tax_dat <- taxonomy_table
  # Get child strains
  ind <- tolower(tax_dat[, input_rank]) %in% tolower(input_taxon)
  strain_list <- unique(tax_dat[, "strain"][ind])
  strain_list <- strain_list[!is.na(strain_list)]
  ## Delete rows with child strains
  new_table <- tax_dat[!(tax_dat[, "strain"] %in% strain_list), ]
  ind <- tolower(new_table[, input_rank]) %in% tolower(input_taxon)
  species_list <- unique(new_table[, "species"][ind])
  species_list <- species_list[!is.na(species_list)]
  children_list <- c(strain_list, species_list)
  return(children_list)
}

#' Filter unmapped reads
#'
#' This function will remove all unmapped reads or lines in a .bam file
#' (warning: overwrites the original file!). This function is needed because
#' combining multiple .bam files from different microbial libraries may lead to
#' some reads that mapped to one library and have unmapped entries from another
#' library. This will remove any unmapped entries and leave all reference mapped
#' lines in the .bam file.
#'
#' It is not intended for direct use.
#'
#' @param bamfile Location for the .bam file to filter & remove all unmapped
#'   reads
#'
#' @return This function will overwrite the existing .bam file with a new .bam
#' file in the same location that has only mapped lines. The function itself
#' returns the output .bam file name.
#'

filter_unmapped_reads <- function(bamfile) {
  sorted_bamfile <- Rsamtools::sortBam(bamfile,
                                       paste(tools::file_path_sans_ext(bamfile),
                                             ".sorted", sep = ""))
  bam_index <- Rsamtools::indexBam(sorted_bamfile)
  filtered_bam <- Rsamtools::filterBam(
    sorted_bamfile,
    destination = bamfile,
    index = bam_index,
    indexDestination = FALSE,
    param = Rsamtools::ScanBamParam(flag = Rsamtools::scanBamFlag(
      isUnmappedQuery = FALSE))
  )
  # clean up
  file.remove(sorted_bamfile)
  file.remove(bam_index)
  # return filtered file name
  return(filtered_bam)
}

#' Create a combined .bam header
#'
#' This function generates a combined header from multiple .bam files from
#' different reference libraries (e.g. a split bacterial library). It is not
#' intended for use by users.
#'
#' @param bam_files A character vector of the locations/file names of .bam files
#'   from which to combine the headers.
#' @param header_file A file name and location for the output file for the
#'   combined header. This will be a .sam format file without any reads.
#'   Defaults to 'header_tmp.sam'.
#'
#' @return This function will return a combined header from all the supplied
#' .bam files.
#'

combined_header <-
  function(bam_files, header_file = "header_tmp.sam") {
    # get first and last line of header
    bam_head <- Rsamtools::scanBamHeader(bam_files[1])
    n <- length(bam_head[[1]]$text)
    last <- c(names(bam_head[[1]]$text)[n], bam_head[[1]]$text[[n]])
    # open and print the first line of header
    head_con <- file(header_file, open = "w")
    writeLines(c(names(bam_head[[1]]$text)[1], bam_head[[1]]$text[[1]]),
               con = head_con,
               sep = "\t")
    writeLines("\n", con = head_con, sep = "")
    # print genomes from all .bam files
    for (bfile in bam_files) {
      bam_head <- Rsamtools::scanBamHeader(bfile)
      for (j in 2:(length(bam_head[[1]]$text) - 1)) {
        writeLines(c(names(bam_head[[1]]$text)[j], bam_head[[1]]$text[[j]]),
                   con = head_con,
                   sep = "\t")
        writeLines("\n", con = head_con, sep = "")
      }
    }
    writeLines(last, con = head_con, sep = "\t")
    close(head_con)
    return(header_file)
  }

#' Replace the header from a .bam file
#'
#' This function replaces the header from one .bam file with a header from a
#' different .sam file. This function mimics the function of the 'reheader'
#' function in samtools. It is not intended for use by users.
#'
#' @param head A file name and location for the .sam file with the new header.
#' @param old_bam A file name and location for the .bam file which you would
#' @param new_bam A file name for the new .bam file with a replaced header.
#'   Defaults to the same base filename plus 'h.bam'. For example, 'example.bam'
#'   will be written as 'exampleh.bam'.
#'
#' @return This function will return a new .bam file with a replaced header. The
#' function also outputs the new .bam filename.
#'

bam_reheader_R <- function(head, old_bam, new_bam = paste(
  tools::file_path_sans_ext(old_bam), "h.bam", sep = "")) {
  new_sam <- paste(tools::file_path_sans_ext(new_bam), ".sam", sep = "")
  new_sam_con <- file(new_sam, open = "w")
  head_con <- file(head, open = "r")
  while (length(oneLine <- readLines(head_con, n = 1, warn = FALSE)) > 0) {
    writeLines(oneLine, new_sam_con)
  }
  close(head_con)
  old_sam <- Rsamtools::asSam(old_bam, overwrite = TRUE)
  old_sam_con <- file(old_sam, open = "r")
  while (length(oneLine <- readLines(old_sam_con, n = 1, warn = FALSE)) > 0) {
    if (substr(oneLine, 1, 1) != "@") {
      writeLines(oneLine, new_sam_con)
    }
  }
  close(new_sam_con)
  close(old_sam_con)
  file.remove(old_sam)
  new_bam <- Rsamtools::asBam(new_sam, overwrite = TRUE)
  file.remove(new_sam)
  file.remove(paste(new_bam, ".bai", sep = ""))
  return(new_bam)
}

#' Merge multiple .bam files
#'
#' This function merges .bam files. It first used the combined_header function
#' to generate a combined header for all the files, reheaders the files, and
#' then merges and sorts the .bam files. It is similar to the 'samtools merge'
#' function, but it allows the .bam files to have different headers. It is not
#' intended for direct use.
#'
#' @param bam_files A list of file names for the .bam files to be merged.
#' @param destination A file name and location for the merged .bam file.
#' @param head_file A file name and location for the combined header file.
#'   Defaults to the destination. For example, 'example.bam' will be written as
#'   'example.bam'.
#' @param quiet Turns off most messages. Default is \code{TRUE}.
#'
#' @return This function merges .bam files and combines them into a single file.
#' The function also outputs the new .bam filename.
#'

merge_bam_files <- function(bam_files, destination,
                            head_file = paste(destination, "_header.sam",
                                              sep = ""), quiet = TRUE) {
  if (!quiet) message("Merging .bam files")
  com_head <- combined_header(bam_files, header_file = head_file)
  # Paths for merged bam files
  unsortbam <- paste0(destination, "_unsorted.bam")
  if (check_samtools_exists()) {
    if (!quiet) message("samtools found on system. Merging with samtools.")
    lapply(bam_files, function(bf) sys::exec_wait(
      "samtools", c("reheader", head_file, bf), std_out = FALSE,
      std_err = TRUE))
    sys::exec_wait("samtools", c("merge", unsortbam, bam_files),
                   std_out = FALSE, std_err = TRUE)
    lapply(c(bam_files, head_file), unlink, force = TRUE, recursive = TRUE)
    if (!quiet) message("Sorting merged bam file")
    merged_bam_sorted <- paste0(destination, ".bam")
    sys::exec_wait("samtools", c("sort", "-o", merged_bam_sorted, unsortbam),
      std_out = TRUE, std_err = TRUE)
    lapply(unsortbam, unlink, force = TRUE, recursive = TRUE)
  } else {
    if (!quiet) message("samtools not found. Merging with Rsamtools instead.")
    bam_files_h <- NULL
    for (i in seq_along(bam_files)) {
      if (!quiet) message("Reheading bam file")
      new_bam_h <- bam_reheader_R(com_head, bam_files[i])
      bam_files_h <- c(bam_files_h, new_bam_h)
      file.remove(bam_files[i])
      # remove .bam and .vcf and .bam.summary files for each alignment
      file.remove(paste(bam_files[i], ".indel.vcf", sep = ""))
      file.remove(paste(bam_files[i], ".summary", sep = ""))
    }
    merged_bam <- Rsamtools::mergeBam(bam_files_h, unsortbam, overwrite = TRUE)
    # clean up
    file.remove(com_head)
    lapply(bam_files_h, file.remove)
    if (!quiet) message("Sorting merged bam file")
    # sort merged bam file
    merged_bam_sorted <- Rsamtools::sortBam(merged_bam, destination)
    file.remove(merged_bam)
  }
  # return merged and sorted bam
  return(merged_bam_sorted)
}

#' Align microbiome reads to a set of reference libraries
#'
#' This is the main MetaScope target library mapping function, using Rsubread
#' and multiple libraries. Aligns to each library separately, filters unmapped
#' reads from each file, and then merges and sorts the .bam files from each
#' library into one output file. If desired, output can be passed to
#' `filter_host()` to remove reads that also map to filter library genomes.
#'
#' @inheritParams align_target_bowtie
#' @param lib_dir Path to the index files for all libraries.
#' @param libs A vector of character strings giving the basenames of the Subread
#'   index files for alignment. If ALL indices to be used are located in the
#'   current working directory, set \code{lib_dir = NULL}. Default is
#'   \code{lib_dir = NULL}.
#' @param subread_options A named \code{list} specifying alignment parameters
#'   for the \code{Rsubread::align()} function, which is called inside
#'   \code{align_target()}. Elements should include type, nthreads,
#'   maxMismatches, nsubreads, phredOffset, unique, and nBestLocations.
#'   Descriptions of these parameters are available under
#'   \code{?Rsubread::align}. Defaults to the global \code{align_details}
#'   object.
#'
#' @return This function writes a merged and sorted .bam file after aligning to
#' all reference libraries given, along with a summary report file, to the
#' user's working directory. The function also outputs the new .bam filename.
#'
#' @export
#'
#' @examples
#' #### Align example reads to an example reference library using Rsubread
#' \donttest{
#' ## Create temporary directory
#' target_ref_temp <- tempfile()
#' dir.create(target_ref_temp)
#'
#' tax <- "Ovine atadenovirus D"
#'
#' ## Create temporary taxonomizr accession
#' tmp_accession <- system.file("extdata", "example_accessions.sql", package = "MetaScope")
#'
#' ## Download genome
#' all_ref <- MetaScope::download_refseq(tax,
#'                                       reference = FALSE,
#'                                       representative = FALSE,
#'                                       compress = TRUE,
#'                                       out_dir = target_ref_temp,
#'                                       caching = TRUE,
#'                                       accession_path = tmp_accession)
#'
#' ## Create subread index
#' ind_out <- mk_subread_index(all_ref)
#'
#' ## Get path to example reads
#' readPath <- system.file("extdata", "reads.fastq",
#'                         package = "MetaScope")
#' ## Copy the example reads to the temp directory
#' refPath <- file.path(target_ref_temp, "reads.fastq")
#' file.copy(from = readPath, to = refPath)
#'
#' ## Modify alignment parameters object
#' data("align_details")
#' align_details[["type"]] <- "rna"
#' align_details[["phredOffset"]] <- 50
#' # Just to get it to align - toy example!
#' align_details[["maxMismatches"]] <- 100
#'
#' ## Run alignment
#' target_map <- align_target(refPath,
#'                            libs = stringr::str_replace_all(tax, " ", "_"),
#'                            lib_dir = target_ref_temp,
#'                            subread_options = align_details)
#'
#' ## Remove temporary folder
#' unlink(target_ref_temp, recursive = TRUE)
#' }

align_target <- function(read1, read2 = NULL, lib_dir = NULL, libs,
                         threads = 1,
                         align_file = tools::file_path_sans_ext(read1),
                         subread_options = align_details, quiet = TRUE) {
  data_env <- new.env(parent = emptyenv())
  utils::data("align_details", envir = data_env, package = "MetaScope")
  align_details <- data_env[["align_details"]]
  if (!is.null(lib_dir)) lib_dir <- tools::file_path_as_absolute(lib_dir)
  bam_files <- numeric(length(libs))
  for (i in seq_along(libs)) {
    if (!quiet) message("Attempting to perform subread alignment on ", libs[i],
            " index")
    bam_files[i] <- paste(tools::file_path_sans_ext(read1), ".", libs[i],
                          ".bam", sep = "")
    Rsubread::align(
      index = file.path(lib_dir, libs[i]),
      readfile1 = read1,
      readfile2 = read2,
      input_format = "fastq",
      output_format = "BAM",
      output_file = bam_files[i],
      type = subread_options[["type"]],
      nthreads = threads,
      maxMismatches = subread_options[["maxMismatches"]],
      nsubreads = subread_options[["nsubreads"]],
      phredOffset = subread_options[["phredOffset"]],
      unique = subread_options[["unique"]],
      nBestLocations = subread_options[["nBestLocations"]]
    )
    ## remove umapped reads
    if (!quiet) message("Filtering unmapped reads")
    filter_unmapped_reads(bam_files[i])
  }
  if (!quiet) message("Library alignment complete")
  # If more than one library was aligned, then combine bam files
  if (length(bam_files) > 1) {
    if (quiet) message("Merging the bam files into ", align_file, ".bam")
    merge_bam_files(bam_files, align_file, quiet=quiet)
  } else {
    file.rename(bam_files, paste(align_file, ".bam", sep = ""))
    # remove Rsubread .vcf and .bam.summary files for now
    file.remove(paste(bam_files, ".indel.vcf", sep = ""))
    file.remove(paste(bam_files, ".summary", sep = ""))
  }
  if (!quiet) message("DONE! Alignments written to ", align_file, ".bam")
  return(paste(align_file, ".bam", sep = ""))
}

#' Align microbiome reads to set of indexed Bowtie2 libraries
#'
#' This is the main MetaScope target library mapping function, using Rbowtie2
#' and multiple libraries. Aligns to each library separately, filters unmapped
#' reads from each file, and then merges and sorts the .bam files from each
#' library into one output file. If desired, output can be passed to
#' `filter_host_bowtie()` to remove reads that also map to filter library
#' genomes.
#'
#' The default parameters are the same that PathoScope 2.0 uses.
#' "--very-sensitive-local -k 100 --score-min L,20,1.0"
#'
#' If you experience any issues with reading the input files, make sure that
#' the file(s) are not located in a read-only folder. This can be circumvented
#' by copying files to a new location before running the function.
#'
#' @param read1 Path to the .fastq file to align.
#' @param read2 Optional: Location of the mate pair .fastq file to align.
#' @param lib_dir Path to the directory that contains the Bowtie2 indexes.
#' @param libs The basename of the Bowtie2 indexes to align against (without
#'   trailing .bt2 or .bt2l extensions).
#' @param align_dir Path to the directory where the output alignment file should
#'   be created.
#' @param align_file The basename of the output alignment file (without trailing
#'   .bam extension).
#' @param  bowtie2_options Optional: Additional parameters that can be passed to
#'   the align_target_bowtie() function. To see all the available parameters use
#'   Rbowtie2::bowtie2_usage(). See Details for default parameters. NOTE: Users
#'   should pass all their parameters as one string and if optional parameters
#'   are given then the user is responsible for entering all the parameters to
#'   be used by Bowtie2. The only parameter that should NOT be specified here is
#'   the number of threads.
#' @param threads The number of threads that can be utilized by the function.
#'   Default is 1 thread.
#' @param overwrite Whether existing files should be overwritten. Default is
#'   FALSE.
#' @param quiet Turns off most messages. Default is \code{TRUE}.
#'
#' @return Returns the path to where the output alignment file is stored.
#'
#' @export
#'
#' @examples
#' #### Align example reads to an example reference library using Rbowtie2
#' \donttest{
#' ## Create temporary directory to store file
#' target_ref_temp <- tempfile()
#' dir.create(target_ref_temp)
#'
#' tmp_accession <- system.file("extdata", "example_accessions.sql", package = "MetaScope")
#'
#' ## Dowload reference genome
#' MetaScope::download_refseq("Morbillivirus hominis",
#'                            reference = FALSE,
#'                            representative = FALSE,
#'                            compress = TRUE,
#'                            out_dir = target_ref_temp,
#'                            caching = TRUE,
#'                            accession_path = tmp_accession
#' )
#'
#' ## Create temporary directory to store the indices
#' index_temp <- tempfile()
#' dir.create(index_temp)
#'
#' ## Create bowtie2 index
#' MetaScope::mk_bowtie_index(
#'   ref_dir = target_ref_temp,
#'   lib_dir = index_temp,
#'   lib_name = "target",
#'   overwrite = TRUE
#' )
#'
#' ## Create temporary directory for final file
#' output_temp <- tempfile()
#' dir.create(output_temp)
#'
#' ## Get path to example reads
#' readPath <- system.file("extdata", "virus_example.fastq",
#'                         package = "MetaScope")
#'
#' ## Align to target genomes
#' target_map <-
#'   MetaScope::align_target_bowtie(
#'     read1 = readPath,
#'     lib_dir = index_temp,
#'     libs = "target",
#'     align_dir = output_temp,
#'     align_file = "bowtie_target",
#'     overwrite = TRUE,
#'     bowtie2_options = "--very-sensitive-local"
#'   )
#'
#' ## Remove extra folders
#' unlink(target_ref_temp, recursive = TRUE)
#' unlink(index_temp, recursive = TRUE)
#' unlink(output_temp, recursive = TRUE)
#' }

align_target_bowtie <- function(read1, read2 = NULL, lib_dir, libs,
                                align_dir, align_file, bowtie2_options = NULL,
                                threads = 1, overwrite = FALSE, quiet = TRUE) {
    # Convert user specified paths to absolute paths for debugging purposes
    lib_dir <- tools::file_path_as_absolute(lib_dir)
    align_dir <- tools::file_path_as_absolute(align_dir)
    # If user does not specify parameters, specify for them
    if (is.null(bowtie2_options)) {
      bowtie2_options <- paste(
        "--local -R 2 -N 0 -L 25 -i S,1,0.75 -k 5 --score-min L,0,1.7",
        "--threads", threads)
    } else bowtie2_options <- paste(bowtie2_options, "--threads", threads)
    bam_files <- numeric(length(libs))
    for (i in seq_along(libs)) {
      # Don't attach .bam extension - Rbowtie2 does this already
      bam_files[i] <- file.path(
        align_dir, paste(basename(tools::file_path_sans_ext(read1)),
                         ".", libs[i], sep = ""))
      if (!quiet) message("Attempting to perform Bowtie2 alignment on ",
                          libs[i], " index")
      Rbowtie2::bowtie2_samtools(
        bt2Index = file.path(lib_dir, libs[i]),
        output = bam_files[i],
        outputType = "bam",
        seq1 = read1,
        seq2 = read2,
        overwrite = overwrite,
        ... = bowtie2_options
      )
      unlink(".bowtie2.cerr.txt")
      # Attach .bam extension to bam files to call function
      if (!quiet) message("Filtering unmapped reads")
      filter_unmapped_reads(paste0(bam_files[i], ".bam"))
    }
    if (!quiet) message("Library alignment complete")
    outputFile <- file.path(align_dir, paste0(align_file, ".bam"))
    bam_files <- paste0(bam_files, ".bam")
    # If more than one library was aligned, then combine bam files
    if (length(bam_files) > 1) {
      if (!quiet) message("Merging the bam files into ", align_file, ".bam")
      merge_bam_files(bam_files, tools::file_path_sans_ext(outputFile), quiet=quiet)
    } else file.rename(bam_files, outputFile)
    if (!quiet) message("DONE! Alignments written to ", outputFile)
    return(outputFile)
  }

getExtension <- function(file){
  ex <- strsplit(basename(file), split="\\.")[[1]]
  return(ex[-1])
}

# download_accessions.R
#' Download indexes required for MetaScope ID and MetaBlast modules
#'
#' This is a necessary step for all samples utilizing NCBI and SILVA databases
#' in the MetaScope pipeline. As specified by the user,
#' \code{download_accessions} will automatically download the NCBI accessions
#' database, the SILVA taxonomy database, and or the NCBI Blast 16S database and
#' prepare consolidated databases for downstream use with the MetaID and
#' MetaBLAST modules. This package relies on the \code{taxonomizr} package.
#'
#' @param ind_dir Character string. Directory filepath where indices should be
#'   saved. Required.
#' @param tmp_dir Character path to directory for storing temp files. (Useful
#'   to avoid redownloading) Defaults to \code{file.path(ind_dir, "tmp")}
#' @param remove_tmp_dir Delete tmp_dir after downloads are complete? Defaults
#'   to \code{TRUE}
#' @param NCBI_accessions_database Logical. Download taxonomizr NCBI accessions
#'   database? Defaults to \code{TRUE}.
#' @param NCBI_accessions_name Character string. Filename (with or without
#'   extension) to save taxonomizr NCBI accessions database. Defaults to
#'   \code{"accessionTaxa.sql"}.
#' @param silva_taxonomy_database Logical. Download SILVA taxonomy database?
#'   Defaults to \code{TRUE}.
#' @param silva_taxonomy_name Character string. Filename (with or without
#'   extension) to save SILVA taxonomy database.
#'   Defaults to the file supplied with the package,
#'   \code{"all_silva_headers.rds"}.
#' @param blast_16S_database Logical. Download BLAST 16S database?
#'   Defaults to \code{TRUE}.
#' @param blast_16S_name Character string. Filename (without extension) to save \
#'   BLAST 16S database.
#'   Defaults to the file supplied with the package,
#'   \code{"16S_ribosomal_RNA"}.
#'
#' @return Exports database(s) with names and to location specified by the user.
#'
#' @export
#'
#' @examples
#' \dontrun{
#'   download_accessions(
#'     ind_dir = "C:/Users/JohnSmith/Research",
#'     tmp_dir = file.path(ind_dir, "tmp"),
#'     remove_tmp_dir = TRUE,
#'     NCBI_accessions_database = TRUE,
#'     NCBI_accessions_name = "accessionTaxa.sql",
#'     silva_taxonomy_database = TRUE,
#'     silva_taxonomy_name = "all_silva_headers.rds")
#' }
#'

download_accessions <- function(ind_dir,
                                tmp_dir = file.path(ind_dir, "tmp"),
                                remove_tmp_dir = TRUE,
                                NCBI_accessions_database = TRUE,
                                NCBI_accessions_name = "accessionTaxa",
                                silva_taxonomy_database = TRUE,
                                silva_taxonomy_name = "all_silva_headers",
                                blast_16S_database = TRUE,
                                blast_16S_name = "16S_ribosomal_RNA") {
  if (!file.exists(ind_dir)) {
    stop("ind_dir does not exist")
  }

  if (!file.exists(tmp_dir)) {
    dir.create(tmp_dir)
  }
  if (NCBI_accessions_database) {
    message("Downloading NCBI accessions database")
    # Check if NCBI_accessions_name has extension
    if (!identical(getExtension(NCBI_accessions_name), "sql")) {
      NCBI_accessions_name <- paste0(NCBI_accessions_name, ".sql")
    }
    taxonomizr::prepareDatabase(file.path(ind_dir, NCBI_accessions_name),
                                tmpDir = tmp_dir)
  }
  if (silva_taxonomy_database) {
    message("Downloading SILVA taxonomy database")
    location <- "https://github.com/wejlab/metascope-docs/raw/main/all_silva_headers.rds"
    # CHeck if SILVA_taxonomy_name has extension
    if (!identical(getExtension(silva_taxonomy_name), "rds")) {
      NCBI_accessions_name <- paste0(silva_taxonomy_name, ".rds")
    }
    destination <- paste(ind_dir, silva_taxonomy_name, sep = "/")
    utils::download.file(location, destination)
  }

  if (remove_tmp_dir) {
    unlink(tmp_dir)
  }

  message("Download completed!")
}

#' @name check_samtools_exists
#' @title Check if samtools exists on the system
#' @description This is an internal function that is not meant to be used
#'   outside of the package. It checks whether samtools exists on the system.
#' @return Returns TRUE if samtools exists on the system, else FALSE.
#'

check_samtools_exists <- function() {
  if (file.exists(Sys.which("samtools"))) return(TRUE)
  return(FALSE)
}

#' Gets Taxonomy IDS from accessions names

#' Returns a dataframe with the original accession, the associated taxonomy ID
#' and associated taxonomy species

#' @param all_fastas Biostrings loaded object of all fastas
#' @param accession_path Path to taxonomizr accessions database
#' @keywords internal
#' @return Data.frame of taxonomy Ids and Species name
#'

find_taxids <- function(all_fastas, accession_path) {
  if (grepl("ti|",names(all_fastas)[1], fixed = TRUE)){
    accessions_taxids <- tibble::tibble(
      "accessions" = unique(names(all_fastas))) |>
      dplyr::mutate(
        "TaxonomyID" = as.numeric(stringr::str_match(!!dplyr::sym("accessions"),
                                                     "ti\\|(\\d+)\\|")[,2]),
        "species" = taxonomizr::getTaxonomy(!!dplyr::sym("TaxonomyID"),
                                            sqlFile = accession_path,
                                            desiredTaxa = "species")[,]
      )
  } else {
    accessions_taxids <- tibble::tibble("accessions" = unique(names(all_fastas))) |>
      dplyr::mutate("TaxonomyID" = taxonomizr::accessionToTaxa(!!dplyr::sym("accessions"), accession_path),
                    "species" = taxonomizr::getTaxonomy(!!dplyr::sym("TaxonomyID"),
                                                        sqlFile = accession_path,
                                                        desiredTaxa = "species")[,])
  }
  return(accessions_taxids)
}

#' Adds in taxa if silva database

#' Returns MetaScope Table with silva taxa in separate columns

#' @param metascope_id_in MetaScope ID file with silva taxa
#' @param caching Boolean for if all_silva_headers.rds is already downloaded
#' @param path_to_write Path to save all_silva_headers.rds
#' @keywords internal
#' @return Data.frame of taxonomy information

add_in_taxa <- function(metascope_id_in, caching, path_to_write) {
  location <- "https://github.com/wejlab/metascope-docs/raw/main/all_silva_headers.rds"
  filename <- "all_silva_headers.rds"
  if (!caching) {
    if (!dir.exists(path_to_write)) dir.create(path_to_write)
    destination <- paste(path_to_write, filename, sep = "/")
    utils::download.file(location, destination)
  } else if (caching) {
    bfc <- .get_cache()
    rid <- BiocFileCache::bfcquery(bfc, filename, "rname")$rid
    if (!length(rid)) {
      rid <- names(BiocFileCache::bfcadd(bfc, filename, location))
    }
    if (!isFALSE(BiocFileCache::bfcneedsupdate(bfc, rid))) {
      BiocFileCache::bfcdownload(bfc, rid)
      BiocFileCache::bfcrpath(bfc, rids = rid)
    } else {
      message("Caching is set to TRUE, ",
              "and it appears that this file is already downloaded ",
              "in the cache. It will not be downloaded again.")
    }
    destination <- BiocFileCache::bfcrpath(bfc, rids = rid)
  }
  all_silva_headers <- readRDS(destination)
  tax_table_pre <- metascope_id_in |>
    dplyr::mutate("TaxonomyID" = as.character(!!dplyr::sym("TaxonomyID"))) |>
    dplyr::distinct(!!dplyr::sym("TaxonomyID"), .keep_all = TRUE) |>
    dplyr::left_join(all_silva_headers, by = c("TaxonomyID")) |>
    dplyr::relocate("genus", "species", .after = "family") |>
    dplyr::relocate("read_count", "Proportion", "readsEM", "EMProportion", .after = "species") |>
    dplyr::select(-"Genome")
  return(tax_table_pre)
}

#' Adds in taxa if input used NCBI database
#'
#' Returns MetaScope Table with NCBI taxa in separate columns
#'
#' @param metascope_id_in MetaScope ID file with NCBI taxa qnames
#' @param BPPARAM An optional BiocParallelParam instance determining the
#'   parallel back-end to be used during evaluation.
#' @inheritParams metascope_blast
#' @keywords internal
#' @return data.frame or tibble of taxonomy information

add_in_taxa_ncbi <- function(metascope_id_in, accession, BPPARAM) {
  taxon_ranks <- c("superkingdom", "kingdom", "phylum", "class",
                   "order", "family", "genus", "species")
  all_ncbi <- plyr::llply(metascope_id_in$TaxonomyID, .fun = class_taxon,
                          accession = accession, .progress = "text")
  # fix unknowns
  unk_tab <- data.frame(name = "unknown", rank = taxon_ranks, id = 0)
  all_ncbi <- lapply(all_ncbi, function(x) if (all(is.na(x))) unk_tab else x)
  # Create table
  taxonomy_table <- BiocParallel::bplapply(all_ncbi, mk_table,
                                           taxon_ranks = taxon_ranks,
                                           BPPARAM = BPPARAM) |>
    dplyr::bind_rows() |> as.data.frame() |>
    magrittr::set_colnames(taxon_ranks) |>
    dplyr::mutate("TaxonomyID" = metascope_id_in$TaxonomyID) |>
    dplyr::relocate("TaxonomyID") |>
    dplyr::distinct(!!dplyr::sym("TaxonomyID"), .keep_all = TRUE) |>
    dplyr::left_join(metascope_id_in, by = c("TaxonomyID")) |>
    dplyr::relocate("genus", "species", .after = "family") |>
    dplyr::relocate("read_count", "Proportion", "readsEM", "EMProportion", .after = "species") |>
    dplyr::select(-"Genome")
  return(taxonomy_table)
}

#' Gets sequences from bam file
#'
#' Returns fasta sequences from a bam file with a given taxonomy ID
#'
#' @param id Taxonomy ID of genome to get sequences from
#' @param bam_file A sorted bam file and index file, loaded with
#'   Rsamtools::bamFile
#' @param n Number of sequences to retrieve
#' @param bam_seqs A list of the sequence IDs from the bam file
#'
#' @return Biostrings format sequences

get_seqs <- function(id, bam_file, n = 10, bam_seqs) {
  rlang::is_installed("GenomicRanges")
  rlang::is_installed("IRanges")
  allGenomes <- stringr::str_detect(id, bam_seqs)
  # Get sequence info (Genome Name) from bam file
  seq_info_df <- as.data.frame(Rsamtools::seqinfo(bam_file)) |>
    tibble::rownames_to_column("seqnames")
  # Sample one of the Genomes that match
  Genome <- sample(seq_info_df$seqnames[allGenomes], 1)
  # Scan Bam file for all sequences that match genome
  param <- Rsamtools::ScanBamParam(what = c("rname", "seq"),
                                   which = GenomicRanges::GRanges(
                                     seq_info_df$seqnames[allGenomes],
                                     IRanges::IRanges(1, 1e+07)))
  allseqs <- Rsamtools::scanBam(bam_file, param = param)[[1]]
  n <- min(n, length(allseqs$seq))
  seqs <- sample(allseqs$seq, n)
  return(seqs)
}


#' Gets multiple sequences from different accessions in a bam file
#'
#' Returns fasta sequences from a bam file with given taxonomy IDs
#'
#' @param ids_n List of vectors with Taxonomy IDs and the number of sequences to get from each
#' @param bam_file A sorted bam file and index file, loaded with
#'   Rsamtools::bamFile
#' @param seq_info_df Dataframe of sequence information from metascope_blast()
#' @param metascope_id_tax Data.frame of taxonomy information
#' @param sorted_bam_file Filepath to sorted bam file
#'
#' @return Biostrings format sequences
#' @keywords internal
get_multi_seqs <- function(ids_n, bam_file, seq_info_df,
                           metascope_id_tax,
                           sorted_bam_file) {
  ids <- ids_n[1]
  requireNamespace("GenomicRanges")
  requireNamespace("IRanges")
  allGenomes <- seq_info_df |>
    dplyr::filter(!!dplyr::sym("seqnames") %in% ids) |>
    dplyr::pull("original_seqids")
  # Define BAM file with smaller yield size
  YS <- 1000
  bf <- Rsamtools::BamFile(sorted_bam_file, yieldSize = YS)
  open(bf)
  param <- Rsamtools::ScanBamParam(what = c("rname", "seq"),
                                   which = GenomicRanges::GRanges(
                                     allGenomes,
                                     IRanges::IRanges(1, 1e+07)))
  max_n <- Rsamtools::countBam(bf, param = param)$records
  close(bf)
  n_seqs <- min(as.integer(ids_n[2]), sum(max_n))

  YIELD <- function(bf) {
    value <- Rsamtools::scanBam(bf, param = param)
    flattened_value <- list()
    flattened_value$rname <- lapply(value, function(x) x$rname)
    flattened_value$seq <- lapply(value, function(x) x$seq |> as.character())
    return(flattened_value)
  }
  MAP <- function(value) {
    if(length(value$rname) == 0) return(dplyr::tibble(NULL))
    y <- dplyr::tibble("UID" = value$rname |> unlist(),
                       "Sequence" = value$seq |> unlist())
    return(y)
  }
  REDUCE <- function(x, y) {
    all_join <- dplyr::bind_rows(x, y)
    # Print all current reads to file
    return(all_join)
  }

  reduceByYield_multiseqs <- function(X, YIELD, MAP, REDUCE, n_end) {
    if (!Rsamtools::isOpen(X)) {
      Rsamtools::open.BamFile(X)
      on.exit(Rsamtools::close.BamFile(X))
    }
    result <- dplyr::tibble(NULL)

    while(nrow(result) < n_end) {
      data <- YIELD(X)
      result <- REDUCE(result, MAP(data))
      message(result)
    }
    return(result)
  }

  seqs <- reduceByYield_multiseqs(bf, YIELD, MAP, REDUCE, n_end = n_seqs)
  #seqs <- seqs[sample(nrow(seqs), n_seqs), ]

  final_seqs <- list(rname = seqs$UID, seq = Biostrings::DNAStringSet(seqs$Sequence))
  return(final_seqs)
}

#' Converts NCBI taxonomy ID to scientific name
#'
#' @param taxids List of NCBI taxids to convert to scientific name
#' @importFrom rlang .data
#' @inheritParams metascope_blast
#' @return Returns a dataframe of blast results for a metascope result
#'

taxid_to_name <- function(taxids, accession_path) {
  taxids <- stringr::str_replace(taxids, ";(.*)$", "") |> as.integer()
  out <- taxonomizr::getTaxonomy(taxids, accession_path)
  out_df <- as.data.frame(out) |> tibble::rownames_to_column("staxid") |>
    dplyr::select("staxid", "genus", "species") |>
    dplyr::mutate("species" = stringr::str_replace(!!dplyr::sym("species"), paste0(!!dplyr::sym("genus"), " "), "")) |>
    dplyr::mutate("staxid" = gsub("\\.", "", !!dplyr::sym("staxid"))) |>
    dplyr::mutate("staxid" = gsub("X", "", !!dplyr::sym("staxid"))) |>
    dplyr::mutate("staxid" = as.integer(!!dplyr::sym("staxid")))
  return(out_df)
}

#' Checks if blastn is installed
#'
#' @title Check if blastn exists on the system
#' @description This is an internal function that is not meant to be used
#'   outside of the package. It checks whether blastn exists on the system.
#' @return Returns TRUE if blastn exists on the system, else FALSE.
#'

check_blastn_exists <- function() {
  if (file.exists(Sys.which("blastn"))) return(TRUE)
  return(FALSE)
}

# BLASTn sequences


blastn_seqs <- function(db_path, fasta_path, res_path, hit_list, num_threads) {
  if (!check_blastn_exists()) {
    stop("BLAST executable not found. Please install before running.")
  }
  sys::exec_wait(
    "blastn", c("-db", db_path, "-query", fasta_path, "-out", res_path,
                "-outfmt", paste("10 qseqid sseqid pident length mismatch gapopen",
                                 "qstart qend sstart send evalue bitscore staxid"),
                "-max_target_seqs", hit_list, "-num_threads", num_threads,
                "-task", "megablast"),
    std_out = FALSE, 
    std_err = FALSE)
}



#' blastn_single_result
#'
#' @param results_table A dataframe of the Metascope results
#' @param bam_file A sorted bam file and index file, loaded with
#'   Rsamtools::bamFile
#' @param which_result Index in results_table for which result to Blast search
#' @param num_reads Number of reads to blast per result
#' @param hit_list Number of how many blast results to fetch per read
#' @param num_threads Number of threads if multithreading
#' @param db_path Blast database path
#' @param bam_seqs A list of the sequence IDs from the bam file
#' @param out_path Path to output results.
#' @param fasta_dir Path to where fasta files are stored.
#' @inheritParams metascope_blast
#'
#' @return Returns a dataframe of blast results for a metascope result
#'

blastn_single_result <- function(results_table, bam_file, which_result,
                                 num_reads = 100, hit_list = 10,
                                 num_threads, db_path, quiet,
                                 accession_path, bam_seqs, out_path,
                                 sample_name, fasta_dir = NULL) {
  res <- tryCatch({ #If any errors, should just skip the organism
    genome_name <- results_table$Genome[which_result]
    if (!quiet) message("Current id: ", genome_name)
    tax_id <- results_table$TaxonomyID[which_result]
    if (!quiet) message("Current ti: ", tax_id)
    # Generate sequences to blast
    if (!is.null(fasta_dir)) {
      fasta_path <- list.files(path = fasta_dir, full.names = TRUE)[which_result]
    } else {
      fasta_path <- get_seqs(id = tax_id, bam_file = bam_file, n = num_reads,
                             bam_seqs = bam_seqs)
    }

    res_path <- file.path(out_path, paste0(sprintf("%05d", which_result), "_",
                                          sample_name, ".csv"))
    blastn_seqs(db_path, fasta_path, res_path = res_path, hit_list, num_threads)
    blast_res <- utils::read.csv(res_path, header = FALSE) |>
      magrittr::set_colnames(c("qseqid", "sseqid", "pident", "length", "mismatch",
                               "gapopen","qstart", "qend", "sstart", "send",
                               "evalue", "bitscore", "staxid")) |>
      dplyr::mutate("MetaScope_Taxid" = tax_id,
                    "MetaScope_Genome" = as.character(genome_name))
      taxize_genome_df <- taxid_to_name(unique(blast_res$staxid),
                                        accession_path = accession_path)
    blast_res <- dplyr::left_join(blast_res, taxize_genome_df, by = "staxid")
    blast_res
  },
  error = function(e) {
    message(conditionMessage(e), "\n")
    this_format <- paste("qseqid sseqid pident length mismatch gapopen",
                         "qstart qend sstart send evalue bitscore staxid")
    all_colnames <- stringr::str_split(this_format, " ")[[1]]
    blast_res <- matrix(NA, ncol = length(all_colnames),
                        dimnames = list(c(), all_colnames)) |> as.data.frame() |>
      dplyr::mutate("MetaScope_Taxid" = results_table[which_result, 1],
                    "MetaScope_Genome" =
                      as.character(results_table[which_result, 2]),
                    "name" = NA)
    blast_res
  }
  )
  return(res)
}

#' Reformat BLASTn results
#'
#' @param results_table data.frame containing the MetaScope results.
#' @param bam_file \code{Rsamtools::bamFile} instance for the given sample.
#' @param num_results Integer; maximum number of Metascope results to BLAST.
#'   Default is 10.
#' @param num_reads_per_result Integer; number of reads to BLAST per result.
#'   Default is 100.
#' @param hit_list Integer; how many BLAST results to fetch for each read.
#'   Default is 10.
#' @param num_threads Integer; how many threads to use if multithreading.
#'   Default is 1.
#' @param db_path Character string; filepath for the location of the
#'   pre-installed BLAST database.
#' @param out_path Character string; Output directory to save CSV output files,
#'   including base name of files. For example, given a sample "X78256",
#'   filepath would be \code{file.path(directory_here, "X78256")} with extension
#'   omitted.
#' @param fasta_dir Character string; Directory where fastas from
#'   \code{metascope_id} are stored.
#' @param BPPARAM An optional BiocParallelParam instance determining the
#'   parallel back-end to be used during evaluation.
#' @inheritParams metascope_blast
#'
#' @return Creates and exports num_results number of csv files with blast
#'   results from local blast

blastn_results <- function(results_table, bam_file, num_results = 10,
                           num_reads_per_result = 100, hit_list = 10,
                           num_threads = 1, db_path, out_path, db = NULL,
                           sample_name = NULL, quiet = quiet,
                           accession_path, fasta_dir = NULL,
                           BPPARAM) {
  # Grab all identifiers
  if (is.null(fasta_dir)) {
    seq_info_df <- as.data.frame(Rsamtools::seqinfo(bam_file)) |>
      tibble::rownames_to_column("seqnames")
    bam_seqs <- find_accessions(seq_info_df$seqnames,
                                quiet = quiet) |>
      plyr::aaply(1, function(x) x[1])
  } else {
    bam_seqs <- NULL
  }
  # Grab results
  num_results2 <- min(num_results, nrow(results_table))
  run_res <- function(i) {
    df <- blastn_single_result(results_table, bam_file, which_result = i,
                               num_reads = num_reads_per_result,
                               hit_list = hit_list, num_threads = num_threads,
                               db_path = db_path, quiet = quiet, bam_seqs = bam_seqs,
                               out_path = out_path, sample_name = sample_name,
                               fasta_dir = fasta_dir, accession_path = accession_path)
    utils::write.csv(df, file.path(out_path,
                                   paste0(sprintf("%05d", i), "_", sample_name, ".csv")),
                     row.names = FALSE)
  }
  BiocParallel::bplapply(seq_len(num_results2), run_res, BPPARAM = BPPARAM)
}

#' Calculates result metrics from a blast results table
#'
#' This function calculates the best hit (genome with most blast read hits),
#' uniqueness score (total number of genomes hit), species percentage hit
#' (percentage of reads where MetaScope species also matched the blast hit
#' species), genus percentage hit (percentage of reads where blast genus matched
#' MetaScope aligned genus) and species contaminant score (percentage of reads
#' that blasted to other species genomes) and genus contaminant score
#' (percentage of reads that blasted to other genus genomes)
#'
#' @param blast_results_table_path path for blast results csv file
#' @inheritParams metascope_blast
#'
#' @return a vector with best_hit, uniqueness_score, species_percentage_hit
#'   genus_percentage_hit, species_contaminant_score, and
#'   genus_contaminant_score

blast_result_metrics <- function(blast_results_table_path, accession_path, db = NULL){
  tryCatch({
    # Load in blast results table
    blast_results_table <- utils::read.csv(blast_results_table_path, header = TRUE)

    # Clean species results
    blast_results_table <- blast_results_table |>
      dplyr::filter(!grepl("sp.", !!dplyr::sym("species"), fixed = TRUE)) |>
      dplyr::filter(!grepl("uncultured", !!dplyr::sym("species"), fixed = TRUE)) |>
      dplyr::filter(!is.na(!!dplyr::sym("genus")))

    # Remove any empty tables
    if (nrow(blast_results_table) < 2) {
      error_vector <- c(rep(0, 5), rep(NA, 3)) |>
        magrittr::set_names(c("uniqueness_score", "species_percentage_hit", "genus_percentage_hit",
                            "species_contaminant_score", "genus_contaminant_score",
                            "best_hit_genus", "best_hit_species", "best_hit_strain"))
    return(error_vector)
    }
    # Clean species results
    blast_results_table <- blast_results_table |>
      dplyr::filter(!grepl("sp.", !!dplyr::sym("species"), fixed = TRUE)) |>
      dplyr::filter(!grepl("uncultured", !!dplyr::sym("species"), fixed = TRUE)) |>
      dplyr::filter(!is.na(!!dplyr::sym("genus")))
    # Adding MetaScope Species and Genus columns
    if (db == "silva") {
      # Cleaning up silva genome names
      #silva_genome <- gsub(";([a-z0-9])", " \\1", blast_results_table$MetaScope_Genome[1])
      silva_genome <- gsub(" uncultured", ";uncultured", blast_results_table$MetaScope_Genome[1])
      # Creating silva genus and species names
      silva_genus <- magrittr::extract(strsplit(silva_genome, split = "_"), 1)[[1]]
      silva_species <- magrittr::extract(strsplit(silva_genome, split = "_"), 2)[[1]]
      blast_results_table_2 <- blast_results_table |>
        dplyr::mutate("MetaScope_genus" = rep(silva_genus, nrow(blast_results_table)),
                      "MetaScope_species" = rep(silva_species, nrow(blast_results_table))) |>
        dplyr::rename("query_genus" = "genus",
                      "query_species" = "species") |>
        # Remove rows with NA
        tidyr::drop_na("query_genus", "query_species") |>
        # Getting best hit per read
        dplyr::group_by(!!dplyr::sym("qseqid")) |>
        dplyr::slice_min(!!dplyr::sym("evalue"), with_ties = TRUE) |>
        # Removing duplicate query num and query species
        dplyr::distinct(!!dplyr::sym("qseqid"), !!dplyr::sym("query_species"), .keep_all = TRUE)

    } else {
      meta_tax <- taxid_to_name(unique(blast_results_table$MetaScope_Taxid),
                                accession_path = accession_path) |>
        dplyr::select(-"staxid")
      blast_results_table_2 <- blast_results_table |>
        dplyr::mutate("MetaScope_genus" = meta_tax$genus[1],
                      "MetaScope_species" = meta_tax$species[1]) |>
        dplyr::rename("query_genus" = "genus",
                      "query_species" = "species") |>
        # Remove rows with NA
        tidyr::drop_na() |>
        # Getting best hit per read
        dplyr::group_by(!!dplyr::sym("qseqid")) |>
        dplyr::slice_min(!!dplyr::sym("evalue"), with_ties = TRUE) |>
        # Removing duplicate query num and query species
        dplyr::distinct(!!dplyr::sym("qseqid"), !!dplyr::sym("query_species"), .keep_all = TRUE)
    }
    # Calculating Metrics
    best_hit_genus <- blast_results_table_2 |>
      dplyr::group_by(!!dplyr::sym("query_genus")) |>
      dplyr::summarise("num_reads" = dplyr::n()) |>
      dplyr::slice_max(!!dplyr::sym("num_reads"), with_ties = FALSE)

    best_hit_species <- blast_results_table_2 |>
      dplyr::group_by(!!dplyr::sym("query_species")) |>
      dplyr::summarise("num_reads" = dplyr::n()) |>
      dplyr::slice_max(!!dplyr::sym("num_reads"), with_ties = FALSE)

    #best_hit_strain <- blast_results_table_2 |>
    #  tidyr::separate("sseqid", c(NA, "gi", NA, NA, NA), sep = "\\|")  |>
    #  dplyr::group_by(!!dplyr::sym("gi")) |>
    #  dplyr::summarise("num_reads" = dplyr::n()) |>
    #  dplyr::slice_max(!!dplyr::sym("num_reads"), with_ties = TRUE)
    #if (nrow(best_hit_strain) > 1 | db == "silva") {
    #  best_strain <- NA
    #} else if (nrow(best_hit_strain) == 1) {
    #  res <- taxonomizr::getTaxonomy(best_hit_strain$gi, sqlFile = accession_path)
    #  best_strain <- attr(res[[1]], "name")
    #}

    uniqueness_score <- blast_results_table_2 |>
      dplyr::group_by(!!dplyr::sym("query_species")) |>
      dplyr::summarise("num_reads" = dplyr::n()) |>
      nrow()

    species_percentage_hit <- blast_results_table_2 |>
      dplyr::filter(!!dplyr::sym("MetaScope_species") == !!dplyr::sym("query_species")) |>
      nrow() / length(unique(blast_results_table_2$qseqid))

    genus_percentage_hit <- blast_results_table_2 |>
      dplyr::filter(!!dplyr::sym("MetaScope_genus") == !!dplyr::sym("query_genus")) |>
      dplyr::distinct(!!dplyr::sym("qseqid"), .keep_all = FALSE) |>
      nrow() / length(unique(blast_results_table_2$qseqid))

    species_contaminant_score <- blast_results_table_2 |>
      dplyr::filter(!!dplyr::sym("MetaScope_species") != !!dplyr::sym("query_species")) |>
      dplyr::distinct(!!dplyr::sym("qseqid"), .keep_all = TRUE) |>
      nrow() / length(unique(blast_results_table_2$qseqid))

    genus_contaminant_score <- blast_results_table_2 |>
      dplyr::filter(!!dplyr::sym("MetaScope_genus") != !!dplyr::sym("query_genus")) |>
      dplyr::distinct(!!dplyr::sym("qseqid"), .keep_all = TRUE) |>
      nrow() / length(unique(blast_results_table_2$qseqid))


    all_results <- c(uniqueness_score, species_percentage_hit, genus_percentage_hit,
                     species_contaminant_score, genus_contaminant_score,
                     best_hit_genus$query_genus, best_hit_species$query_species)
    names(all_results) <- c("uniqueness_score", "species_percentage_hit", "genus_percentage_hit",
                            "species_contaminant_score", "genus_contaminant_score",
                            "best_hit_genus", "best_hit_species")
    return(all_results)
  },
  error = function(e)
  {
    message("Error", conditionMessage(e), "/n")
    error_vector <- c(0,0,0,0,0, NA, NA)
    names(error_vector) <- c("uniqueness_score", "species_percentage_hit", "genus_percentage_hit",
                             "species_contaminant_score", "genus_contaminant_score",
                             "best_hit_genus", "best_hit_species")

    return(error_vector)
  }
  )
}


#' Blast reads from MetaScope aligned files
#'
#' This function allows the user to check a subset of identified reads against
#' NCBI BLAST and the nucleotide database to confirm or contradict results
#' provided by MetaScope. It aligns the top `metascope_id()` results to NCBI
#' BLAST database. It REQUIRES that command-line BLAST and a separate nucleotide
#' database have already been installed on the host machine. It returns a csv
#' file updated with BLAST result metrics.
#'
#' This function assumes that you used the NCBI nucleotide database to process
#' samples, with a download date of 2021 or later. This is necessary for
#' compatibility with the bam file headers.
#'
#' This is highly computationally intensive and should be ran with multiple
#' cores, submitted as a multi-threaded computing job if possible.
#'
#' Note, if best_hit_strain is FALSE, then no strain was observed more often
#' among the BLAST results.
#'
#' @param metascope_id_path Character string; path to a csv file output by
#'   `metascope_id()`.
#' @param bam_file_path Character string; full path to bam file for the same
#'   sample processed by `metascope_id`. Note that the `filter_bam` function
#'   must have output a bam file, and not a .csv.gz file. See
#'   `?filter_bam_bowtie` for more details. Defaults to
#'   \code{list.files(file_temp, ".updated.bam$")[1]}.
#' @param tmp_dir Character string, a temporary directory in which to host
#'   files.
#' @param out_dir Character string, path to output directory.
#' @param sample_name Character string, sample name for output files.
#' @param num_results Integer, the maximum number of taxa from the metascope_id
#'   output to check reads. Takes the top n taxa, i.e. those with largest
#'   abundance. Default is 10.
#' @param num_reads Integer, the maximum number of reads to blast per microbe.
#'   If the true number of reads assigned to a given taxon is smaller, then the
#'   smaller number will be chosen. Default is 100. Too many reads will involve
#'   more processing time.
#' @param hit_list Integer, number of blast hit results to keep. Default is 10.
#' @param num_threads Integer, number of threads if running in parallel
#'   (recommended). Default is 1.
#' @param db_path Character string. The database file to be searched (including
#'   basename, but without file extension). For example, if the nt database is
#'   in the nt folder, this would be /filepath/nt/nt assuming that the database
#'   files have the nt basename. Check this path if you get an error message
#'   stating "No alias or index file found".
#' @param quiet Logical, whether to print out more informative messages. Default
#'   is FALSE.
#' @param db Currently accepts one of \code{c("ncbi", "silva", "other")} Default
#'   is \code{"ncbi"}, appropriate for samples aligned against indices compiled
#'   from NCBI whole genome databases. Alternatively, usage of an alternate
#'   database (like Greengenes2) should be specified with \code{"other"}.
#' @param fasta_dir Directory where fasta files for blast will be stored.
#' @param accession_path (character) Filepath to NCBI accessions SQL
#'   database. See \code{taxonomzr::prepareDatabase()}.
#'
#' @returns This function writes an updated csv file with metrics.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' ### Create temporary directory
#' file_temp <- tempfile()
#' dir.create(file_temp)
#'
#' bamPath <- system.file("extdata", "bowtie_target.bam",
#'                        package = "MetaScope")
#' file.copy(bamPath, file_temp)
#'
#' metascope_id(file.path(file_temp, "bowtie_target.bam"), aligner = "bowtie2",
#'              input_type = "bam", out_dir = file_temp, num_species_plot = 0,
#'              update_bam = TRUE)
#'
#' ## Run metascope blast
#' ### Get export name and metascope id results
#' out_base <- bamPath |> basename() |> tools::file_path_sans_ext() |>
#'   tools::file_path_sans_ext()
#' metascope_id_path <- file.path(file_temp, paste0(out_base, ".metascope_id.csv"))
#'
#' # NOTE: change db_path to the location where your BLAST database is stored!
#' db <- "/restricted/projectnb/pathoscope/data/blastdb/nt/nt"
#'
#' tmp_accession <- system.file("extdata", "example_accessions.sql", package = "MetaScope"
#'
#' metascope_blast(metascope_id_path,
#'                 bam_file_path = file.path(file_temp, "bowtie_target.bam"),
#'                 tmp_dir = file_temp,
#'                 out_dir = file_temp,
#'                 sample_name = out_base,
#'                 db_path = db,
#'                 num_results = 10,
#'                 num_reads = 5,
#'                 hit_list = 10,
#'                 num_threads = 3,
#'                 db = "ncbi",
#'                 quiet = FALSE,
#'                 fasta_dir = NULL,
#'                 accession_path = tmp_accession)
#'
#' ## Remove temporary directory
#' unlink(file_temp, recursive = TRUE)
#'}
#'

metascope_blast <- function(metascope_id_path,
                            bam_file_path = list.files(tmp_dir, ".updated.bam$",
                                                       full.names = TRUE)[1],
                            tmp_dir, out_dir, sample_name, fasta_dir = NULL,
                            num_results = 10, num_reads = 100, hit_list = 10,
                            num_threads = 1, db_path, quiet = FALSE,
                            db = NULL, accession_path = NULL) {
  # Parallelization settings
  if (!is.numeric(num_threads)) num_threads <- 1
  BPPARAM <- BiocParallel::MulticoreParam(workers = num_threads)

  # Sort and index bam file
  sorted_bam_file_path <- file.path(tmp_dir, paste0(sample_name, "_sorted"))
  message("Sorting bam file to ", sorted_bam_file_path, ".bam")
  Rsamtools::sortBam(bam_file_path, destination = sorted_bam_file_path,
                     overwrite = TRUE)
  sorted_bam_file <- paste0(sorted_bam_file_path, ".bam")
  Rsamtools::indexBam(sorted_bam_file)
  bam_file <- Rsamtools::BamFile(sorted_bam_file, index = sorted_bam_file)

  # Generate fasta file from sorted bam file
  command_string <- paste0("samtools view ", file.path(tmp_dir, sample_name),
                           "_sorted.bam | awk \'{print \">\"$3 \"\\n\" $10}\' > ",
                           file.path(tmp_dir, sample_name), "_aligned.fasta")
  system(command_string)
  all_fastas <- Biostrings::readDNAStringSet(paste0(file.path(tmp_dir, sample_name), "_aligned.fasta"))
  accessions_taxids <- find_taxids(all_fastas, accession_path)


  # Load in metascope id file and clean unknown genomes
  metascope_id_in <- utils::read.csv(metascope_id_path, header = TRUE)

  # Group metascope id by species and create metascope species id
  if (db == "silva") {
    metascope_id_species <- add_in_taxa(metascope_id_in, caching = FALSE,           # TODO Add initialization steps to install required databases
                                    path_to_write = tmp_dir)
  } else if (db == "ncbi") {
    tax_df <- taxonomizr::getTaxonomy(metascope_id_in$TaxonomyID, sqlFile = accession_path)
    rownames(tax_df) <- trimws(rownames(tax_df))
    tax_df <- as.data.frame(tax_df) |> 
      tibble::rownames_to_column("TaxonomyID") |>
      dplyr::mutate("TaxonomyID" = ifelse(grepl("NA", !!dplyr::sym("TaxonomyID"), fixed=TRUE), 
                                        NA, 
                                        gsub("^X[.]?", "", !!dplyr::sym("TaxonomyID")))) |>
      dplyr::filter(!is.na(!!dplyr::sym("TaxonomyID"))) |>
      dplyr::mutate("TaxonomyID" = as.integer(!!dplyr::sym("TaxonomyID")))
      
    metascope_id_species <- dplyr::left_join(metascope_id_in, tax_df,
                                             by = "TaxonomyID")
  }

  # Create fasta directory in tmp directory to save fasta sequences
  fastas_tmp_dir <- file.path(tmp_dir, "fastas")
  if(!dir.exists(fastas_tmp_dir)) dir.create(fastas_tmp_dir,
                                             recursive = TRUE)
  unlink(paste0(fastas_tmp_dir, "/*"), recursive = TRUE)
  message("Generating fasta sequences from bam file")
  # How many taxa
  num_taxa_loop <- min(nrow(metascope_id_species), num_results)

  # Sampling num_reads number of fastas per genome
  sample_fastas <- function(i) {
    current_species <- metascope_id_species$species[i]
    if (!is.na(current_species)) {
      current_accessions <- accessions_taxids |>
        dplyr::filter(!!dplyr::sym("species") == current_species) |>
        dplyr::pull("accessions")
    } else {
      accession_id <- gsub("unknown genome; accession ID is", "", 
                           metascope_id_species$Genome[i])
      current_accessions <- accessions_taxids$accessions[
        which(grepl(accession_id, accessions_taxids$accessions))[1]
      ]
    }
    seqs <- all_fastas[names(all_fastas) %in% current_accessions]
    num_seqs <- min(length(seqs), num_reads) 
    num_digits <- nchar(as.character(num_reads))
    seqs <- sample(seqs, size = num_seqs)
    if (length(seqs) > 0) {
      names(seqs) <- paste0(
        "ms_index", i, "|",
        names(seqs), 
        "|read_", 
        sprintf(paste0("%0", num_digits, "d"), seq_along(seqs)))
    }
    return(seqs)
  }

  # Parallelize sampling of fastas
  all_seqs <- BiocParallel::bplapply(seq_len(num_taxa_loop), sample_fastas, BPPARAM = BPPARAM)
  fasta_path <- file.path(fastas_tmp_dir, "blast_input_fastas.fa")
  
  Biostrings::writeXStringSet(
    do.call(c,all_seqs), # flattens list  
    filepath = fasta_path)
  
  # Create blast directory in tmp directory to save blast results in
  blast_tmp_dir <- file.path(tmp_dir, "blast")
  if(!dir.exists(blast_tmp_dir)) dir.create(blast_tmp_dir, recursive = TRUE)
  unlink(paste0(blast_tmp_dir, "/*"), recursive = TRUE)

  # Run rBlast on all metascope microbes
  if (!check_blastn_exists()) {
    stop("BLAST executable not found. Please install before running.")
  } else {
    message("BLASTN executable found on system.")
  }
  res_path <- file.path(blast_tmp_dir, "blast_output.csv")
  message("Running BLASTN on all sequences")
  sys::exec_wait(
    "blastn", c("-db", db_path, "-query", fasta_path, "-out", res_path,
                "-outfmt", paste("10 qseqid sseqid pident length mismatch gapopen",
                                 "qstart qend sstart send evalue bitscore staxid"),
                "-max_target_seqs", hit_list, "-max_hsps", 1, 
                "-num_threads", num_threads,
                "-task", "megablast"))

  # Calculate Blast metrics
  message("Running BLAST metrics on all blast results")
  blast_result_metrics_df <- utils::read.csv(res_path, header = FALSE)
  colnames(blast_result_metrics_df) <- c("qseqid", "sseqid", "pident", "length", "mismatch", "gapopen",
                                         "qstart", "qend", "sstart", "send", "evalue", "bitscore", "staxid")
  ## Extract MetaScope Taxid
  blast_result_metrics_df <- blast_result_metrics_df |>
    dplyr::mutate(
      ms_index = stringr::str_extract(!!dplyr::sym("qseqid"), "(?<=ms_index)\\d+"), 
      metascope_taxid = purrr::map_chr(stringr::str_split(!!dplyr::sym("qseqid"), pattern = "\\|"), ~ .x[3]),
      read_id = purrr::map_chr(stringr::str_split(!!dplyr::sym("qseqid"), pattern = "\\|"), ~ .x[8]))
  ## Calculate MetaScope species and Blast Species
  ms_tax <- taxonomizr::getTaxonomy(unique(blast_result_metrics_df$metascope_taxid), sqlFile = accession_path) |>
    as.data.frame() |>
    tibble::rownames_to_column("metascope_taxid") |>
    dplyr::mutate(metascope_taxid = stringr::str_trim(!!dplyr::sym("metascope_taxid")))
  colnames(ms_tax) <- c("metascope_taxid", "ms_domain", "ms_phylum", "ms_class", "ms_order", "ms_family", "ms_genus", "ms_species")
  b_tax <- taxonomizr::getTaxonomy(unique(blast_result_metrics_df$staxid), sqlFile = accession_path) |>
    as.data.frame() |>
    tibble::rownames_to_column("staxid") |>
    dplyr::mutate(staxid = as.integer(stringr::str_trim(!!dplyr::sym("staxid"))))
  colnames(b_tax) <- c("staxid", "b_domain", "b_phylum", "b_class", "b_order", "b_family", "b_genus", "b_species")
  blast_result_metrics_df <- dplyr::left_join(blast_result_metrics_df, ms_tax, by = "metascope_taxid")
  blast_result_metrics_df <- dplyr::left_join(blast_result_metrics_df, b_tax, by = "staxid")
  # Calculate Blast Metrics grouped by qseqid
  ms_blast_results <- data.frame(
    ms_index = as.character(seq(1, num_results))
  )
  ## Best Hit Genus
  best_hit_genus <- blast_result_metrics_df |>
    dplyr::filter(!is.na(!!dplyr::sym("b_genus"))) |> 
    dplyr::group_by(!!dplyr::sym("qseqid"), 
                    !!dplyr::sym("ms_index"),
                    !!dplyr::sym("read_id"),
                    !!dplyr::sym("b_genus")) |>
    dplyr::summarize(genus_counts = dplyr::n()) |>
    dplyr::ungroup() |>
    dplyr::group_by(!!dplyr::sym("ms_index"),
                    !!dplyr::sym("read_id")) |>
    dplyr::slice_max(!!dplyr::sym("genus_counts")) |>
    dplyr::ungroup() |>
    dplyr::group_by(!!dplyr::sym("ms_index")) |>
    dplyr::count(!!dplyr::sym("ms_index"), !!dplyr::sym("b_genus"), name = "genus_counts") |> 
    dplyr::slice_max(order_by = !!dplyr::sym("genus_counts"), n = 1, with_ties = FALSE) |>
    dplyr::rename("best_hit_genus" = "b_genus")
  ms_blast_results <- dplyr::left_join(ms_blast_results, best_hit_genus, by = "ms_index")
  ## Best Hit Species 
  best_hit_species <- blast_result_metrics_df |>
    dplyr::filter(!is.na(!!dplyr::sym("b_species"))) |> 
    dplyr::group_by(!!dplyr::sym("qseqid"), 
                    !!dplyr::sym("ms_index"),
                    !!dplyr::sym("read_id"),
                    !!dplyr::sym("b_species")) |>
    dplyr::summarize(species_counts = dplyr::n()) |>
    dplyr::ungroup() |>
    dplyr::group_by(!!dplyr::sym("ms_index"),
                    !!dplyr::sym("read_id")) |>
    dplyr::slice_max(!!dplyr::sym("species_counts")) |>
    dplyr::ungroup() |>
    dplyr::group_by(!!dplyr::sym("ms_index")) |>
    dplyr::count(!!dplyr::sym("ms_index"), !!dplyr::sym("b_species"), name = "species_counts") |> 
    dplyr::slice_max(order_by = !!dplyr::sym("species_counts"), n = 1, with_ties = FALSE) |>
    dplyr::rename("best_hit_species" = "b_species")
  ms_blast_results <- dplyr::left_join(ms_blast_results, best_hit_species, by = "ms_index")
  ## Uniqueness Score
  uniqueness_score <- blast_result_metrics_df |>
    dplyr::filter(!is.na(!!dplyr::sym("b_species"))) |> 
    dplyr::group_by(!!dplyr::sym("qseqid"), 
                    !!dplyr::sym("ms_index"),
                    !!dplyr::sym("read_id"),
                    !!dplyr::sym("b_species")) |>
    dplyr::summarize(species_counts = dplyr::n()) |>
    dplyr::ungroup() |>
    dplyr::group_by(!!dplyr::sym("ms_index"),
                    !!dplyr::sym("read_id")) |>
    dplyr::slice_max(!!dplyr::sym("species_counts")) |> # Up to here is best hit species
    dplyr::ungroup() |>
    dplyr::group_by(!!dplyr::sym("ms_index")) |>
    dplyr::summarise(uniqueness_score = dplyr::n_distinct(!!dplyr::sym("b_species")))
  ms_blast_results <- dplyr::left_join(ms_blast_results, uniqueness_score, by = "ms_index")
  ## Species Percentage Hit
  species_percentage_hit <- blast_result_metrics_df |>
    dplyr::filter(!is.na(!!dplyr::sym("b_species"))) |> 
    dplyr::mutate(species_match = !!dplyr::sym("b_species") == !!dplyr::sym("ms_species")) |> 
    dplyr::group_by(!!dplyr::sym("ms_index"), !!dplyr::sym("read_id")) |> 
    dplyr::summarize(has_match = any(!!dplyr::sym("species_match")), .groups = "drop") |> # Check if read has at least one match
    dplyr::group_by(!!dplyr::sym("ms_index")) |>
    dplyr::summarize(species_percentage_hit = mean(!!dplyr::sym("has_match")), .groups = "drop")
  ms_blast_results <- dplyr::left_join(ms_blast_results, species_percentage_hit, by = "ms_index")
  ## Genus Percentage Hit
  genus_percentage_hit <- blast_result_metrics_df |>
    dplyr::filter(!is.na(!!dplyr::sym("b_genus"))) |> 
    dplyr::mutate(genus_match = !!dplyr::sym("b_genus") == !!dplyr::sym("ms_genus")) |> 
    dplyr::group_by(!!dplyr::sym("ms_index"), !!dplyr::sym("read_id")) |> 
    dplyr::summarize(has_match = any(!!dplyr::sym("genus_match")), .groups = "drop") |> # Check if read has at least one match
    dplyr::group_by(!!dplyr::sym("ms_index")) |>
    dplyr::summarize(genus_percentage_hit = mean(!!dplyr::sym("has_match")), .groups = "drop")
  ms_blast_results <- dplyr::left_join(ms_blast_results, genus_percentage_hit, by = "ms_index")
  ## Species Contaminant Score
  species_contaminant_score <- blast_result_metrics_df |>
    dplyr::filter(!is.na(!!dplyr::sym("b_species"))) |>
    dplyr::mutate(species_match = !!dplyr::sym("b_species") == !!dplyr::sym("ms_species")) |>
    dplyr::group_by(!!dplyr::sym("ms_index"), !!dplyr::sym("read_id")) |>
    dplyr::summarize(has_mismatch = any(!(!!dplyr::sym("species_match"))), .groups = "drop") |>
    dplyr::group_by(!!dplyr::sym("ms_index")) |>
    dplyr::summarize(species_contaminant_score = mean(!!dplyr::sym("has_mismatch")), .groups = "drop")
  ms_blast_results <- dplyr::left_join(ms_blast_results, species_contaminant_score, by = "ms_index")
  ## Genus Contaminant Score
  genus_contaminant_score <- blast_result_metrics_df |>
    dplyr::filter(!is.na(!!dplyr::sym("b_genus"))) |>
    dplyr::mutate(genus_match = !!dplyr::sym("b_genus") == !!dplyr::sym("ms_genus")) |>
    dplyr::group_by(!!dplyr::sym("ms_index"), !!dplyr::sym("read_id")) |>
    dplyr::summarize(has_mismatch = any(!(!!dplyr::sym("genus_match"))), .groups = "drop") |>
    dplyr::group_by(!!dplyr::sym("ms_index")) |>
    dplyr::summarize(genus_contaminant_score = mean(!!dplyr::sym("has_mismatch")), .groups = "drop")
  ms_blast_results <- dplyr::left_join(ms_blast_results, genus_contaminant_score, by = "ms_index")
  
  # Append Blast Metrics to MetaScope results
  metascope_id_species <- metascope_id_species |>
    dplyr::mutate(ms_index = as.character(dplyr::row_number()))
  metascope_blast_df <- dplyr::left_join(metascope_id_species, ms_blast_results, by = "ms_index") |>
    dplyr::select(-!!dplyr::sym("ms_index"))
  metascope_blast_df[is.na(metascope_blast_df)] <- 0
  
  # Export File.
  print_file <- file.path(out_dir, paste0(sample_name, ".metascope_blast.csv"))
  utils::write.csv(metascope_blast_df, print_file, row.names = FALSE)
  message("Results written to ", print_file)
  
  utils::write.csv(blast_result_metrics_df, file.path(blast_tmp_dir, "blast_output_taxonomy.csv"), row.names = FALSE)
  return(metascope_blast_df)
}

#' Reassign reads from MetaScope BLASTn alignment
#'
#' Using the output from \code{metascope_blast()}, the
#' \code{blast_reassignment()} function takes the results and alters the
#' original \code{metascope_id()} output to reassign reads that were invalidated
#' by the BLAST findings. Currently, the implementation of this function only
#' reassigns reads to a taxon that was already found in the sample at a higher
#' abundance.
#'
#' @param metascope_blast_path Character string. The filepath to a
#'   \code{metascope_blast} CSV output file.
#' @param species_threshold Numeric. A number between 0 and 1 indicating
#'   the minimum proportion of reads needed for a taxon to be considered
#'   validated from the BLAST results. Default is 0.2, or 20\%.
#' @param num_hits Integer. The number of hits for which to assess validation.
#'   Default is 10, i.e., only the top 10 taxa will be assessed.
#' @param blast_tmp_dir Character string. Filepath of the directory where
#'   BLAST results were output from the \code{metascope_blast} function.
#'   Referencing the arguments from \code{metascope_blast}, this would be
#'   \code{file.path(tmp_dir, "blast")}
#' @param reassign_validated Logical. Should reads from validated accessions be
#'   reassigned to other validated accessions. Defaults to \code{FALSE}
#' @param reassign_to_validated Logical. Should reads only be re-assigned to
#'   validated accessions or to any accession with more reads than the current
#'   accession. Defaults to \code{TRUE}
#' @inheritParams metascope_blast
#'
#' @returns Returns a \code{data.frame} with the reassigned taxa and read counts.
#'
#' @export

blast_reassignment <- function(metascope_blast_path, 
                               species_threshold, num_hits,
                               blast_tmp_dir, out_dir, sample_name,
                               reassign_validated = FALSE,
                               reassign_to_validated = TRUE) {
  # Create Validated Column based on species thresholds
  metascope_blast_df <- data.table::fread(metascope_blast_path) |>
    tibble::as_tibble() |>
    dplyr::mutate(blast_validated = (!!dplyr::sym("species_percentage_hit") > species_threshold))
  # Create Index for later reassignment
  metascope_blast_df$ms_index <- rownames(metascope_blast_df) |> as.numeric()
  # Tidy columns read_counts to read_count
  if ("read_counts" %in% colnames(metascope_blast_df)) {
    ind <- colnames(metascope_blast_df) == "read_counts"
    colnames(metascope_blast_df)[ind] <- "read_count"
  }

  reassigned_metascope_blast <- metascope_blast_df

  blast_file <- list.files(blast_tmp_dir, pattern = "blast_output_taxonomy.csv", full.names = TRUE)
  blast_result_metrics_df <- data.table::fread(blast_file)
  # Create vector of indices that have been reassigned
  drop_indices <- c()

  # test all indices if reassign_validated, otherwise only test unvalidated indices
  if (reassign_validated) {
    test_indices <- c(2:min(num_hits, nrow(metascope_blast_df)))
  } else {
    test_indices <- which(!metascope_blast_df$blast_validated)
    test_indices <- test_indices[test_indices <= num_hits]
  }

  # define helper function get_blast_summary
  get_blast_summary <- function(i) {
    blast_summary <-
      tryCatch(
        {
          blast_summary <- data.table::fread(blast_file) |>
            tibble::as_tibble() |>
            dplyr::filter(!!dplyr::sym("ms_index") == i) |>
            dplyr::group_by(!!dplyr::sym("qseqid")) |>
            # Select lowest evalues
            dplyr::slice_min(!!dplyr::sym("evalue"), with_ties = TRUE) |>
            # Removing duplicate query num and query species
            dplyr::distinct(!!dplyr::sym("qseqid"), !!dplyr::sym("b_species"), .keep_all = TRUE) |>
            dplyr::ungroup() |>
            dplyr::group_by(!!dplyr::sym("b_species")) |>
            dplyr::summarise("num_reads" = dplyr::n(), .groups="keep") |>
            dplyr::slice_max(order_by = !!dplyr::sym("num_reads"), with_ties = TRUE) |>
            dplyr::left_join(metascope_blast_df[seq_len(num_hits), ],
                             by = dplyr::join_by("b_species" == "species")) |>
            # Do not reassign below metascope's current reassignment
            dplyr::filter(!is.na(!!dplyr::sym("ms_index"))) |>
            dplyr::filter(!!dplyr::sym("ms_index") < i)
          # If reassign_to_validated, then only reassign reads to blast_validated accesions
          if (reassign_to_validated) {
            blast_summary <- blast_summary |>
              dplyr::filter(!!dplyr::sym("blast_validated") == TRUE)
          }
          blast_summary <- blast_summary |>
            dplyr::ungroup() |>
            dplyr::mutate(reassignment_proportion = !!dplyr::sym("num_reads") / sum(!!dplyr::sym("num_reads")),
                          reassigned_read_count = metascope_blast_df$read_count[i] * !!dplyr::sym("reassignment_proportion"),
                          reassigned_Proportion = metascope_blast_df$Proportion[i] * !!dplyr::sym("reassignment_proportion"),
                          reassigned_readsEM = metascope_blast_df$readsEM[i] * !!dplyr::sym("reassignment_proportion"),
                          reassigned_EMProportion = metascope_blast_df$EMProportion[i] * !!dplyr::sym("reassignment_proportion"))
          blast_summary
        },
        error = function(cond) {
          message("Problem with blast summary for current file: ")
          message(blast_file[i])
          message(cond)
          # Choose a return value in case of error
          blast_summary <- dplyr::tibble(
            "genus" = numeric(),
            "species" = numeric(),
            "num_reads" = numeric()
          )
          blast_summary
        }
      )
    return(blast_summary)
  }
  for (i in test_indices){
    if (!metascope_blast_df$blast_validated[i]){
      blast_summary <- get_blast_summary(i)
      if (nrow(blast_summary) > 0) {
        # Updated reassigned_metascope_blast df for every row in blast_summary
        for (n in 1:nrow(blast_summary)) {
          metascope_index <- blast_summary$ms_index[n]
          #print(reassigned_metascope_blast$read_count[metascope_index])
          #print(blast_summary$reassigned_read_count[n])
          reassigned_metascope_blast$read_count[metascope_index] <-
            reassigned_metascope_blast$read_count[metascope_index] + blast_summary$reassigned_read_count[n]
          reassigned_metascope_blast$Proportion[metascope_index] <-
            reassigned_metascope_blast$Proportion[metascope_index] + blast_summary$reassigned_Proportion[n]
          reassigned_metascope_blast$readsEM[metascope_index] <-
            reassigned_metascope_blast$readsEM[metascope_index] + blast_summary$reassigned_readsEM[n]
          reassigned_metascope_blast$EMProportion[metascope_index] <-
            reassigned_metascope_blast$EMProportion[metascope_index] + blast_summary$reassigned_EMProportion[n]
        }
        drop_indices <- append(drop_indices, i)
      }
    }
  }
  reassigned_metascope_blast <- reassigned_metascope_blast[-drop_indices,] |>
    dplyr::select(-"ms_index")

  print_file <- file.path(out_dir, paste0(sample_name, ".metascope_blast_reassigned.csv"))
  data.table::fwrite(reassigned_metascope_blast, print_file, row.names = FALSE)
  message("Results written to ", print_file)
  return(reassigned_metascope_blast)
}

#' Create a multi-assay experiment from PathoScope 2.0 output for usage with
#' animalcules
#'
#' This function serves as a legacy integration method for usage with
#' PathoScope 2.0 outputs. Upon completion of the PathoScope 2.0 pipeline, users can analyze and visualize
#' abundances in their samples using the animalcules package. After running this function, the user should save the returned MAE
#' to an RDS file using a function like \code{saveRDS} to upload the output into
#' the \code{animalcules} package.
#'
#' @param patho_counts Character string, a directory filepath to the counts ID CSVs output by
#'   \code{metascope_id()}.
#' @param annot_path The filepath to the CSV annotation file for the samples.
#' @param end_string The end string used at the end of the metascope_id files.
#'   Default is "-sam-report.tsv".
#' @param which_annot_col The name of the column of the annotation file
#'   containing the sample IDs. These should be the same as the
#'   \code{meta_counts} root filenames.
#' @returns Returns a MultiAssay Experiment file of combined sample counts data.
#' The MultiAssay Experiment will have a counts assay ("MGX").
#' @export
#' @importFrom rlang .data
#' 
convert_animalcules_patho <- function(patho_counts, annot_path,
                                      which_annot_col,
                                      end_string = "-sam-report.tsv") {
  count_table <- animalcules::read_pathoscope_data(
    input_dir = patho_counts,
    pathoreport_file_suffix = end_string)$countdat
  # Choose only the samples in metadata that have counts data as well
  metadata_table <- readr::read_csv(annot_path, show_col_types = FALSE) |>
    as.data.frame()
  rownames(metadata_table) <- metadata_table[, which_annot_col]
  sample_overlap <- base::intersect(colnames(count_table), rownames(metadata_table))
  if (length(sample_overlap) < length(colnames(count_table))) {
    no_info <- paste(colnames(count_table)[which(!colnames(count_table) %in% sample_overlap)],
                     collapse = ",")
    message("The following samples don't have metadata info:", no_info)
    
    count_table <- count_table[,which(colnames(count_table) %in% sample_overlap)]
  }
  metadata_table <- metadata_table[match(colnames(count_table),
                                         rownames(metadata_table)), , drop = FALSE]
  # Test and fix the constant/zero row
  row.remove.index <- c()
  if (sum(base::rowSums(as.matrix(count_table)) == 0) > 0){
    row.remove.index <- which(base::rowSums(as.matrix(count_table)) == 0)
    count_table <- count_table[-row.remove.index,]
  }
  ids <- rownames(count_table)
  tids <- unlist(lapply(ids, FUN = animalcules::grep_tid))
  if (sum(is.na(tids)) > 0){
    tid_remove <- which(is.na(tids))
    ids <- ids[-tid_remove]
    tids <- tids[-tid_remove]
    count_table <- count_table[-tid_remove,]
  }
  taxonLevels <- animalcules::find_taxonomy(tids)
  tax_table <- animalcules::find_taxon_mat(ids, taxonLevels)
  # replace spaces in tax name with underscore
  tax_table <- as.data.frame(apply(tax_table,
                                   2,
                                   function(x) gsub('\\s+', '_',x)))
  
  # create MAE object
  se_mgx <-
    count_table |>
    base::data.matrix() |>
    S4Vectors::SimpleList() |>
    magrittr::set_names("MGX")
  
  # Read in metadata
  se_colData <-
    metadata_table |>
    S4Vectors::DataFrame()
  
  se_rowData <-
    tax_table |>
    base::data.frame() |>
    dplyr::mutate_all(as.character) |>
    dplyr::select("superkingdom", "phylum", "class", "order",
                  "family", "genus", "species") |>
    S4Vectors::DataFrame()
  
  microbe_se <-
    SummarizedExperiment::SummarizedExperiment(assays = se_mgx,
                                               colData = se_colData,
                                               rowData = se_rowData)
  mae_experiments <-
    S4Vectors::SimpleList(MicrobeGenetics = microbe_se)
  
  MAE <-
    MultiAssayExperiment::MultiAssayExperiment(experiments = mae_experiments,
                                               colData = se_colData)
  
  return(MAE)
}

# Helper function
comp_line <- function(this_line, tax_cols) {
  new_line <- this_line
  empty_ind <- which(is.na(this_line))
  if (length(empty_ind) == length(tax_cols)) return(rep("unknown", length(tax_cols)))
  if (length(empty_ind) == 0) return(new_line)
  
  for (i in empty_ind) {
    filltax <- i - 1
    if (filltax %in% empty_ind) {
      new_line[i] <- new_line[filltax]
    } else {
      new_line[i] <- paste(tolower(substr(tax_cols[filltax], 1, 1)),
                           new_line[filltax], sep = "_")
    }
  }
  return(new_line)
}

# Helper function
read_in_id_silva <- function(path_id_counts, end_string,
                             species_headers_all) {
  name_file <- utils::tail(stringr::str_split(path_id_counts, "/")[[1]],
                           n = 1)
  output <- readr::read_csv(path_id_counts, show_col_types = FALSE) %>%
    dplyr::filter(!is.na(.data$TaxonomyID)) %>%
    dplyr::select("read_count", "TaxonomyID", "Genome") %>%
    dplyr::mutate(sample = stringr::str_remove(name_file, end_string))
  counts_table <- output %>% dplyr::select(-"Genome")
  tax_table <- output %>% dplyr::select("TaxonomyID", "Genome")
    #species_headers_all %>%
    #dplyr::filter("TaxonomyID" %in% output$TaxonomyID)
  final_output <- list(counts_table, tax_table)
    return(final_output)
}

# Helper function
organize_tax_counts <- function(combined_list, tax_table) {
  order_ind <- match(combined_list$TaxonomyID, tax_table$TaxonomyID)
  tax_table_1 <- tax_table[order_ind, ]  %>%
    dplyr::mutate(dplyr::across(dplyr::everything(),
                                function(x) stringr::str_remove_all(x, "\\[|\\]")),
                  dplyr::across(dplyr::everything(), stringr::str_to_sentence))
  tax_table_2 <- apply(tax_table_1, 1, comp_line,
                       tax_cols = colnames(tax_table_1)[seq_len(8)]) %>%
    t() %>%
    magrittr::set_rownames(tax_table_1$TaxonomyID) %>%
    magrittr::set_colnames(colnames(tax_table_1)) %>%
    as.data.frame()
  prelim_counts <- combined_list %>%
    tibble::column_to_rownames("TaxonomyID")
  # Upsample counts
  counts_up <- animalcules::upsample_counts(prelim_counts, tax_table_2, "species")
  tax_up <- tax_table_2[!duplicated(tax_table_2[, "species"]), ]
  rownames(tax_up) <- tax_up$species
  order_ind_2 <- match(rownames(counts_up), rownames(tax_up))
  tax_final <- tax_up[order_ind_2, ]
  return(list(counts = counts_up, tax = tax_final))
}

# Helper function
add_in_taxa_silva <- function(combined_pre, caching, path_to_write) {
  location <- "https://github.com/wejlab/metascope-docs/raw/main/all_silva_headers.rds"
  filename <- "all_silva_headers.rds"
  if (!caching) {
    if (!dir.exists(path_to_write)) dir.create(path_to_write)
    destination <- paste(path_to_write, filename, sep = "/")
    utils::download.file(location, destination)
  } else if (caching) {
    bfc <- .get_cache()
    rid <- BiocFileCache::bfcquery(bfc, filename, "rname")$rid
    if (!length(rid)) {
      rid <- names(BiocFileCache::bfcadd(bfc, filename, location))
    }
    if (!isFALSE(BiocFileCache::bfcneedsupdate(bfc, rid))) {
      BiocFileCache::bfcdownload(bfc, rid)
      BiocFileCache::bfcrpath(bfc, rids = rid)
    } else {
      message("Caching is set to TRUE, ",
              "and it appears that this file is already downloaded ",
              "in the cache. It will not be downloaded again.")
    }
    destination <- BiocFileCache::bfcrpath(bfc, rids = rid)
  }
  all_silva_headers <- readRDS(destination)
  tax_table_pre_1 <- lapply(combined_pre, function(x) x[[2]]) %>%
    dplyr::bind_rows()
  tax_table_pre <- tax_table_pre_1[!duplicated(tax_table_pre_1[, "TaxonomyID"]), ] %>%
    dplyr::left_join(all_silva_headers, by = c("TaxonomyID")) %>%
    dplyr::relocate("genus", "species", "TaxonomyID", .after = "family") %>%
    dplyr::select(-"Genome")
  return(tax_table_pre)
}

# Helper function
transform_cols <- function(x) {
  stringr::str_split_i(x, "__", i = -1) %>% stringr::str_replace_all("-", "_")
}

#' Create a multi-assay experiment from MetaScope output for usage with
#' animalcules with the SILVA 13_8 database
#'
#' Upon completion of the MetaScope pipeline, users can analyze and visualize
#' abundances in their samples using the animalcules package. This function
#' allows interoperability of \code{metascope_id} output with both animalcules
#' and QIIME. After running this function, the user should save the returned MAE
#' to an RDS file using a function like \code{saveRDS} to upload the output into
#' the \code{animalcules} package. NOTE: This function is for outputs that were
#' generated with the SILVA 13_8 database.
#'
#' @param meta_counts A vector of filepaths to the counts ID CSVs output by
#'   \code{metascope_id()} created with the SILVA database.
#' @inheritParams convert_animalcules
#' @inheritParams download_refseq
#' @returns Returns a MultiAssay Experiment file of combined sample counts data
#'   and/or saved biom file and mapping file for analysis with QIIME. The 
#'   MultiAssayExperiment will have a counts assay ("MGX").
#' @export
#' @importFrom rlang .data
#' @examples
#' tempfolder <- tempfile()
#' dir.create(tempfolder)
#'
#' # Create three different samples
#' samp_names <- c("X123", "X456", "X789")
#' all_files <- file.path(tempfolder,
#'                        paste0(samp_names, ".csv"))
#'
#' create_IDcsv <- function (out_file) {
#'   final_taxids <- c("AY846380.1.2583", "AY909584.1.2313", "HG531388.1.1375")
#'   final_genomes <- rep("Genome name", 3)
#'   best_hit <- sample(seq(100, 1050), 3)
#'   proportion <- best_hit/sum(best_hit) |> round(2)
#'   EMreads <- best_hit + round(runif(3), 1)
#'   EMprop <- proportion + 0.003
#'   dplyr::tibble("TaxonomyID" = final_taxids,
#'                 "Genome" = final_genomes,
#'                 "read_count" = best_hit, "Proportion" = proportion,
#'                 "EMreads" = EMreads, "EMProportion" = EMprop) |>
#'     dplyr::arrange(dplyr::desc(.data$read_count)) |>
#'     utils::write.csv(file = out_file, row.names = FALSE)
#'   message("Done!")
#'   return(out_file)
#' }
#' out_files <- vapply(all_files, create_IDcsv, FUN.VALUE = character(1))
#'
#' # Create annotation data for samples
#' annot_dat <- file.path(tempfolder, "annot.csv")
#' dplyr::tibble(Sample = samp_names, RSV = c("pos", "neg", "pos"),
#'               month = c("March", "July", "Aug"),
#'               yrsold = c(0.5, 0.6, 0.2)) |>
#'   utils::write.csv(file = annot_dat,
#'                    row.names = FALSE)
#'
#' # Convert samples to MAE
#' outMAE <- convert_animalcules_silva(meta_counts = out_files,
#'                                     annot_path = annot_dat,
#'                                     which_annot_col = "Sample",
#'                                     end_string = ".metascope_id.csv",
#'                                     qiime_biom_out = FALSE,
#'                                     caching = TRUE)
#'
#' unlink(tempfolder, recursive = TRUE)
#' 

convert_animalcules_silva <- function(meta_counts, annot_path, which_annot_col,
                                      end_string = ".metascope_id.csv",
                                      qiime_biom_out = FALSE, path_to_write = ".",
                                      caching = TRUE) {
  combined_pre <- lapply(meta_counts, read_in_id_silva, end_string = end_string)
  combined_list <- lapply(combined_pre, function(x) x[[1]]) %>%
    data.table::rbindlist() %>% as.data.frame() %>%
    dplyr::mutate(sample = stringr::str_remove_all(sample, ".csv")) %>%
    dplyr::select("read_count", "TaxonomyID", "sample") %>%
    tidyr::pivot_wider(
      id_cols = "TaxonomyID", names_from = "sample",
      values_from = "read_count", values_fill = 0, id_expand = TRUE) %>%
    dplyr::group_by(.data$TaxonomyID) %>%
    dplyr::summarise(dplyr::across(dplyr::where(is.numeric), sum))
  
  tax_table_pre <- add_in_taxa_silva(combined_pre, caching, path_to_write)
  final_res <- organize_tax_counts(combined_list, tax_table_pre)
  MAE <- create_MAE(annot_path, which_annot_col, final_res$counts,
                    final_res$tax, path_to_write, qiime_biom_out)
  return(MAE)
}

#' Make a Subread index
#'
#' This function is a wrapper for the \code{Rsubread::buildindex} function. It
#' will generate one or more Subread indexes from a .fasta file. If the library
#' is too large (default >4GB) it will automatically be split into multiple
#' indexes, with _1, _2, etc at the end of the ref_lib basename.
#'
#' @param ref_lib The name/location of the reference library file, in
#'   (uncompressed) .fasta format.
#' @param split The maximum allowed size of the genome file (in GB). If the
#'   \code{ref_lib} file is larger than this, the function will split the
#'   library into multiple parts.
#' @param mem The maximum amount of memory (in MB) that can be used by the index
#'   generation process (used by the Rsubread::buildindex function).
#' @param quiet Turns off most messages. Default is \code{TRUE}.
#'
#' @return Creates one or more Subread indexes for the supplied reference .fasta
#'   file. If multiple indexes are created, the libraries will be named the
#'   \code{ref_lib} basename + "_1", "_2", etc. The function returns the names of the
#'   folders holding these files.
#'
#' @export
#'
#' @examples
#' #### Create a subread index from the example reference library
#'
#' ## Create a temporary directory to store the reference library
#' ref_temp <- tempfile()
#' dir.create(ref_temp)
#' tmp_accession <- system.file("extdata", "example_accessions.sql", package = "MetaScope")
#'
#' ## Download reference genome
#' out_fasta <- download_refseq('Orthoebolavirus zairense', reference = FALSE,
#'                              representative = FALSE, out_dir = ref_temp,
#'                              compress = TRUE, patho_out = FALSE,
#'                              caching = TRUE, accession_path = tmp_accession)
#'
#' ## Make subread index of reference library
#' mk_subread_index(out_fasta)
#' unlink(ref_temp)
#'

mk_subread_index <- function(ref_lib, split = 4, mem = 8000, quiet = TRUE) {
  gb <- 1073741824
  ref_size <- file.info(ref_lib)$size
  split_libs <- ceiling(ref_size / (split * gb))
  if (ref_size > (split * gb)) {
    if (!quiet) message("Library size is ", round(ref_size / gb, 1),
                        " GBs. Splitting the library into ", split_libs,
                        " separate components.")
    # Making connections to the .fasta library
    # generating the new split .fasta files
    con <- file(ref_lib, open = "r")
    out_cons <- paste("outcon_", seq_len(split_libs), sep = "")
    out_files <- paste(tools::file_path_sans_ext(ref_lib),
                       "_", seq_len(split_libs), ".",
                       tools::file_ext(ref_lib), sep = "")
    for (i in seq_len(split_libs)) {
      assign(out_cons[i], file(out_files[i], open = "w"))
    }
    ## Reading ref library and splitting to split libraries
    ngenomes <- -1
    while ((length(one_line <-
                   readLines(con, n = 1, warn = FALSE)) > 0)) {
      if (substr(one_line, 1, 1) == ">") {
        ngenomes <- ngenomes + 1
      }
      writeLines(one_line, get(out_cons[(ngenomes %% split_libs) + 1]))
    }
    if (!quiet) message("Printed ", ngenomes + 1, " sequences to ",
                        split_libs, " genome files")
    close(con)
    for (ocon in out_cons) close(get(ocon))
    for (lib in out_files) Rsubread::buildindex(
      basename = tools::file_path_sans_ext(lib),
      reference = lib, memory = mem)
  } else {
    if (!quiet) message("Library size is ", round(ref_size / gb, 1),
                        " GBs. Building the Rsubread index")
    Rsubread::buildindex(
      basename = tools::file_path_sans_ext(ref_lib, compression = TRUE),
      reference = ref_lib, memory = mem)
  }
  return(
    paste(tools::file_path_sans_ext(ref_lib, compression = TRUE),
          "_", seq_len(split_libs), sep = ""))
}

#' Download RefSeq 16S rRNA Bacterial and Archael libraries
#'
#' This function will automatically download the 16S rRNA RefSeq libraries from
#' \code{https://ftp.ncbi.nlm.nih.gov/refseq/TargetedLoci/} and combine them
#' into a single .fna file
#'
#' @param out_dir Character string giving the name of the directory to which
#'   libraries should be output. **Required**
#' @param combined_name Name of output combined file. Default is
#'   \code{"refseq_16S.fna"}
#'
#' @return Returns a character string of the path to the combined 16S rRNA .fna
#'    file
#'
#' @note This function requires the suggested packages \pkg{RCurl} and
#'    \pkg{R.utils}. If they are not installed, you will need to install
#'    them manually.
#'
#' @export
#' @examples
#' \dontrun{
#' #### Download 16S rRNA Genomes
#'
#' download_refseq_16S(out_dir = "out_dir",
#'   combined_name = "refseq_16S.fna")
#'   }


download_refseq_16S <- function(out_dir,
                                combined_name = "refseq_16S.fna") {
  # Check for required suggested packages
  if (!requireNamespace("RCurl", quietly = TRUE)) {
    stop("Package 'RCurl' is required for download_refseq_16s_only(). Please install it.", call. = FALSE)
  }
  if (!requireNamespace("R.utils", quietly = TRUE)) {
    stop("Package 'R.utils' is required for download_refseq_16s_only(). Please install it.", call. = FALSE)
  }
  if (!requireNamespace("stringr", quietly = TRUE)) {
    stop("Package 'stringr' is required for download_refseq_16s_only(). Please install it.", call. = FALSE)
  }

  # Create Directory
  dir.create(out_dir, showWarnings = FALSE, recursive = TRUE)

  base_urls <- c(
    "ftp://ftp.ncbi.nlm.nih.gov/refseq/TargetedLoci/Bacteria/",
    "ftp://ftp.ncbi.nlm.nih.gov/refseq/TargetedLoci/Archaea/"
  )

  download_from_url <- function(url, local_dir) {
    listing <- stringr::str_split(RCurl::getURL(url, ftp.use.epsv = TRUE, dirlistonly = TRUE), "\n")[[1]]
    listing <- stringr::str_trim(listing)
    listing <- listing[listing != ""]

    for (item in listing) {
      if (grepl("/$", item)) {
        # Recurse into subdirectories
        new_url <- paste0(url, item)
        download_from_url(new_url, local_dir)
      } else if (grepl("16S.*\\.fna\\.gz$", item)) {
        # Only download 16S FASTA files
        file_url <- paste0(url, item)
        dest_file <- file.path(local_dir, item)
        tryCatch({
          utils::download.file(file_url, destfile = dest_file, quiet = TRUE, mode = "wb")
          message("Downloaded: ", file_url)
        }, error = function(e) {
          message("Failed to download: ", file_url)
        })
      }
    }
  }

  # Download all 16S files
  for (base_url in base_urls) {
    download_from_url(base_url, out_dir)
  }

  # Gunzip all downloaded files
  gz_files <- list.files(out_dir, pattern = "\\.fna\\.gz$", full.names = TRUE)
  message("Decompressing ", length(gz_files), " files...")

  for (gz in gz_files) {
    tryCatch({
      R.utils::gunzip(gz, overwrite = TRUE, remove = TRUE)  # decompress & remove .gz
      message("Unzipped: ", gz)
    }, error = function(e) {
      message("Failed to unzip: ", gz)
    })
  }

  # Concatenate into one combined FASTA
  fna_files <- list.files(out_dir, pattern = "\\.fna$", full.names = TRUE)
  combined_path <- file.path(out_dir, combined_name)

  message("Concatenating into: ", combined_path)
  file.create(combined_path)  # ensure file exists
  for (fna in fna_files) {
    system(paste("cat", shQuote(fna), ">>", shQuote(combined_path)))
  }

  message("All 16S FASTAs combined into: ", combined_path)

  # Delete original fna_files
  unlink(fna_files)

  return(combined_path)
}

globalVariables("count")

get_max_index_matrix <- function(mat) {
  stopifnot(inherits(mat, "Matrix"))  # Ensure sparse Matrix
  # Convert to triplet format for efficient row-wise max
  trip <- Matrix::summary(mat)
  row_max <- tapply(trip$x, trip$i, max)
  # Keep only entries that match the row max
  keep <- trip$x == row_max[as.character(trip$i)]
  # Build a new sparse logical matrix marking max positions
  is_max <- Matrix::sparseMatrix(i = trip$i[keep],
               j = trip$j[keep],
               x = TRUE,
               dims = dim(mat))
  return(is_max)
}

obtain_reads <- function(input_file, input_type, aligner, quiet) {
  to_pull <- c("qname", "rname", "cigar", "qwidth", "pos")
  if (identical(input_type, "bam")) {
    if (!quiet) message("Reading .bam file: ", input_file)
    if (identical(aligner, "bowtie2")) {
      params <- Rsamtools::ScanBamParam(what = to_pull, tag = c("AS"))
    } else if (identical(aligner, "subread")) {
      params <- Rsamtools::ScanBamParam(what = to_pull, tag = c("NM"))
    } else if (identical(aligner, "other")) {
      params <- Rsamtools::ScanBamParam(what = to_pull)
    }
    reads <- Rsamtools::scanBam(input_file, param = params)
  } else if (identical(input_type, "csv.gz")) {
    if (!quiet) message("Reading .csv.gz file: ", input_file)
    reads <- data.table::fread(input_file, sep = ",", header = FALSE) %>%
      magrittr::set_colnames(c("qname", "rname", "cigar", "qwidth", "pos", "tag")) %>% as.list() %>% list()
    if (identical(aligner, "bowtie2")) {
      reads[[1]]$tag <- list("AS" = reads[[1]]$tag)
    } else if (identical(aligner, "subread")) {
      reads[[1]]$tag <- list("NM" = reads[[1]]$tag)
    }
  }
  return(reads)
}

identify_rnames <- function(reads, unmapped = NULL) {
  reads_in <- reads[[1]]$rname
  if(!is.null(unmapped)) reads_in <- reads[[1]]$rname[!unmapped]
  # FOR 2015 - if >50% have 8 pipe symbols
  if (mean(stringr::str_count(reads_in, "\\|") == 8) > 0.5) {
    mapped_rname <- stringr::str_split(reads_in, "\\|") |>
      plyr::laply(function(x) magrittr::extract(x, 8))
    return(mapped_rname)
  }
  # https://www.ncbi.nlm.nih.gov/books/NBK21091/table/ch18.T.refseq_accession_numbers_and_mole
  prefixes <- c("AC", "NC", "NG", "NT", "NW", "NZ", "NR") %>%
    paste0("_") %>%
    paste(collapse = "|")
  this_read <- stringr::str_split_i(reads_in, prefixes, -1) |>
    stringr::str_split_i("[| ]", 1)
  read_prefix <- stringr::str_extract(reads_in, prefixes)
  mapped_rname <- paste0(read_prefix, this_read)
  return(mapped_rname)
}


find_accessions <- function(accessions, accession_path, quiet) {
  # Convert accessions to taxids and get genome names
  if (!quiet) message("Obtaining taxonomy and genome names")
  taxids <- taxonomizr::accessionToTaxa(accessions, sqlFile = accession_path)
  genome_names <- apply(taxonomizr::getTaxonomy(taxids, sqlFile = accession_path,
                                                desiredTaxa = c("superkingdom", "phylum", "class",
                                                                "order", "family", "genus", "species", "strain")),
                        1,function(x) paste0(x, collapse = ";"))
  return(genome_names)
}

get_alignscore <- function(aligner, cigar_strings, count_matches, scores,
                           qwidths) {
  #Subread alignment scores: CIGAR string matches - edit score
  if (identical(aligner, "subread")) {
    num_match <- unlist(vapply(cigar_strings, count_matches,
                               USE.NAMES = FALSE, double(1)))
    alignment_scores <- num_match - scores
    scaling_factor <- 1.0 / max(alignment_scores)
    relative_alignment_scores <- alignment_scores - min(alignment_scores)
    exp_alignment_scores <- exp(relative_alignment_scores * scaling_factor)
  } else if (identical(aligner, "bowtie2")) {
    # Bowtie2 alignment scores: AS value + read length (qwidths)
    alignment_scores <- scores + qwidths
    scaling_factor <- 1.0 / max(alignment_scores)
    relative_alignment_scores <- alignment_scores - min(alignment_scores)
    exp_alignment_scores <- exp(relative_alignment_scores * scaling_factor)
  } else if (identical(aligner, "other")) {
    # Other alignment scores: No assumptions
    exp_alignment_scores <- 1
  }
  return(exp_alignment_scores)
}


get_assignments <- function(combined, convEM, maxitsEM, unique_taxids,
                            unique_genome_names, update_bam = TRUE, input_file, priors_df, quiet) {
  combined$index <- seq.int(1, nrow(combined))
  input_distinct <- dplyr::distinct(combined, .data$qname, .data$rname,
                                    .keep_all = TRUE)
  qname_inds_2 <- input_distinct$qname
  rname_tax_inds_2 <- input_distinct$rname
  scores_2 <- input_distinct$scores
  non_unique_read_ind <- unique(combined[[1]][(
    duplicated(input_distinct[, 1]) | duplicated(input_distinct[, 1],
                                                 fromLast = TRUE))])
  # 1 if read is multimapping, 0 if read is unique
  y_ind_2 <- as.numeric(unique(input_distinct[[1]]) %in% non_unique_read_ind)
  gammas <- Matrix::sparseMatrix(qname_inds_2, rname_tax_inds_2, x = scores_2)
  pi_old <- rep(1 / nrow(gammas), ncol(gammas))
  pi_new <- theta_new <- Matrix::colMeans(gammas)
  if (!is.null(priors_df)) {
    if (colnames(priors_df)[1] != "species") {
      stop("priors_df must have two columns. \n",
           "The first column is named 'species' and contains the species names. \n",
           "The second column is named 'prior_weights' and contains the weights.")
    }
    colnames(priors_df)[2] <- "prior_weights"
    weights <- tidyr::as_tibble(unique_genome_names) |>
      dplyr::rename("species" = "value") |>
      dplyr::left_join(priors_df, by = "species") |>
      tidyr::replace_na(replace = list(prior_weights = 0))
    unique_reads <- weights$prior_weights * max(combined$qname)
    exp_weights <- exp(weights$prior_weights)
    #weighted_pi_new <- pi_new + weights / (1 + weights)
    posterior <- pi_new * exp_weights
    pi_new <- posterior
  } else {
    unique_reads = 0
  }
  epsilon = 1e-8
  conv <- max(abs(pi_new - pi_old) / (pi_old + epsilon))
  it <- 0
  if (!quiet) message("Starting EM iterations")
  while (conv > convEM && it < maxitsEM) {
    # Expectation Step: Estimate expected value for each read to ea genome
    pi_mat <- Matrix::sparseMatrix(qname_inds_2, rname_tax_inds_2,
                                   x = pi_new[rname_tax_inds_2])
    theta_mat <- Matrix::sparseMatrix(qname_inds_2, rname_tax_inds_2,
                                      x = theta_new[rname_tax_inds_2])
    weighted_gamma <- gammas * pi_mat * theta_mat
    weighted_gamma_sums <- Matrix::rowSums(weighted_gamma)
    gammas_new <- weighted_gamma / weighted_gamma_sums
    # Maximization step: proportion of reads to each genome
    pi_new <- (Matrix::colSums(gammas_new) + unique_reads) / (sum(Matrix::colSums(gammas_new)) + sum(unique_reads))
    theta_new_num <- (Matrix::colSums(y_ind_2 * gammas_new) + 1)
    theta_new <- theta_new_num / (nrow(gammas_new) + 1)
    # Check convergence
    it <- it + 1
    conv <- max(abs(pi_new - pi_old) / (pi_old + epsilon), na.rm = TRUE)
    pi_old <- pi_new
    if (!quiet) message(c(it, " ", conv))
  }
  if (!quiet) message("DONE! Converged in ", it, " iterations.")
  hit_which <- get_max_index_matrix(gammas_new)
  combined <- combined |>
    dplyr::mutate("hit" = hit_which[cbind(!!rlang::sym("qname"), !!rlang::sym("rname"))])
  combined_single <- combined |>
    dplyr::group_by(!!rlang::sym("qname"), !!rlang::sym("rname")) |>
    dplyr::mutate("best_hit" = !!rlang::sym("hit") &
                    !!rlang::sym("scores") == max(!!rlang::sym("scores"))) |>
    dplyr::ungroup() |>
    dplyr::group_by(!!rlang::sym("qname")) |>
    dplyr::mutate("single_hit" = dplyr::row_number() == min(dplyr::row_number()[!!rlang::sym("best_hit")]))
  combined_distinct <- dplyr::distinct(combined, !!rlang::sym("qname"),
                                       !!rlang::sym("rname"), .keep_all = TRUE) |>
    dplyr::filter(!!rlang::sym("hit") == TRUE)
  best_hit <- Matrix::colSums(hit_which)
  names(best_hit) <- seq_along(best_hit)
  best_hit <- best_hit[best_hit != 0]
  hits_ind <- as.numeric(names(best_hit))
  final_taxids <- unique_taxids[hits_ind]
  final_genomes <- unique_genome_names[hits_ind]
  proportion <- best_hit / sum(best_hit)
  gammasums <- Matrix::colSums(gammas_new)
  readsEM <- round(gammasums[hits_ind], 1)
  propEM <- gammasums[hits_ind] / sum(gammas_new)
  results_tibble <- dplyr::tibble(TaxonomyID = final_taxids, Genome = final_genomes,
                                  read_count = best_hit, Proportion = proportion,
                                  readsEM = readsEM, EMProportion = propEM,
                                  hits_ind = hits_ind) %>%
    dplyr::arrange(dplyr::desc(.data$read_count))
  if (!quiet) message("Found reads for ", nrow(results_tibble), " genomes")

  return(list(results_tibble, combined_distinct, combined_single))
}

#' Count the number of base lengths in a CIGAR string for a given operation
#'
#' The 'CIGAR' (Compact Idiosyncratic Gapped Alignment Report) string is how the
#' SAM/BAM format represents spliced alignments. This function will accept a
#' CIGAR string for a single read and a single character indicating the
#' operation to be parsed in the string. An operation is a type of column that
#' appears in the alignment, e.g. a match or gap. The integer following the
#' operator specifies a number of consecutive operations. The
#' \code{count_matches()} function will identify all occurrences of the operator
#' in the string input, add them, and return an integer number representing the
#' total number of operations for the read that was summarized by the input
#' CIGAR string.
#'
#' This function is best used on a vector of CIGAR strings using an apply
#' function (see examples).
#'
#' @param x Character. A CIGAR string for a read to be parsed. Examples of
#' possible operators include "M", "D", "I", "S", "H", "=", "P", and "X".
#' @param char A single letter representing the operation to total for the
#' given string.
#'
#' @return an integer number representing the total number of alignment
#' operations for the read that was summarized by the input CIGAR string.
#'
#' @export
#'
#' @examples
#' # A single cigar string: 3M + 3M + 5M
#' cigar1 <- "3M1I3M1D5M"
#' count_matches(cigar1, char = "M")
#'
#' # Parse with operator "P": 2P
#' cigar2 <- "4M1I2P9M"
#' count_matches(cigar2, char = "P")
#'
#' # Apply to multiple strings: 1I + 1I + 5I
#' cigar3 <- c("3M1I3M1D5M", "4M1I1P9M", "76M13M5I")
#' vapply(cigar3, count_matches, char = "I",
#'        FUN.VALUE = numeric(1))
#'

count_matches <- function(x, char = "M") {
  if (length(char) != 1) {
    stop("Please provide a single character ",
         "operator with which to parse.")
  } else if (length(x) != 1) {
    stop("Please provide a single CIGAR string to be parsed.")
  }
  pattern <- paste0("\\d+", char)
  ind <- gregexpr(pattern, x)[[1]]
  start <- as.numeric(ind)
  end <- start + attr(ind, "match.length") - 2
  out <- cbind(start, end) %>% apply(
    1, function(y) substr(x, start = y[1], stop = y[2])) %>%
    as.numeric() %>% sum()
  return(data.table::fifelse(is.na(out[1]), yes = 0, no = out[1]))
}

#' Helper Function for MetaScope ID
#'
#' Used to create plots of genome coverage for any number of accession numbers
#'
#' @param which_taxid Which taxid to plot
#' @param which_genome Which genome to plot
#' @param accessions List of accessions from \code{metascope_id()}
#' @param taxids List of accessions from \code{metascope_id()}
#' @param reads List of reads from input file
#' @param out_base The basename of the input file
#' @param out_dir The path to the input file
#'
#' @return A plot of the read coverage for a given genome

locations <- function(which_taxid, which_genome,
                      accessions, taxids, reads, out_base, out_dir) {
  plots_save <- file.path(out_dir, paste(out_base, "cov_plots",
                                         sep = "_"))
  if (!dir.exists(plots_save)) dir.create(plots_save)
  # map back to accessions
  choose_acc <- paste(accessions[which(as.numeric(taxids) %in% which_taxid)])
  # map back to BAM
  map2bam_acc <- which(reads[[1]]$rname %in% choose_acc)
  # Split genome name to make digestible
  this_genome <- strsplit(which_genome, " ")[[1]][c(1, 2)]
  use_name <- paste(this_genome, collapse = " ") %>% stringr::str_replace(",", "")
  coverage <- round(mean(seq_len(338099) %in% unique(
    reads[[1]]$pos[map2bam_acc])), 3)
  # Plotting
  dfplot <- dplyr::tibble(x = reads[[1]]$pos[map2bam_acc])
  ggplot2::ggplot(dfplot, ggplot2::aes(.data$x, fill = ggplot2::after_stat(count))) +
    ggplot2::geom_histogram(bins = 50) +
    ggplot2::scale_fill_gradient(low = 'red', high = 'yellow') +
    ggplot2::theme_classic() +
    ggplot2::labs(title = bquote("Positions of reads mapped to"~italic(.(use_name))),
                  xlab = "Aligned position across genome (leftmost read position)",
                  ylab = "Read Count",
                  caption = paste0("Accession Number: ", choose_acc))

  ggplot2::ggsave(paste0(plots_save, "/",
                         stringr::str_replace(use_name, " ", "_"), ".png"),
                  device = "png")
}

#' Identify which genomes are represented in a processed sample
#'
#' This function will read in a .bam or .csv.gz file, annotate the taxonomy and
#' genome names, reduce the mapping ambiguity using a mixture model, and output
#' a .csv file with the results. Currently, it assumes that the genome
#' library/.bam files use NCBI accession names for reference names (rnames in
#' .bam file).
#'
#' @param input_file The .bam or .csv.gz file of sample reads to be identified.
#' @param input_type Extension of file input. Should be either "bam" or
#'   "csv.gz". Default is "csv.gz".
#' @param aligner The aligner which was used to create the bam file. Default is
#'   "bowtie2" but can also be set to "subread" or "other".
#' @param db Currently accepts one of \code{c("ncbi", "silva", "other")} Default
#'   is \code{"ncbi"}, appropriate for samples aligned against indices compiled
#'   from NCBI whole genome databases. Alternatively, usage of an alternate
#'   database (like Greengenes2) should be specified with \code{"other"}.
#' @param db_feature_table If \code{db = "other"}, a data.frame must be supplied
#'   with two columns, "Feature ID" matching the names of the alignment indices,
#'   and a second \code{character} column supplying the taxon identifying
#'   information.
#' @param accession_path (character) Filepath to NCBI accessions SQL database.
#'   See \code{taxonomzr::prepareDatabase()}.
#' @param out_dir The directory to which the .csv output file will be output.
#'   Defaults to \code{dirname(input_file)}.
#' @param convEM The convergence parameter of the EM algorithm. Default set at
#'   \code{1/10000}.
#' @param maxitsEM The maximum number of EM iterations, regardless of whether
#'   the convEM is below the threshhold. Default set at \code{50}. If set at
#'   \code{0}, the algorithm skips the EM step and summarizes the .bam file 'as
#'   is'.
#' @param num_species_plot The number of genome coverage plots to be saved.
#'   Default is \code{NULL}, which saves coverage plots for the ten most highly
#'   abundant species.
#' @param update_bam Whether to update BAM file with new read assignments.
#'   Default is \code{FALSE}. If \code{TRUE}, requires \code{input_type = "bam"}
#'   such that a BAM file is the input to the function.
#' @param quiet Turns off most messages. Default is \code{TRUE}.
#' @param tmp_dir Path to a directory to which bam and updated bam files can be
#'   saved. Required.
#' @param priors_df data.frame containing priors data. The data.frame consists
#'   of two columns, 'species' containing species name, and 'prior_weights'
#'   containing the prior weights (as a percent; integer).
#' @param group_by_taxa Character. Taxonomy level at which accessions should be
#'   grouped. Defaults to \code{"species"}
#'
#' @return This function exports a .csv file with annotated read counts to
#'   genomes with mapped reads to the location returned by the function.
#'   Depending on the parameters specified, can also output an updated BAM
#'   file, and fasta files for additional analysis downstream.
#'
#' @export
#'
#' @examples
#' #### Align reads to reference library and then apply metascope_id()
#' ## Assuming filtered bam files already exist
#'
#' ## Create temporary directory
#' file_temp <- tempfile()
#' dir.create(file_temp)
#'
#' ## Get temporary accessions database
#' tmp_accession <- system.file("extdata", "example_accessions.sql", package = "MetaScope")
#'
#' #### Subread aligned bam file
#'
#' ## Create object with path to filtered subread csv.gz file
#' filt_file <- "subread_target.filtered.csv.gz"
#' bamPath <- system.file("extdata", filt_file, package = "MetaScope")
#' file.copy(bamPath, file_temp)
#'
#' ## Run metascope id with the aligner option set to subread
#' metascope_id(input_file = file.path(file_temp, filt_file),
#'              aligner = "subread", num_species_plot = 0,
#'              input_type = "csv.gz", accession_path = tmp_accession)
#'
#' #### Bowtie 2 aligned .csv.gz file
#'
#' ## Create object with path to filtered bowtie2 bam file
#' bowtie_file <- "bowtie_target.filtered.csv.gz"
#' bamPath <- system.file("extdata", bowtie_file, package = "MetaScope")
#' file.copy(bamPath, file_temp)
#'
#' ## Run metascope id with the aligner option set to bowtie2
#' metascope_id(file.path(file_temp, bowtie_file), aligner = "bowtie2",
#'              num_species_plot = 0, input_type = "csv.gz",
#'              accession_path = tmp_accession)
#'
#' ## Remove temporary directory
#' unlink(file_temp, recursive = TRUE)
#'

metascope_id <- function(input_file, input_type = "csv.gz",
                         aligner = "bowtie2",
                         db = "ncbi",
                         db_feature_table = NULL,
                         accession_path = NULL,
                         priors_df = NULL,
                         tmp_dir = dirname(input_file),
                         out_dir = dirname(input_file),
                         convEM = 1 / 10000,
                         maxitsEM = 25,
                         update_bam = FALSE,
                         num_species_plot = NULL,
                         group_by_taxa = "species",
                         quiet = TRUE)  {

  out_base <- input_file %>% base::basename() %>% strsplit(split = "\\.") %>%
    magrittr::extract2(1) %>% magrittr::extract(1)
  out_file <- file.path(out_dir, paste0(out_base, ".metascope_id.csv"))
  # Check to make sure valid aligner is specified
  if (!(aligner %in% c("bowtie2", "subread", "other"))) {
    stop("Please make sure aligner is set to either 'bowtie2', 'subread',",
         " or 'other'")
  }
  if (db == "other" && is.null(db_feature_table)) {
    stop("Please supply a data.frame for db_feature_table if 'db = other'")
  }
  if (input_type == "csv.gz") {
    if (!quiet) message("Note, cannot generate updated_bam from csv.gz file")
    update_bam <- FALSE
    if (update_bam) stop("`update_bam` is set to TRUE but input is a `csv.gz`")
  }
  # Check that tmp_dir is specified
  if (is.null(tmp_dir)) stop("Please supply a directory for 'tmp_dir' argument.")
  reads <- obtain_reads(input_file, input_type, aligner, quiet)
  unmapped <- is.na(reads[[1]]$rname)
  if (db == "ncbi") reads[[1]]$rname <- identify_rnames(reads)
  mapped_rname <- as.character(reads[[1]]$rname[!unmapped])
  mapped_qname <- reads[[1]]$qname[!unmapped]
  mapped_cigar <- reads[[1]]$cigar[!unmapped]
  mapped_qwidth <- reads[[1]]$qwidth[!unmapped]
  if (aligner == "bowtie2") {
    # mapped alignments used
    map_edit_or_align <- reads[[1]][["tag"]][["AS"]][!unmapped]
  } else if (aligner == "subread") {
    map_edit_or_align <-
      reads[[1]][["tag"]][["NM"]][!unmapped] # mapped edits used
  }
  read_names <- unique(mapped_qname)
  accessions <- as.character(unique(mapped_rname))

  # Let user decide which taxa level to group by
  if (is.null(group_by_taxa)) {
    group_by_taxa <- "species"
  }
  if (db == "ncbi") {
    if (is.null(accession_path)) stop("Please provide a valid accession_path argument")
    taxonomy_indices <- tidyr::tibble(accessions) |>
      dplyr::mutate("taxids" = taxonomizr::accessionToTaxa(accessions, sqlFile = accession_path)) |>
      dplyr::mutate("taxa_names" = taxonomizr::getTaxonomy(taxids, sqlFile = accession_path,
                                                         desiredTaxa = group_by_taxa)[,]) |>
      dplyr::mutate("taxa_names" = ifelse(is.na(!!dplyr::sym("taxa_names")), paste0("unknown genome; accession ID is", accessions), !!dplyr::sym("taxa_names"))) |>
      dplyr::mutate("taxa_index" = match(!!dplyr::sym("taxa_names"), unique(!!dplyr::sym("taxa_names"))))

  } else if (db == "silva") {
    tax_id_all <- stringr::str_split(accessions, ";", n =2)
    taxids <- sapply(tax_id_all, `[[`, 1)
    genome_names <- sapply(tax_id_all, `[[`, 2)
    # Fix names
    mapped_rname <- stringr::str_split(mapped_rname, ";", n = 2) %>%
      sapply(`[[`, 1)
    accessions <- as.character(unique(mapped_rname))
  } else if (db == "other") {
    tax_id_all <- dplyr::tibble(`Feature ID` = accessions) %>%
      dplyr::left_join(db_feature_table, by = "Feature ID")
    taxids <- tax_id_all %>% dplyr::select(1) %>% unlist() %>% unname()
    genome_names <- tax_id_all %>% dplyr::select(2) %>% unlist() %>%
      unname()
  }


  # Make an alignment matrix (rows: reads, cols: unique taxids)
  if (!quiet) message("Setting up the EM algorithm")
  qname_inds <- match(mapped_qname, read_names)
  if (db == "ncbi") {
    rname_tax_inds <- taxonomy_indices$taxa_index[match(mapped_rname, taxonomy_indices$accessions)]
    if (!quiet) message("\tFound ", length(taxonomy_indices$taxa_index),
                        " unique taxa")
    unique_tax_data <- taxonomy_indices |>
      dplyr::mutate("taxid_rank" = suppressWarnings(as.numeric(taxids))) |>
      dplyr::group_by(!!dplyr::sym("taxa_index")) |>
      dplyr::slice(which.min(ifelse(is.na(!!dplyr::sym("taxid_rank")),
                                    Inf, !!dplyr::sym("taxid_rank")))) |>
      dplyr::ungroup()

    unique_taxids <- unique_tax_data$taxids
    unique_genome_names <- unique_tax_data$taxa_names
  }
  # Order based on read names
  rname_tax_inds <- rname_tax_inds[order(qname_inds)]
  cigar_strings <- mapped_cigar[order(qname_inds)]
  qwidths <- mapped_qwidth[order(qname_inds)]

  if (aligner == "bowtie2") {
    # mapped alignments used
    scores <- map_edit_or_align[order(qname_inds)]
  } else if (aligner == "subread") {
    # mapped edits used
    scores <- map_edit_or_align[order(qname_inds)]
  } else if (aligner == "other") scores <- 1
  qname_inds <- sort(qname_inds)
  exp_alignment_scores <- get_alignscore(aligner, cigar_strings,
                                         count_matches, scores, qwidths)
  combined <- dplyr::bind_cols("qname" = qname_inds,
                               "rname" = rname_tax_inds,
                               "scores" = exp_alignment_scores)
  results <- get_assignments(combined, convEM, maxitsEM, unique_taxids,
                             unique_genome_names, quiet = quiet, priors_df = priors_df)
  metascope_id_file <- results[[1]] %>% dplyr::select("TaxonomyID", "Genome",
                                                      "read_count", "Proportion",
                                                      "readsEM", "EMProportion")
  utils::write.csv(metascope_id_file, file = out_file, row.names = FALSE)
  if (!quiet) message("Results written to ", out_file)

  if (update_bam) {
    # Convert accessions back into taxids. Some taxids may be from different accessions
    # This step may produce false alignments to accessions of the same taxid
    if (db == "ncbi") {
      accessions_taxids <- taxonomy_indices |>
        dplyr::rename(rname = "taxa_index",
                      rname_names = accessions)
    } else {
      accessions_taxids <- tidyr::tibble(rname_names = accessions, taxids, rname = match(taxids, unique(taxids)))
    }
    combined_distinct <- results[[2]]
    combined_distinct <- combined_distinct |>
      dplyr::right_join(accessions_taxids, combined_distinct, by = "rname", relationship = "many-to-many") |>
      dplyr::mutate(qname_names = read_names[.data$qname])

    if (db == "ncbi") {
      bam_index_df <- data.frame("original_bam_index" = seq_along(reads[[1]]$qname),
                                 "qname_names" = reads[[1]]$qname,
                                 "rname_names" = as.character(reads[[1]]$rname))
    }
    if (db == "silva"){
      bam_index_df <- data.frame("original_bam_index" = seq_along(reads[[1]]$qname),
                                 "qname_names" = reads[[1]]$qname,
                                 "rname_names" = sub(';.*$','', as.character(reads[[1]]$rname)))
    }

    combined_bam_index <- dplyr::right_join(bam_index_df, combined_distinct,
                                            by = (c("qname_names", "rname_names"))) |>
      dplyr::mutate("qname_rname" = paste(.data$qname, .data$rname, sep = "_"),
                    "first_qname_rname" = !duplicated(.data$qname_rname)) |>
      dplyr::filter(.data$first_qname_rname == TRUE)

    filter_which <- rep(FALSE, nrow(bam_index_df))
    filter_which[combined_bam_index$original_bam_index] <- TRUE
    # Filter based on batches and row 
    i <- 0L
    keep_rule <- S4Vectors::FilterRules(list(
      keep = function(rec) {
        n <- NROW(rec$qname)
        idx <- seq.int(i + 1L, length.out = n)
        i <<- i + n
        filter_which[idx]
      }
    ))
    param <- Rsamtools::ScanBamParam(what = "qname")  # just to know chunk size
    
    Rsamtools::indexBam(files = input_file)
    input_bam <- Rsamtools::BamFile(input_file, yieldSize = 1e6)
    
    bam_out <- file.path(tmp_dir, paste0(out_base, ".updated.bam"))
    
    Rsamtools::filterBam(
      input_bam,
      destination = bam_out,
      param = param,
      filter = keep_rule,
      indexDestination = TRUE
    )
    
  }
  # Plotting genome locations
  num_plot <- num_species_plot
  if (is.null(num_plot)) num_plot <- min(nrow(metascope_id_file), 10)
  if (num_plot > 0) {
    if (!quiet) message("Creating coverage plots at ",
                        out_base, "_cov_plots")
    lapply(seq_along(metascope_id_file$TaxonomyID)[seq_len(num_plot)],
           function(x) {
             locations(as.numeric(metascope_id_file$TaxonomyID)[x],
                       which_genome = metascope_id_file$Genome[x],
                       accessions, taxids, reads, out_base, out_dir)})
  } else if (!quiet) message("No coverage plots created")
  return(out_file)
}

#' A universal parameter settings object for Rsubread alignment
#'
#' This object is a named vector of multiple options that can be chosen for
#' functions that involve alignment with Rsubread, namely \code{align_target()}
#' and \code{filter_host()}. Both functions take an object for the parameter
#' \code{settings}, which are provided by \code{align_details} by default, or
#' may be given by a user-created object containing the same information.
#'
#' The default options included in \code{align_details} are \code{type = "dna"},
#' \code{maxMismatches = 3}, \code{nsubreads = 10}, \code{phredOffset = 33},
#' \code{unique = FALSE}, and \code{nBestLocations = 16}. Full descriptions of
#' these parameters can be read by accessing \code{?Rsubread::align}.
#' @name align_details
#' @docType data
#' @format list
#' @usage data(align_details)
#' @keywords datasets
#' @examples
#' data("align_details")
#'
"align_details"

#' A universal parameter object for Bowtie 2 16S alignment
#'
#' This character string provides several Bowtie 2 options to provide an
#' optimized alignment specifically optimized for 16S amplicon sequencing data.
#' This object can be used with functions that use the Bowtie 2 aligner through
#' the \code{Rbowtie2} package, namely \code{align_target_bowtie()} and
#' \code{filter_host_bowtie}. These settings can be substituted for default
#' settings by passing \code{} to the \code{bowtie2_options} argument.
#'
#' The default parameters listed in this object are
#' "--local -R 2 -N 0 -L 25 -i S,1,0.75 -k 5 --score-min L,0,1.88"
#' 
#' Note that k is actually 10 and is doubled internally from 5.
#' The \code{score-min} function was chosen such that the minimum alignment score
#' allowed requires 98% identity.
#'
#' Further delineation of Bowtie 2 parameters is provided in the
#' \href{https://bowtie-bio.sourceforge.net/bowtie2/manual.shtml}{Bowtie 2
#' manual}.
#'
#' @name bt2_16S_params
#' @docType data
#' @format list
#' @usage data(bt2_16S_params)
#' @keywords datasets
#' @examples
#' data("bt2_16S_params")
#'
"bt2_16S_params"

#' A universal parameter object for Bowtie 2 metagenomic or non-16S alignment
#'
#' This character string provides several Bowtie 2 options to provide a 95% identity
#' alignment useful for metagenomes. This object can be used with functions that
#' use the Bowtie 2 aligner through the \code{Rbowtie2} package, namely
#' \code{align_target_bowtie()} and \code{filter_host_bowtie}. These settings
#' can be substituted for default settings by passing \code{} to the
#' \code{bowtie2_options} argument.
#'
#' The default parameters listed in this object are
#' "--local -R 2 -N 0 -L 25 -i S,1,0.75 -k 5 --score-min L,0,1.7".
#'
#' Further delineation of Bowtie 2 parameters is provided in the
#' \href{https://bowtie-bio.sourceforge.net/bowtie2/manual.shtml}{Bowtie 2
#' manual}.
#'
#' @name bt2_regular_params
#' @docType data
#' @format list
#' @usage data(bt2_regular_params)
#' @keywords datasets
#' @examples
#' data("bt2_regular_params")
#'
"bt2_regular_params"

#' A universal parameter object for Bowtie 2 metagenomic alignment where the host genome is thought to be absent from the reference database
#'
#' This character string provides several Bowtie 2 options to conduct an 
#' alignment useful for metagenomes, especially in the case where a genome may
#' not be present in the reference database. This object can be used with functions that
#' use the Bowtie 2 aligner through the \code{Rbowtie2} package, namely
#' \code{align_target_bowtie()} and \code{filter_host_bowtie}. These settings
#' can be substituted for default settings by passing \code{} to the
#' \code{bowtie2_options} argument.
#'
#' The default parameters listed in this object are
#' "--local -R 2 -N 0 -L 25 -i S,1,0.75 -k 5 --score-min L,0,1.4".
#'
#' Further delineation of Bowtie 2 parameters is provided in the
#' \href{https://bowtie-bio.sourceforge.net/bowtie2/manual.shtml}{Bowtie 2
#' manual}.
#'
#' @name bt2_missing_params
#' @docType data
#' @format list
#' @usage data(bt2_missing_params)
#' @keywords datasets
#' @examples
#' data("bt2_missing_params")
#'
"bt2_missing_params"

# Helper functions for convert_animalcules

# Create biom file and mapping file
create_qiime_biom <- function(se_colData, taxonomy_table, which_annot_col,
                              counts_table, path_to_write) {
  # Consolidate counts based on genus
  counts_table_g <- counts_table |>
    dplyr::mutate(genus = taxonomy_table$genus) |>
    dplyr::group_by(!!dplyr::sym("genus")) |>
    dplyr::summarise(dplyr::across(.fns = sum)) |> as.data.frame()
  rownames(counts_table_g) <- counts_table_g$genus
  counts_table_g <- counts_table_g |> dplyr::select(-"genus")
  # Match Indices
  ind <- base::match(se_colData[, which_annot_col], colnames(counts_table_g))
  counts_table_match <- counts_table_g[, c(ind)]
  # Formatting
  to_rm <- "[^[:alnum:] _.\\-+%;/:,]"
  func <- function(x) stringr::str_remove_all(x, to_rm)
  alt_col <- se_colData |> dplyr::as_tibble() |>
    dplyr::mutate("#SampleID" = se_colData[, which_annot_col],
                  # Sample Column: alphanumeric characters and periods ONLY
                  "#SampleID" = stringr::str_replace_all(!!dplyr::sym("#SampleID"), c(
                    "-" = "\\.", "_" = "\\.", " " = "\\.")),
                  "#SampleID" = stringr::str_remove_all(!!dplyr::sym("#SampleID"),
                                                        "[^[:alnum:].]")) |>
    # Metadata columns: Only alphanumeric and [_.-+% ;:,/] characters
    apply(2, func) 
    alt_col <- ifelse(alt_col == " " | alt_col == "<NA>", NA, alt_col)
  # Remove unecessary columns
  ind <- colnames(alt_col) == which_annot_col
  alt_col2 <- alt_col[, -ind] |> dplyr::as_tibble() |>
    dplyr::mutate(BarcodeSequence = "-", LinkerPrimerSequence = "-",
                  Description = "-") |> dplyr::relocate(
                    "#SampleID", "BarcodeSequence", "LinkerPrimerSequence")
  # Write files
  out_map <- paste(path_to_write, "QIIME_metadata_map.tsv", sep = "/")
  message("Writing TSV mapping file to ", out_map)
  utils::write.table(alt_col2, out_map, sep = "\t",
                     row.names = FALSE, quote = FALSE)
  # Write counts table to biom
  colnames(counts_table_match) <- alt_col2$`#SampleID`
  biom_obj <- biomformat::make_biom(counts_table_match)
  out_biom <- paste(path_to_write, "QIIME_featuretable.biom", sep = "/")
  message("Writing QIIME counts biom file to ", out_biom)
  biomformat::write_biom(biom_obj, out_biom)
}

# Create MAE Object
create_MAE <- function(annot_path, which_annot_col,
                       counts_table, taxonomy_table, path_to_write,
                       qiime_biom_out) {
  annot_dat <- readr::read_csv(annot_path, show_col_types = FALSE)
  se_colData <- annot_dat |> # Only keep present samples in annotation data
    dplyr::mutate("sampcol" = unlist(annot_dat[, which_annot_col])) |>
    dplyr::filter(!!dplyr::sym("sampcol") %in% colnames(counts_table)) |>
    dplyr::select(-"sampcol") |> S4Vectors::DataFrame()
  rownames(se_colData) <- se_colData[, which_annot_col]
  se_mgx <- counts_table |> base::data.matrix() |>
    S4Vectors::SimpleList() |> magrittr::set_names("MGX")
  # Reorder colData according to se_mgx
  ind <- match(rownames(se_colData), colnames(se_mgx$MGX))
  se_colData <- se_colData[order(ind), ]
  se_rowData <- taxonomy_table |> base::data.frame() |>
    dplyr::mutate_all(as.character) |> S4Vectors::DataFrame()
  microbe_se <- SummarizedExperiment::SummarizedExperiment(
    assays = se_mgx, colData = se_colData, rowData = se_rowData)
  MAE <- MultiAssayExperiment::MultiAssayExperiment(
    experiments = S4Vectors::SimpleList(MicrobeGenetics = microbe_se),
    colData = se_colData
  )
  if (qiime_biom_out) create_qiime_biom(
    se_colData, taxonomy_table, which_annot_col, counts_table, path_to_write)
  return(MAE)
}

# Read in the MetaScope_id CSVs
read_in_id <- function(path_id_counts, end_string, which_annot_col) {
  name_file <- basename(path_id_counts)
  readr::read_csv(path_id_counts, show_col_types = FALSE) |>
    dplyr::filter(!is.na(!!dplyr::sym("TaxonomyID"))) |>
    dplyr::select("read_count", "TaxonomyID") |>
    dplyr::mutate(sample = stringr::str_remove(name_file, end_string))
}

# Get input taxon phylogeny
class_taxon <- function(taxon, accession_path) {
  taxon_ranks <- c("superkingdom", "kingdom", "phylum", "class", "order",
                   "family", "genus", "species", "strain")
  na_table <- data.frame(name = "Unknown", rank = taxon_ranks, id = 0)
  if (is.na(taxon)) return(na_table)

  classification_table <- taxonomizr::getTaxonomy(taxon,accession_path) |>
    as.data.frame()

  classification_table <- classification_table |>
    tidyr::pivot_longer(cols = colnames(classification_table), names_to = "rank", values_to = "name") |>
    dplyr::relocate(rank, .after = "name")
  classification_table$id <- c(rep(NA, nrow(classification_table) - 1), taxon)
  return(classification_table)
}

# Function written to account for accessions that weren't identified
# in metascope_id step
fill_in_missing <- function(combined_pre, accession_path) {
  NAs <- "NA;NA;NA;NA;NA;NA;NA;NA"
  na_ind <- combined_pre$TaxonomyID |>
    as.double() |> is.na() |> which() |> suppressWarnings()
  if(length(na_ind) > 0) {
    acc_list <- combined_pre$TaxonomyID[na_ind]
    result <- acc_list |>
      find_accessions(quiet = TRUE, accession_path = accession_path) |>
      stats::setNames(acc_list) |>
      plyr::aaply(1, function(x) x[1]) |> unname()
    if (NAs %in% result) {
      na_ind_2 <- which(result %in% NAs)
      result[na_ind_2] <- acc_list[na_ind_2]
    }
    combined_pre$TaxonomyID[na_ind] <- result
  }
  return(combined_pre)
}

#' Create a multi-assay experiment from MetaScope output for usage with
#' animalcules
#'
#' Upon completion of the MetaScope pipeline, users can analyze and visualize
#' abundances in their samples using the animalcules package. This function
#' allows interoperability of \code{metascope_id} output with both animalcules
#' and QIIME. After running this function, the user should save the returned MAE
#' to an RDS file using a function like \code{saveRDS} to upload the output into
#' the \code{animalcules} package.
#'
#' @param meta_counts A vector of filepaths to the counts ID CSVs output by
#'   \code{metascope_id()}.
#' @param annot_path The filepath to the CSV annotation file for the samples.
#' This CSV metadata/annotation file should contain at least two columns,
#' one with names of all samples WITHOUT the extension listed in \code{end_string},
#' e.g. for output file "sample_x76.metascope_id.csv", the column specified in
#' \code{which_annot_col} should contain the entry "sample_x76". Sample names
#' containing characters "_", "-", and "." are fine, however sample names
#' beginning with numbers should be renamed to have a prefix, e.g. "777897sample"
#' should be renamed to "X777897sample" for both the output file name and the
#' annotation name.
#' @param end_string The end string used at the end of the metascope_id files.
#'   Default is ".metascope_id.csv".
#' @param which_annot_col The name of the column of the annotation file
#'   containing the sample IDs. These should be the same as the
#'   \code{meta_counts} root filenames.
#' @param qiime_biom_out Would you also like a qiime-compatible biom file
#'   output? If yes, two files will be saved: one is a biom file of the counts
#'   table, and the other is a specifically formatted mapping file of metadata
#'   information. Default is \code{FALSE}.
#' @param path_to_write If \code{qiime_biom_out = TRUE}, where should output QIIME
#'   files be written? Should be a character string of the folder path. Default is
#'   '.', i.e. the current working directory.
#' @param accession_path (character) Path to taxonomizr accessions. See
#'   \code{taxonomizr::prepareDatabase()}.
#' @returns Returns a MultiAssay Experiment file of combined sample counts data
#'   and/or biom file and mapping file for analysis with QIIME. The MultiAssay
#'   Experiment will have a counts assay ("MGX").
#' @export
#' @importFrom rlang .data
#'
#' @examples
#' tempfolder <- tempfile()
#' dir.create(tempfolder)
#'
#' # Create three different samples
#' samp_names <- c("X123", "X456", "X789")
#' all_files <- file.path(tempfolder,
#'                        paste0(samp_names, ".csv"))
#'
#' create_IDcsv <- function (out_file) {
#'   final_taxids <- c("273036", "418127", "11234")
#'   final_genomes <- c(
#'     "Staphylococcus aureus RF122, complete sequence",
#'     "Staphylococcus aureus subsp. aureus Mu3, complete sequence",
#'     "Measles virus, complete genome")
#'   best_hit <- sample(seq(100, 1050), 3)
#'   proportion <- best_hit/sum(best_hit) |> round(2)
#'   EMreads <- best_hit + round(runif(3), 1)
#'   EMprop <- proportion + 0.003
#'   dplyr::tibble(TaxonomyID = final_taxids,
#'                 Genome = final_genomes,
#'                 read_count = best_hit, Proportion = proportion,
#'                 EMreads = EMreads, EMProportion = EMprop) |>
#'     dplyr::arrange(dplyr::desc(.data$read_count)) |>
#'     utils::write.csv(file = out_file, row.names = FALSE)
#'   message("Done!")
#'   return(out_file)
#' }
#' out_files <- vapply(all_files, create_IDcsv, FUN.VALUE = character(1))
#'
#' # Create annotation data for samples
#' annot_dat <- file.path(tempfolder, "annot.csv")
#' dplyr::tibble(Sample = samp_names, RSV = c("pos", "neg", "pos"),
#'               month = c("March", "July", "Aug"),
#'               yrsold = c(0.5, 0.6, 0.2)) |>
#'   utils::write.csv(file = annot_dat,
#'                    row.names = FALSE)
#'
#' # Create temporary taxonomizr accession
#' tmp_accession <- system.file("extdata", "example_accessions.sql", package = "MetaScope")
#'
#' # Convert samples to MAE
#' outMAE <- convert_animalcules(meta_counts = out_files,
#'                               annot_path = annot_dat,
#'                               which_annot_col = "Sample",
#'                               end_string = ".metascope_id.csv",
#'                               qiime_biom_out = FALSE,
#'                               accession_path = tmp_accession)
#'
#' unlink(tempfolder, recursive = TRUE)

convert_animalcules <- function(meta_counts, annot_path, which_annot_col,
                                end_string = ".metascope_id.csv",
                                qiime_biom_out = FALSE, path_to_write = ".",
                                accession_path = NULL) {
  combined_pre <- lapply(meta_counts, read_in_id, end_string = end_string,
                          which_annot_col = which_annot_col) |>
    data.table::rbindlist() |> dplyr::ungroup() |> as.data.frame() |>
    dplyr::mutate(sample = stringr::str_remove_all(sample, ".csv")) |>
    dplyr::select("read_count", "TaxonomyID", "sample") |>
    tidyr::pivot_wider(
      id_cols = "TaxonomyID", names_from = "sample",
      values_from = "read_count", values_fill = 0, id_expand = TRUE)
  # Which entries are not numeric, and try running them through genbank2uid
  combined_list <- fill_in_missing(combined_pre, accession_path) |>
    dplyr::mutate("TaxonomyID" = as.numeric(!!dplyr::sym("TaxonomyID")))
  # Create taxonomy, counts tables
  taxon_ranks <- c("superkingdom", "kingdom", "phylum", "class", "order",
                   "family", "genus", "species", "strain")
  message("Looking up taxon UIDs in NCBI database")
  all_ncbi <- plyr::llply(combined_list$TaxonomyID, .fun = class_taxon,
                          accession_path = accession_path)
  # fix unknowns
  na_ind <- which(is.na(all_ncbi))
  unk_tab <- data.frame(name = "unknown", rank = taxon_ranks, id = 0)
  if (length(na_ind) > 0) for (i in na_ind) all_ncbi[[i]] <- unk_tab
  # Create table
  taxonomy_table <- plyr::llply(all_ncbi, mk_table, taxon_ranks) |>
    dplyr::bind_rows() |> as.data.frame()
  colnames(taxonomy_table) <- taxon_ranks
  counts_table <- combined_list |> dplyr::select(-"TaxonomyID") |>
    as.data.frame()
  # Remove any brackets
  taxonomy_table$species <- gsub("\\[|\\]", "", taxonomy_table$species)
  # Remove duplicated species
  na_ind <- which(is.na(taxonomy_table$species))
  taxonomy_table$species[na_ind] <- paste0("g_", taxonomy_table$genus[na_ind])
  counts_tab_uniq <- animalcules::upsample_counts(counts_table, taxonomy_table,
                                                  higher_level = "species")
  taxonomy_table_uniq <- taxonomy_table |>
    dplyr::distinct(!!dplyr::sym("species"), .keep_all = TRUE) |>
    dplyr::filter(!!dplyr::sym("species") %in% rownames(counts_tab_uniq))
  
  order_ind <- match(rownames(counts_tab_uniq), taxonomy_table_uniq$species)
  taxonomy_table_f <- taxonomy_table_uniq[order_ind, ]
  # all.equal(taxonomy_table_f$species, rownames(counts_tab_uniq))
  rownames(taxonomy_table_f) <- stringr::str_replace_all(taxonomy_table_f$species, " ", "_")
  MAE <- create_MAE(
    annot_path, which_annot_col, counts_table,
    taxonomy_table, path_to_write, qiime_biom_out)
  return(MAE)
}

# Helper function for remove_matches
reduceByYield_RM <- function(X, YIELD, MAP, DONE, filter_names, num_reads) {
  if (!Rsamtools::isOpen(X)) {
    Rsamtools::open.BamFile(X)
    on.exit(Rsamtools::close.BamFile(X))
  }
  pb <- utils::txtProgressBar(min = 0,  max = num_reads, style = 3)
  k <- 0
  repeat {
    if (DONE(data <- YIELD(X))) break
    k <- k + nrow(data)
    MAP(data, filter_names)
    utils::setTxtProgressBar(pb, k)
  }
  base::close(pb)
}

# Helper function for mk_interim_fastq()
reduceByYield_iterate <- function(X, YIELD, MAP, REDUCE, init) {
  if (!Rsamtools::isOpen(X)) {
    Rsamtools::open.BamFile(X)
    on.exit(Rsamtools::close.BamFile(X))
  }
  # Result must be a data frame
  result <- init
  pb <- utils::txtProgressBar(min = 0,
                              max = mx <- nrow(result),
                              style = 3)
  repeat {
    if (nrow(result) == 0)
      break
    data <- YIELD(X)
    value <- MAP(data)
    result <- REDUCE(result, value)
    utils::setTxtProgressBar(pb, mx - nrow(result))
  }
  close(pb)
}

# Helper function for filter_host_bowtie to write interim fastq file
mk_interim_fastq <- function(reads_bam, read_loc, YS, quiet) {
  unlink(read_loc, recursive = FALSE)
  if (!quiet) message("Writing Intermediate FASTQ file to ", read_loc)
  # Pull out all query names
  bf_init <- Rsamtools::BamFile(reads_bam, yieldSize = 100000000)
  allqname <-
    Rsamtools::scanBam(bf_init, param = Rsamtools::ScanBamParam(
      what = "qname"))[[1]]$qname %>% unique()
  init_df <- dplyr::tibble("Header" = unname(unlist(allqname))) %>%
    dplyr::distinct("Header", .keep_all = TRUE)
  # Define BAM file with smaller yield size
  bf <- Rsamtools::BamFile(reads_bam, yieldSize = YS)
  # Init file connection
  gz_con <- gzfile(read_loc, "a")
  # Define functions
  YIELD <- function(bf) {
    to_pull <- c("qname", "qual", "seq")
    value <- Rsamtools::scanBam(bf, param = Rsamtools::ScanBamParam(
      what = to_pull))[[1]]
    return(value)
  }
  MAP <- function(value) {
    y <- dplyr::tibble("Header" = value$qname |> unlist(),
                       "Sequence" = as.character(value$seq),
                       "Quality" = as.character(value$qual)) %>%
      dplyr::distinct(.data$Header, .keep_all = TRUE)
    return(y)
  }
  REDUCE <- function(x, y) {
    all_join <- dplyr::left_join(x, y, by = "Header")
    # Print all current reads to file
    all_join %>% dplyr::filter(!is.na(.data$Sequence)) %>%
      dplyr::mutate("Plus" = "+",
                    "Header" = stringr::str_c("@", .data$Header)) %>%
      dplyr::select("Header", "Sequence", "Plus", "Quality") %>% 
      as.matrix() %>% t() %>%
      as.character() %>%
      writeLines(con = gz_con)
    empty_vals <- all_join %>% dplyr::filter(is.na(.data$Sequence)) %>%
      dplyr::select(.data$Header)
    return(empty_vals)
  }
  reduceByYield_iterate(bf, YIELD, MAP, REDUCE, init = init_df)
  close(gz_con)
  if (!quiet) message("Intermediate FASTQ file written to", read_loc)
}

#' Helper function to remove reads matched to filter libraries
#'
#' Using the \code{filter_host()} function, we align our sequencing sample to
#' all filter libraries of interest. The \code{remove_matches()} function allows
#' for removal of any target reads that are also aligned to filter libraries.
#'
#' This function is not intended for direct use.
#'
#' @param reads_bam The name of a merged, sorted .bam file that has previously
#'   been aligned to a reference library. Likely, the output from running an
#'   instance of \code{align_target()}.
#' @param read_names A \code{list} of target query names from \code{reads_bam}
#'   that have also aligned to a filter reference library. Each \code{list}
#'   element should be a vector of read names.
#' @param output The name of the .bam or .csv.gz file that to which the filtered
#'   alignments will be written.
#' @inheritParams filter_host_bowtie
#' @inheritParams filter_host
#' @param threads The number of threads to be used in filtering the bam file.
#'   Default is 1.
#' @param aligner The aligner which was used to create the bam file.
#'
#' @return Depending on input \code{make_bam}, either the name of a filtered,
#'   sorted .bam file written to the user's current working directory, or an RDS
#'   file containing a data frame of only requisite information to run
#'   \code{metascope_id()}.
#'

remove_matches <- function(reads_bam, read_names, output, YS, threads,
                           aligner, make_bam, quiet) {
  if (!quiet) message("Removing reads mapped to host indices")
  # some aligned reads may be duplicated; remove these, and unlist
  filter_names <- sort(unique(unlist(read_names)))
  if (make_bam) {
    name_out <- paste0(output, ".bam")
    if (file.exists(name_out)) file.remove(name_out)
    # obtain vector of target query names from .bam file
    target_reads <- Rsamtools::scanBam(reads_bam)[[1]]$qname
    # define logical vector of which reads to keep, based on query names
    filter_which <- !(target_reads %in% filter_names)
    # index target bam file (reads_BAM already sorted by chromosome)
    bam_index <- Rsamtools::indexBam(reads_bam)
    # create BamFile instance to set yieldSize
    bf <- Rsamtools::BamFile(reads_bam, yieldSize = length(filter_which))
    Rsamtools::filterBam(
      bf, destination = name_out, index = bam_index,
      indexDestination = FALSE, filter = filter_which,
      param = Rsamtools::ScanBamParam(what = "qname"))
    # file.remove(bam_index) keep bam index
  } else if (!make_bam) {
    name_out <- paste0(output, ".csv.gz")
    numread <- Rsamtools::BamFile(reads_bam, yieldSize = 100000000) %>%
      Rsamtools::scanBam(param = Rsamtools::ScanBamParam(
        what = "pos")) %>% magrittr::extract2(1) %>% unlist() %>% length()
    # Define BAM file with smaller yield size
    bf <- Rsamtools::BamFile(reads_bam, yieldSize = YS)
    # Initialize connection for output file
    if(file.exists(name_out)) file.remove(name_out)
    con_gz <- gzfile(name_out, open = "a")
    YIELD <- function(X) {
      to_pull <- c("qname", "rname", "cigar", "qwidth", "pos")
      if (identical(aligner, "bowtie2")) {
        out <- Rsamtools::scanBam(X, param = Rsamtools::ScanBamParam(
          what = to_pull, tag = c("AS")))[[1]]
        out[to_pull] %>% dplyr::as_tibble() %>%
          dplyr::mutate(tag = out$tag$AS) %>% return()
      } else if (identical(aligner, "subread")) {
        out <- Rsamtools::scanBam(X, param = Rsamtools::ScanBamParam(
          what = to_pull, tag = c("NM")))[[1]]
        out[to_pull] %>% dplyr::as_tibble() %>%
          dplyr::mutate(tag = out$tag$NM) %>% return()
      }
    }
    MAP <- function(value, filter_names) {
      value %>%
        dplyr::filter(!(.data$qname %in% filter_names)) %>%
        apply(1, function(row) paste(row, collapse = ",")) %>%
        writeLines(con_gz)
      return("DONE!")
    }
    DONE <- function(data) nrow(data) == 0
    reduceByYield_RM(bf, YIELD, MAP, DONE, filter_names, numread)
    close(con_gz)
  }
  if (!quiet) message("DONE! Alignments written to ", name_out)
  return(name_out)
}

#' Use Rsubread to align reads against one or more filter libraries and
#' subsequently remove mapped reads
#'
#' After aligning your sample to a target library with \code{align_target()},
#' use \code{filter_host()} to remove unwelcome host contamination using filter
#' reference libraries. This function takes as input the name of the .bam file
#' produced via \code{align_target()}, and produces a sorted .bam file with any
#' reads that match the filter libraries removed. This resulting .bam file may
#' be used upstream for further analysis. This function uses Rsubread. For the
#' Rbowtie2 equivalent of this function, see \code{filter_host_bowtie}.
#'
#' A compressed .csv can be created to produce a smaller output file that is
#' created more efficiently and is still compatible with \code{metascope_id()}.
#'
#' @param reads_bam The name of a merged, sorted .bam file that has previously
#'   been aligned to a reference library. Likely, the output from running an
#'   instance of \code{align_target()}.
#' @param lib_dir Path to the directory that contains the filter Subread index
#'   files.
#' @param libs The basename of the filter libraries (without index
#'   extension).
#' @inheritParams filter_host_bowtie
#' @inheritParams align_target
#' @return The name of a filtered, sorted .bam file written to the user's
#'   current working directory. Or, if \code{make_bam = FALSE}, a .csv.gz file
#'   containing a data frame of only requisite information to run
#'   \code{metascope_id()}.
#'
#' @export
#'
#' @examples
#' #### Filter reads from bam file that align to any of the filter libraries
#'
#' ## Assuming a bam file has been created previously with align_target()
#' \donttest{
#' ## Create temporary directory
#' filter_ref_temp <- tempfile()
#' dir.create(filter_ref_temp)
#'
#' ## Create temporary taxonomizr accession
#' tmp_accession <- system.file("extdata", "example_accessions.sql", package = "MetaScope")
#'
#'
#' ## Download filter genome
#' all_species <- c("Staphylococcus aureus subsp. aureus str. Newman")
#' all_ref <- vapply(all_species, MetaScope::download_refseq,
#'                   reference = FALSE, representative = FALSE, compress = TRUE,
#'                   out_dir = filter_ref_temp, caching = FALSE,
#'                   accession_path = tmp_accession,
#'                   FUN.VALUE = character(1))
#' ind_out <- vapply(all_ref, mk_subread_index, FUN.VALUE = character(1))
#'
#' ## Get path to example reads
#' readPath <- system.file("extdata", "subread_target.bam",
#'                         package = "MetaScope")
#'
#' ## Copy the example reads to the temp directory
#' refPath <- file.path(filter_ref_temp, "subread_target.bam")
#' file.copy(from = readPath, to = refPath)
#' utils::data("align_details")
#' align_details[["type"]] <- "rna"
#' align_details[["phredOffset"]] <- 10
#' # Just to get it to align - toy example!
#' align_details[["maxMismatches"]] <- 10
#'
#' ## Align and filter reads
#' filtered_map <- filter_host(
#'   refPath, lib_dir = filter_ref_temp,
#'   libs = stringr::str_replace_all(all_species, " ", "_"),
#'   threads = 1, subread_options = align_details)
#'
#' ## Remove temporary directory
#' unlink(filter_ref_temp, recursive = TRUE)
#' }

filter_host <- function(reads_bam, lib_dir = NULL, libs, make_bam = FALSE,
                        output = paste(tools::file_path_sans_ext(reads_bam),
                                       "filtered", sep = "."),
                        subread_options = align_details, YS = 100000,
                        threads = 1, quiet = TRUE) {
  data_env <- new.env(parent = emptyenv())
  utils::data("align_details", envir = data_env, package = "MetaScope")
  align_details <- data_env[["align_details"]]
  # Convert reads_bam into fastq
  read_loc <- file.path(dirname(output), "intermediate.fastq.gz")
  mk_interim_fastq(reads_bam, read_loc, YS, quiet = quiet)
  # Align
  read_names <- vector(mode = "list", length(libs))
  for (i in seq_along(libs)) {
    # Create output file name for BAM
    lib_file <- paste(tools::file_path_sans_ext(reads_bam), ".",
                      libs[i], ".bam", sep = "")
    # Align BAM to the lib & generate new file
    Rsubread::align(index = file.path(lib_dir, libs[i]),
                    readfile1 = read_loc,
                    input_format = "fastq",
                    output_file = lib_file,
                    type = subread_options[["type"]],
                    nthreads = threads,
                    maxMismatches = subread_options[["maxMismatches"]],
                    nsubreads = subread_options[["nsubreads"]],
                    phredOffset = subread_options[["phredOffset"]],
                    unique = subread_options[["unique"]],
                    nBestLocations = subread_options[["nBestLocations"]])
    # Extract target query names from mapped BAM file
    read_names[[i]] <- Rsamtools::scanBam(
      lib_file, param = Rsamtools::ScanBamParam(
        what = c("qname"),
        flag = Rsamtools::scanBamFlag(isUnmappedQuery = FALSE)))
    # Throw away BAM, vcf file
    file.remove(lib_file)
    file.remove(paste(lib_file, ".indel.vcf", sep = ""))
    file.remove(paste(lib_file, ".summary", sep = ""))
  }
  # remove intermediate fastq file
  file.remove(read_loc)
  # Filter out host-aligned reads
  name_out <- remove_matches(reads_bam, read_names, output, YS, threads,
                             "subread", make_bam, quiet = quiet)
  return(name_out)
}

#' Use Rbowtie2 to align reads against one or more filter libraries and
#' subsequently remove mapped reads
#'
#' After a sample is aligned to a target library with
#' \code{align_target_bowtie()}, we may use \code{filter_host_bowtie()} to
#' remove unwelcome host contamination using filter reference libraries. This
#' function takes as input the name of the .bam file produced via
#' \code{align_target_bowtie()}, and produces a sorted .bam or .csv.gz file with
#' any reads that match the filter libraries removed. This resulting .bam file
#' may be used downstream for further analysis. This function uses Rbowtie2 For
#' the Rsubread equivalent of this function, see \code{filter_host}.
#'
#' A compressed .csv can be created to produce a smaller output file that is
#' created more efficiently and is still compatible with \code{metascope_id()}.
#'
#' The default parameters are the same that PathoScope 2.0 uses.
#' "--very-sensitive-local -k 100 --score-min L,20,1.0"
#'
#' @param reads_bam The name of a merged, sorted .bam file that has previously
#'   been aligned to a reference library. Likely, the output from running an
#'   instance of \code{align_target_bowtie()}.
#' @param lib_dir Path to the directory that contains the filter Bowtie2 index
#'   files.
#' @param libs The basename of the filter libraries (without .bt2 or .bt2l
#'   extension).
#' @param make_bam Logical, whether to also output a bam file with host reads
#'   filtered out. A .csv.gz file will be created instead if \code{FALSE}.
#'   Creating a bam file is costly on resources over creating a compressed
#'   csv file with only relevant information, so default is \code{FALSE}.
#' @param output The desired name of the output .bam or .csv.gz file. Extension is
#'   automatically defined by whether \code{make_bam = TRUE.} Default is the
#'   basename of \code{unfiltered_bam} + \code{.filtered} + extension.
#' @param bowtie2_options Optional: Additional parameters that can be passed to
#'   the filter_host_bowtie() function. To see all the available parameters use
#'   Rbowtie2::bowtie2_usage(). See Details for default parameters.
#'   NOTE: Users should pass all their parameters as
#'   one string and if optional parameters are given then the user is
#'   responsible for entering all the parameters to be used by Bowtie2.
#'   The only parameters that should NOT be specified here is the threads.
#' @param YS yieldSize, an integer. The number of alignments to be read in from
#'   the bam file at once for chunked functions. Default is 100000.
#' @param threads The amount of threads available for the function. Default is 1
#'   thread.
#' @param overwrite Whether existing files should be overwritten. Default is
#'   FALSE.
#' @param quiet Turns off most messages. Default is \code{TRUE}.
#'
#' @return The name of a filtered, sorted .bam file written to the user's
#'   current working directory. Or, if \code{make_bam = FALSE}, a .csv.gz file
#'   containing a data frame of only requisite information to run
#'   \code{metascope_id()}.
#'
#' @export
#'
#' @examples
#' #### Filter reads from bam file that align to any of the filter libraries
#'
#' ## Assuming a bam file has already been created with align_target_bowtie()
#' # Create temporary filter library
#' filter_ref_temp <- tempfile()
#' dir.create(filter_ref_temp)
#'
#' ## Create temporary taxonomizr accession
#' tmp_accession <- system.file("extdata", "example_accessions.sql", package = "MetaScope")
#'
#' ## Download reference genome
#' MetaScope::download_refseq("Orthoebolavirus zairense",
#'                            reference = FALSE,
#'                            representative = FALSE,
#'                            compress = TRUE,
#'                            out_dir = filter_ref_temp,
#'                            caching = TRUE,
#'                            accession_path = tmp_accession)
#'
#' ## Create temp directory to store the indices
#' index_temp <- tempfile()
#' dir.create(index_temp)
#'
#' ## Create filter index
#' MetaScope::mk_bowtie_index(
#'   ref_dir = filter_ref_temp,
#'   lib_dir = index_temp,
#'   lib_name = "filter",
#'   overwrite = TRUE
#' )
#'
#' ## Create temporary folder to hold final output file
#' output_temp <- tempfile()
#' dir.create(output_temp)
#'
#' ## Get path to example bam
#' bamPath <- system.file("extdata", "bowtie_target.bam",
#'                        package = "MetaScope")
#' target_copied <- file.path(output_temp, "bowtie_target.bam")
#' file.copy(bamPath, target_copied)
#'
#' ## Align and filter reads
#' filter_out <-
#'   filter_host_bowtie(
#'     reads_bam = target_copied,
#'     lib_dir = index_temp,
#'     libs = "filter",
#'     threads = 1
#'   )
#'
#' ## Remove temporary directories
#' unlink(filter_ref_temp, recursive = TRUE)
#' unlink(index_temp, recursive = TRUE)
#' unlink(output_temp, recursive = TRUE)

filter_host_bowtie <- function(reads_bam, lib_dir, libs, make_bam = FALSE,
                               output = paste(
                                 tools::file_path_sans_ext(reads_bam),
                                 "filtered", sep = "."),
                               bowtie2_options = NULL, YS = 100000,
                               threads = 1, overwrite = FALSE, quiet = TRUE) {
  # If user does not specify parameters, specify for them
  if (is.null(bowtie2_options)) {
    bowtie2_options <- paste("--local -R 2 -N 0 -L 25 -i S,1,0.75 -k 5 --score-min L,0,1.7",
                             "--threads", threads)
  } else bowtie2_options <- paste(bowtie2_options, "--threads", threads)
  # Convert reads_bam into fastq
  read_loc <- file.path(dirname(output), "intermediate.fastq.gz")
  mk_interim_fastq(reads_bam, read_loc, YS, quiet = quiet)
  # Init list of names
  read_names <- vector(mode = "list", length(libs))
  for (i in seq_along(libs)) {
    # Create output file name for BAM
    lib_file <- paste(tools::file_path_sans_ext(reads_bam), ".",
                      libs[i], ".bam", sep = "")
    if (!quiet) message("Attempting to perform Bowtie2 alignment on ",
                        libs[i], " index")
    # Align reads to lib and generate new filter BAM file
    Rbowtie2::bowtie2_samtools(
      bt2Index = file.path(lib_dir, libs[i]),
      output = tools::file_path_sans_ext(lib_file),
      outputType = "bam", seq1 = read_loc, ... = bowtie2_options,
      overwrite = overwrite)
    # Extract target query names from mapped BAM file
    read_names[[i]] <- Rsamtools::scanBam(
      lib_file, param = Rsamtools::ScanBamParam(
        what = c("qname"),
        flag = Rsamtools::scanBamFlag(isUnmappedQuery = FALSE)))
    # Throw away BAM file
    file.remove(lib_file)
    unlink(".bowtie2.cerr.txt")
  }
  # remove intermediate fastq file
  file.remove(read_loc)
  # Filter out host-aligned reads
  name_out <- remove_matches(reads_bam, read_names, output, YS, threads,
                             "bowtie2", make_bam, quiet = quiet)
  return(name_out)
}

#' Run the MetaScope workflow on a directory of FASTQ files
#'
#' This wrapper function streamlines the MetaScope pipeline by automatically
#' downloading references, building a Bowtie2 index, aligning reads, and running
#' \code{metascope_id()} for taxonomic identification. It supports both
#' single-end and paired-end sequencing data and is currently implemented only
#' for 16S rRNA sequencing.
#'
#' @param fastq_dir Path to the directory containing FASTQ (or FASTQ.GZ) files.
#'   For paired-end runs, files must follow the \code{_R1}/\code{_R2}
#'   naming convention.
#' @param sequencing_type Character string specifying the sequencing type.
#'   Currently only \code{"16S"} is supported.
#' @param ref_dir Path to the reference directory where the 16S RefSeq FASTA,
#'   Bowtie2 indices, and accession database are stored. Defaults to
#'   \code{"~/MetaScope/ref_dir"}.
#' @param paired_reads Logical. If \code{FALSE} (default), runs in single-end
#'   mode; if \code{TRUE}, assumes paired-end FASTQ files with matching
#'   \code{_R1} and \code{_R2} filenames.
#' @param tmp_dir Path to a directory for temporary alignment files (SAM/BAM and
#'   intermediate results). Defaults to \code{"~/MetaScope/tmp"}.
#' @param out_dir Path to the directory for final MetaScope output files.
#'   Defaults to \code{"~/MetaScope/out"}.
#' @param threads Integer. Number of CPU threads to use for Bowtie2 alignment
#'   and indexing. Default is \code{1}.
#' @param combine_results Boolean to merge all samples into one feature table
#' @param ... Additional arguments passed to \code{metascope_id()}.
#'
#' @details
#' The workflow proceeds through the following steps:
#' \enumerate{
#'   \item Ensures the NCBI 16S RefSeq FASTA file is downloaded
#'         (via \code{download_refseq_16S()}).
#'   \item Builds a Bowtie2 index of the 16S reference if it does not exist
#'         (via \code{mk_bowtie_index()}).
#'   \item Ensures that the NCBI accession-to-taxonomy database is present
#'         (via \code{download_accessions()}).
#'   \item Aligns all FASTQ files in \code{fastq_dir} against the reference
#'         (via \code{align_target_bowtie()}).
#'   \item Runs \code{metascope_id()} on each alignment BAM file and writes
#'         annotated taxonomic identification results to \code{out_dir}.
#' }
#'
#' @return A named list of output file paths generated by \code{metascope_id()},
#'   with sample names as list element names. Results are also written to the
#'   specified \code{out_dir}.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' # Run MetaScope on single-end FASTQ files
#' run_metascope(fastq_dir = "data/fastq", threads = 4)
#'
#' # Run MetaScope on paired-end FASTQ files
#' run_metascope(fastq_dir = "data/fastq", paired_reads = TRUE, threads = 8)
#' }

run_metascope <- function(
    fastq_dir,
    sequencing_type = "16S",
    ref_dir = "~/MetaScope/ref_dir",
    paired_reads = FALSE, 
    tmp_dir = "~/MetaScope/tmp", 
    out_dir = "~/MetaScope/out",
    threads = 1,
    combine_results = TRUE,
    ...) {
  
  # Parse ... arguments
  extra_args <- list(...)
  # Currently only works for 16S
  if (sequencing_type != "16S") {
    stop("Currently only 16S sequencing is supported.")
  } 
  
  # Create Directories if they don't already exist
  dir.create(ref_dir, showWarnings = FALSE, recursive = TRUE)
  dir.create(tmp_dir, showWarnings = FALSE, recursive = TRUE)
  dir.create(out_dir, showWarnings = FALSE, recursive = TRUE)
  
  # Run download_refseq_16S if it is not already downloaded (Currently assumes default refseq name) 
  ref_fna <- file.path(ref_dir, "refseq_16S.fna")
  if (!file.exists(ref_fna)) {
    message("16S Refseq Not Found, Dowloading from NCBI FTP")
    download_refseq_16S(out_dir = ref_dir, combined_name = "refseq_16S.fna")
  } 
  else {
    message("16S Refseq Found in: ", ref_fna)
  }
  
  # Run mk_bowtie_index if the bowtie index is not already made (Currently only uses bowtie)
  bt2_files <- file.path(ref_dir, c("16S_ribosomal_RNA.1.bt2",
                                    "16S_ribosomal_RNA.2.bt2",
                                    "16S_ribosomal_RNA.3.bt2",
                                    "16S_ribosomal_RNA.4.bt2",
                                    "16S_ribosomal_RNA.rev.1.bt2",
                                    "16S_ribosomal_RNA.rev.2.bt2"))
  
  if (!all(file.exists(bt2_files))) { 
    message("16S Ribosomal RNA bowtie Indices Not Found. Building Bowtie Indices:")
    extra_args_mk_bowtie_index <- extra_args[names(extra_args) %in% names(formals(mk_bowtie_index))]
    do.call(mk_bowtie_index, c(list(ref_dir = ref_dir, 
                                    lib_dir = ref_dir, 
                                    lib_name = "16S_ribosomal_RNA"), 
                               extra_args_mk_bowtie_index))
  } else {
    message("16S Ribosoaml Index Found in: ", ref_dir)
  }
  
  # Download Taxonomy Accessions
  if (!(file.exists(file.path(ref_dir, "accessionTaxa.sql")))) {
    message("Taxonomy Accessions Database Not Found. Downloading from NCBI FTP. This may take a while")
    download_accessions(ref_dir,
                        tmp_dir = file.path(ref_dir, "tmp"),
                        remove_tmp_dir = TRUE,
                        NCBI_accessions_database = TRUE,
                        NCBI_accessions_name = "accessionTaxa",
                        silva_taxonomy_database = FALSE,
                        silva_taxonomy_name = "all_silva_headers",
                        blast_16S_database = FALSE,
                        blast_16S_name = "16S_ribosomal_RNA")
  }
  else {
    message("Accessions Database found in: ", file.path(ref_dir, "accessionTaxa.sql"))
  }
  
  # Run align targets for all FASTQ files in input directory
  ## Collect FASTQ Files
  fq_files <- list.files(fastq_dir, full.names = TRUE)
  if (length(fq_files) == 0) stop("No FASTQ files found in: ", fastq_dir)
  
  ## Derive Sample Names from FASTQ file names
  sample_name <- function(fq) {
    gsub("\\.fastq(.gz)?$|\\.fq(.gz)?$", "", basename(fq))
  }
  
  ## Single End Reads
  if (!paired_reads) {
    message("Running single-end alignment and MetaScopeID...")
    results <- lapply(fq_files, function(fq) {
      tmp_out <- file.path(tmp_dir, paste0(sample_name(fq)))
      dir.create(tmp_out, showWarnings = FALSE) 
      extra_args_align_target_bowtie <- extra_args[names(extra_args) %in% names(formals(align_target_bowtie))]
      bam_out <- do.call(align_target_bowtie, c(list(read1 = fq, 
                                                     read2 = NULL, 
                                                     lib_dir = ref_dir, 
                                                     libs = "16S_ribosomal_RNA",
                                                     align_dir = tmp_out,
                                                     align_file = sample_name(fq), 
                                                     threads = threads), 
                                                 extra_args_align_target_bowtie))
      extra_args_metascope_id <- extra_args[names(extra_args) %in% names(formals(metascope_id))]
      id_out <- do.call(metascope_id, c(list(input_file = bam_out, 
                                             input_type = "bam",
                                             aligner = "bowtie2",
                                             db = "ncbi",
                                             accession_path = file.path(ref_dir, "accessionTaxa.sql"),
                                             tmp_dir = tmp_out,
                                             out_dir = out_dir), 
                                        extra_args_metascope_id))
    })
    
    names(results) <- basename(fq_files) |>
      sub("\\.fastq(.gz)?$|\\.fq(.gz)?$", "", x = _)
    
  } else {
    message("Running paired-end alignment and MetaScopeID...")
    # assumes R1/R2 naming convention
    fq_R1 <- fq_files[grepl("_R1", fq_files)]
    fq_R2 <- fq_files[grepl("_R2", fq_files)]
    if (length(fq_R1) != length(fq_R2)) stop("Mismatched R1/R2 files")
    
    results <- mapply(function(r1, r2) {
      sample_base_name <- sub("_R1.*", "", basename(r1))
      tmp_out <- file.path(tmp_dir, sample_base_name)
      dir.create(tmp_out, showWarnings = FALSE) 
      extra_args_align_target_bowtie <- extra_args[names(extra_args) %in% names(formals(align_target_bowtie))]
      bam_out <- do.call(align_target_bowtie, c(list(read1 = r1, 
                                                     read2 = r2, 
                                                     lib_dir = ref_dir, 
                                                     libs = "16S_ribosomal_RNA",
                                                     align_dir = tmp_out,
                                                     align_file = sample_base_name, 
                                                     threads = threads), 
                                                 extra_args_align_target_bowtie))
      
      extra_args_metascope_id <- extra_args[names(extra_args) %in% names(formals(metascope_id))]
      id_out <- do.call(metascope_id, c(list(input_file = bam_out, 
                                             input_type = "bam",
                                             aligner = "bowtie2",
                                             db = "ncbi",
                                             accession_path = file.path(ref_dir, "accessionTaxa.sql"),
                                             tmp_dir = tmp_out,
                                             out_dir = out_dir), 
                                        extra_args_metascope_id))
    }, fq_R1, fq_R2)
  names(results) <- sub("_R1.*", "", basename(fq_R1))
  }
  if (!combine_results) {
    return(results)
  }
  else {
    if (length(out_dir) == 0) stop("No MetaScope ID outputs found in: ", out_dir)
    # Read in all results
    all_res <- results |>
      purrr::map(function(path) {
        sample_name = sub(".metascope_id.csv", "", basename(path))
        utils::read.csv(path) |>
          dplyr::select(!!dplyr::sym("TaxonomyID"), !!dplyr::sym("Genome"), !!dplyr::sym("readsEM")) |>
          dplyr::rename_with(~ sample_name, .cols = !!dplyr::sym("readsEM"))
      })
    # Merge together and rename column
    merged_res <- purrr::reduce(all_res, dplyr::full_join, by = c("TaxonomyID", "Genome")) |>
      dplyr::mutate(
        dplyr::across(
          -dplyr::all_of(c("TaxonomyID", "Genome")),
          ~ tidyr::replace_na(.x, 0)
        )
      )
    
    utils::write.csv(merged_res, file = file.path(out_dir, "metascope_results.csv"))
    return(merged_res)
  }
}

#' Inspect a Bowtie index
#'
#' This function can be use to call the \code{bowtie2-inspect} wrapper which
#' wraps the \code{bowtie2-inspect-s} and \code{bowtie2-inspect-l} binaries.
#'
#' @param bt2_base The path to the Bowtie2 index files minus minus trailing
#' .1.bt2/.2.bt2 extension names.
#' @param ... Required additional arguments to be passed on to the \code{bowtie2-inspect}
#' binaries. See below for details.
#'
#' @details All additional arguments in ... are intepreted as
#' parameters to be passed on to \code{bowtie2-inspect} wrapper. All of them
#' \code{Character} or \code{Numeric} scalar. You can put all additional
#' arguments in one \code{Character} (e.g. "--across 60 --names") with white
#' space separation, or put them in different \code{Character}
#' (e.g. "--across","60","--names"). See the output of
#' \code{metascope_inspect_usage()} for details about available parameters.
#'
#' @references
#' {
#' Langmead B, Wilks C, Antonescu V, Charles R. Scaling
#' read aligners to hundreds of threads on general-purpose processors.
#' Bioinformatics. 2018 Jul 18. doi: 10.1093/bioinformatics/bty648.
#'
#' Langmead B, Salzberg SL. Fast gapped-read alignment with
#' Bowtie 2. Nature Methods. 2012 Mar 4;9(4):357-9. doi: 10.1038/nmeth.1923.
#' }
#' @export metascope_inspect
#' @examples
#' \dontrun{
#' metascope_inspect(bt2_base)
#' }



metascope_inspect <- function(bt2_base, ...) {

    # Handle additional arguments
    arg_options <- c(...)
    if (!is.null(arg_options)) {
        arg_options <- trimws(arg_options)
        arg_options <- arg_options[nzchar(arg_options)]
        if (length(arg_options) == 0) return(NULL)
    }


    # Combine arguments together
    bt2_base <- shQuote(bt2_base)
    args <- c(arg_options, bt2_base)

    # Call bowtie2-inspect wrapper
    script_path <- system.file("bowtie2-inspect", package = "MetaScope")
    if (script_path == "" || !file.exists(script_path)) {
        stop("bowtie2-inspect script not found in package")
    }

    tryCatch({
        cmd <- paste("python3", shQuote(script_path), paste(args, collapse = " "))
        system(cmd)
    },
    error = function(e) {
        cmd <- paste("python", shQuote(script_path), paste(args, collapse = " "))
        system(cmd)
    })


}

#' @name metascope_inspect_usage
#' @title Print available arguments that can be passed to metascope_inspect()
#' @description Calling metascope_inspect_usage() prints the available arguments
#' that can be passed to the ... argument of the metascope_inspect() function of
#' the package. Note that some arguments are invalid if they are already
#' handled as explicit function arguments.
#' @return Information about available arguments that can be passed to
#' metascope_inspect()
#' @references
#' {
#' Langmead B, Wilks C, Antonescu V, Charles R. Scaling
#' read aligners to hundreds of threads on general-purpose processors.
#' Bioinformatics. 2018 Jul 18. doi: 10.1093/bioinformatics/bty648.
#'
#' Langmead B, Salzberg SL. Fast gapped-read alignment with
#' Bowtie 2. Nature Methods. 2012 Mar 4;9(4):357-9. doi: 10.1038/nmeth.1923.
#' }
#' @export metascope_inspect_usage
#' @examples
#' metascope_inspect_usage()


metascope_inspect_usage <- function() {
    if(R.Version()$arch=="i386"){
        return("bowtie2 is not available for 32bit, please use 64bit R instead")
    }

    script_path <- system.file("bowtie2-inspect", package = "MetaScope")
    if (script_path == "" || !file.exists(script_path)) {
        stop("bowtie2-inspect script not found in package")
    }

    tryCatch({
        message("Calling Bowtie2-inspect usage:")
        cmd <- paste("python3", shQuote(script_path), "-h")
        system(cmd)
    },
    error = function(e) {
        message("Calling Bowtie2-inspect usage:")
        cmd <- paste("python", shQuote(script_path), "-h")
        system(cmd)
    })
}

## ----show_install, eval = FALSE-----------------------------------------------
# if (!requireNamespace("BiocManager", quietly = TRUE))
#   install.packages("BiocManager")
# BiocManager::install("MetaScope")

## ----load_packages, eval = TRUE-----------------------------------------------
suppressPackageStartupMessages({
  library(MetaScope)
  library(magrittr)
})

## ----meta_demultiplex, message = FALSE----------------------------------------
# Get barcode, index, and read data locations
barcodePath <-
  system.file("extdata", "barcodes.txt", package = "MetaScope")
indexPath <- system.file("extdata", "virus_example_index.fastq",
                         package = "MetaScope")
readPath <-
  system.file("extdata", "virus_example.fastq", package = "MetaScope")

# Get barcode, index, and read data locations
demult <-
  meta_demultiplex(barcodePath,
                   indexPath,
                   readPath,
                   rcBarcodes = FALSE,
                   hammingDist = 2,
                   location = tempfile())
demult

## ----eval = FALSE-------------------------------------------------------------
# download_accessions(
#   ind_dir = "C:/Users/JohnSmith/ResearchIndices",
#   tmp_dir = file_path(ind_dir, "tmp"),
#   remove_tmp_dir = TRUE,
#   # For NCBI RefSeq or nucleotide database
#   NCBI_accessions_database = TRUE,
#   NCBI_accessions_name = "accessionTaxa.sql",
#   # For SILVA 16S database
#   silva_taxonomy_database = TRUE,
#   silva_taxonomy_name = "all_silva_headers.rds"
# )

## ----taxonomy_db, eval = TRUE, warning = FALSE, message = FALSE---------------
tmp_accession <- system.file("extdata","example_accessions.sql", package = "MetaScope")

## ----target_lib, eval = TRUE, warning = FALSE, message = FALSE----------------
target_ref_temp <- tempfile()
dir.create(target_ref_temp)

all_species <- c("Staphylococcus aureus subsp. aureus Mu50",
                 "Staphylococcus aureus subsp. aureus Mu3",
                 "Staphylococcus aureus subsp. aureus str. Newman",
                 "Staphylococcus aureus subsp. aureus N315",
                 "Staphylococcus aureus RF122", 
                 "Staphylococcus aureus subsp. aureus ST398")
sapply(all_species, download_refseq, 
       reference = FALSE, representative = FALSE, compress = TRUE,
       out_dir = target_ref_temp, caching = TRUE, accession_path = tmp_accession)

## ----filter_lib, warning = FALSE, message = FALSE-----------------------------
filter_ref_temp <- tempfile()
dir.create(filter_ref_temp)

download_refseq(
  taxon = "Staphylococcus epidermidis RP62A",
  representative = FALSE, reference = FALSE,
  compress = TRUE, out_dir = filter_ref_temp,
  caching = TRUE,
  accession_path = tmp_accession)

## ----make_indexes, eval = TRUE------------------------------------------------
# Create temp directory to store the Bowtie2 indices
index_temp <- tempfile()
dir.create(index_temp)

# Create target index
mk_bowtie_index(
  ref_dir = target_ref_temp,
  lib_dir = index_temp,
  lib_name = "target",
  overwrite = TRUE
)

# Create filter index
mk_bowtie_index(
  ref_dir = filter_ref_temp,
  lib_dir = index_temp,
  lib_name = "filter",
  overwrite = TRUE
)

## ----alignment_align----------------------------------------------------------
# Create a temp directory to store output bam file
output_temp <- tempfile()
dir.create(output_temp)

# Get path to example reads
readPath <-
  system.file("extdata", "reads.fastq", package = "MetaScope")

# Align reads to the target genomes
target_map <-
  align_target_bowtie(
    read1 = readPath,
    lib_dir = index_temp,
    libs = "target",
    align_dir = output_temp,
    align_file = "bowtie_target",
    overwrite = TRUE
  )

## ----alignment_filter---------------------------------------------------------
final_map <-
  filter_host_bowtie(
    reads_bam = target_map,
    lib_dir = index_temp,
    libs = "filter",
    make_bam = TRUE, # Set to true to create BAM output
    # Default is to create simplified .csv.gz output
    # The .csv.gz output is much quicker to create!
    overwrite = TRUE,
    threads = 1
  )

## ----bam_primary_alignment----------------------------------------------------
bamFile <- Rsamtools::BamFile(final_map)

param <-
  Rsamtools::ScanBamParam(
    flag = Rsamtools::scanBamFlag(isSecondaryAlignment = FALSE),
    what = c("flag", "rname")
  ) #Gets info about primary alignments

aln <- Rsamtools::scanBam(bamFile, param = param)
accession_all <- aln[[1]]$rname
genome_name_all <- accession_all |> 
  taxonomizr::accessionToTaxa(tmp_accession) |>
  taxonomizr::getTaxonomy(sqlFile = tmp_accession, desiredTaxa = "strain")
read_count_table <- sort(table(genome_name_all), decreasing = TRUE)
knitr::kable(
  read_count_table,
  col.names = c("Genome Assigned", "Read Count"))

## ----bam_secondary_alignment--------------------------------------------------
bamFile <- Rsamtools::BamFile(final_map)

param <-
  Rsamtools::ScanBamParam(
    flag = Rsamtools::scanBamFlag(isSecondaryAlignment = TRUE),
    what = c("flag", "rname")
  ) #Gets info about secondary alignments

aln <- Rsamtools::scanBam(bamFile, param = param)
accession_all <- aln[[1]]$rname
genome_name_all <- accession_all |> 
  taxonomizr::accessionToTaxa(tmp_accession) |>
  taxonomizr::getTaxonomy(sqlFile = tmp_accession, desiredTaxa = "strain")
read_count_table <- sort(table(genome_name_all), decreasing = TRUE)
knitr::kable(
  read_count_table,
  col.names = c("Genome Assigned", "Read Count"))

## ----identification, message = FALSE------------------------------------------
output <- metascope_id(
  final_map,
  input_type = "bam",
  # change input_type to "csv.gz" when not creating a BAM
  aligner = "bowtie2",
  num_species_plot = 0,
  accession_path = tmp_accession
)

knitr::kable(output,
             format = "html",
             digits = 2,
             caption = "Table of MetaScope ID results")

## ----CSV_summary--------------------------------------------------------------
relevant_col <- dirname(final_map) %>%
  file.path("bowtie_target.metascope_id.csv") %>%
  read.csv() %>% dplyr::select(2:4)

relevant_col |>
  dplyr::mutate(
    Genome = stringr::str_replace_all(Genome, ',.*', ''),
    Genome = stringr::str_replace_all(Genome, "(ST398).*", "\\1"),
    Genome = stringr::str_replace_all(Genome, "(N315).*", "\\1"),
    Genome = stringr::str_replace_all(Genome, "(Newman).*", "\\1"),
    Genome = stringr::str_replace_all(Genome, "(Mu3).*", "\\1"),
    Genome = stringr::str_replace_all(Genome, "(RF122).*", "\\1")
  ) |>
  knitr::kable()
unlink(".bowtie2.cerr.txt")

## ----session_info-------------------------------------------------------------
sessionInfo()

---
title: "Explanation of extdata files"
author: "Aubrey Odom"
date: '2022-10-27'
output: html_document
---
## Toy data

### Demultiplexing

`inst/extdata/barcodes.txt`

`inst/extdata/virus_example.fastq`

`inst/extdata/virus_example_index.fastq`

These files consist of randomly selected sequencing reads from viral stock generated at the National Emerging Infectious Diseases Laboratories (NEIDL; Boston, MA). The data is unpublished. The `virus_example.fastq` file contains 150 reads from six different viral stocks. The files `barcodes.txt` and `virus_example_index.fastq` are used in the demultiplexing process. Detailed information on how to reproduce this data is unfortunately unavailable, but it is a relatively small dataset and only intended for use as a toy example.

### Vignette data

`inst/extdata/reads.fastq`

The example reads provided in this package are simulated reads obtained through the wgsim function of SAMtools ([https://doi.org/10.1093/bioinformatics/btp352).](https://doi.org/10.1093/bioinformatics/btp352).) Using wgsim, 1000 simulated reads of Staphylococcus aureus strain RF122 were obtained. In addition, 500 simulated reads of Staphylococcus epidermidis strain RP62A were also obtained to act as a filter target.

Below are the steps to recreating the reads.

#### Inside R

```{R, eval = FALSE}
library(MetaScope)

download_refseq(taxon = "Staphylococcus aureus RF122", representative = FALSE, reference = FALSE, compress = FALSE)

download_refseq(taxon = "Staphylococcus epidermidis RP62A", representative = FALSE, reference = FALSE, compress = FALSE)

```

#### Command line code (UNIX)

Make sure that samtools is installed / loaded to use wgsim.

```{bash, eval = FALSE}
mv Staphylococcus\\ aureus\\ RF122.fasta Staphylococcus_aureus_RF122.fasta

mv Staphylococcus\\ epidermidis\\ RP62A.fasta Staphylococcus_epidermidis_RP62A.fasta

wgsim -N 1000 -1 250 -2 250 -S 32 Staphylococcus_aureus_RF122.fasta read1.fq read2.fq

wgsim -N 500 -1 250 -2 250 -S 32 Staphylococcus_epidermidis_RP62A.fasta read3.fq read4.fq

cat read1.fq read3.fq > reads.fastq

```

Files created: reads.fastq

### Files used in examples

`bowtie_target.bam`

```{R, eval = FALSE}
#### Align example reads to an example reference library using Rbowtie2

target_ref_temp <- tempfile()
dir.create(target_ref_temp)
MetaScope::download_refseq("Measles morbillivirus",
                           reference = TRUE,
                           representative = FALSE,
                           compress = TRUE,
                           out_dir = target_ref_temp
)

MetaScope::download_refseq("Zaire ebolavirus",
                           reference = FALSE,
                           representative = FALSE,
                           compress = TRUE,
                           out_dir = target_ref_temp
)

# Create temp directory to store the indices
index_temp <- tempfile()
dir.create(index_temp)
MetaScope::mk_bowtie_index(
  ref_dir = target_ref_temp,
  lib_dir = index_temp,
  lib_name = "target",
  overwrite = TRUE
)

output_temp <- tempfile()
dir.create(output_temp)

# Get path to example reads
readPath <- system.file("extdata", "virus_example.fastq",
                        package = "MetaScope")
data("bt2_loose_params")
target_map <-
  MetaScope::align_target_bowtie(
    read1 = readPath,
    lib_dir = index_temp,
    libs = "target",
    align_dir = output_temp,
    align_file = "bowtie_target",
    overwrite = TRUE,
    bowtie2_options = bt2_loose_params
  )

# Copy the file from output_temp into your directory
stem <- getwd()
file.copy(target_map, file.path(stem, "..",
                                "extdata", "bowtie_target.bam"))

unlink(target_ref_temp, recursive = TRUE)
unlink(index_temp, recursive = TRUE)
unlink(output_temp, recursive = TRUE)
```

`bowtie_target.filtered.csv.gz`

```{R, eval = FALSE}
# Create temporary filter library
filter_ref_temp <- tempfile()
dir.create(filter_ref_temp)

MetaScope::download_refseq("Zaire ebolavirus",
                           reference = FALSE,
                           representative = FALSE,
                           compress = TRUE,
                           out_dir = filter_ref_temp
)

# Create temp directory to store the indices
index_temp <- tempfile()
dir.create(index_temp)

MetaScope::mk_bowtie_index(
  ref_dir = filter_ref_temp,
  lib_dir = index_temp,
  lib_name = "filter",
  overwrite = TRUE
)

output_temp <- tempfile()
dir.create(output_temp)

# Get path to example bam
bamPath <- system.file("extdata", "bowtie_target.bam",
                       package = "MetaScope")
target_copied <- file.path(output_temp, "bowtie_target.bam")
file.copy(bamPath, target_copied)

filter_out <-
  filter_host_bowtie(
    reads_bam = target_copied,
    lib_dir = index_temp,
    libs = "filter",
    threads = 1
  )

stem <- getwd()
file.copy(target_map,
          file.path(stem, "..",
                    "extdata", "bowtie_target.filtered.csv.gz"))

unlink(filter_ref_temp, recursive = TRUE)
unlink(index_temp, recursive = TRUE)
unlink(output_temp, recursive = TRUE)
```

`subread_target.bam`

```{R}
## Make and align to multiple reference libraries                  
target_ref_temp <- tempfile()
dir.create(target_ref_temp)
all_species <- c("Staphylococcus aureus RF122",
                 "Staphylococcus aureus subsp. aureus Mu3")
all_ref <- vapply(all_species, MetaScope::download_refseq, 
       reference = FALSE, representative = FALSE, compress = TRUE,
       out_dir = target_ref_temp, FUN.VALUE = character(1))
ind_out <- vapply(all_ref, mk_subread_index, FUN.VALUE = character(1))
# Get path to example reads
readPath <- system.file("extdata", "reads.fastq",
                        package = "MetaScope")
## Copy the example reads to the temp directory
refPath <- file.path(target_ref_temp, "reads.fastq")
file.copy(from = readPath, to = refPath)
data("align_details")
align_details[["type"]] <- "rna"
align_details[["phredOffset"]] <- 50
# Just to get it to align - toy example!
align_details[["maxMismatches"]] <- 10

target_map <- align_target(refPath,
                           libs = stringr::str_replace_all(all_species, " ", "_"),
                           lib_dir = target_ref_temp,
                           subread_options = align_details)

stem <- getwd()
file.copy(target_map,
          file.path(stem, "..",
                    "extdata", "subread_target.bam"))

unlink(target_ref_temp, recursive = TRUE)
unlink(index_temp, recursive = TRUE)
unlink(output_temp, recursive = TRUE)
```

`subread_target.filtered.csv.gz`

```{R}
## Make and align to multiple reference libraries                  
filter_ref_temp <- tempfile()
dir.create(filter_ref_temp)

all_species <- c("Staphylococcus aureus subsp. aureus str. Newman")
all_ref <- vapply(all_species, MetaScope::download_refseq, 
       reference = FALSE, representative = FALSE, compress = TRUE,
       out_dir = filter_ref_temp, FUN.VALUE = character(1))
ind_out <- vapply(all_ref, mk_subread_index, FUN.VALUE = character(1))
# Get path to example reads
readPath <- system.file("extdata", "subread_target.bam",
                        package = "MetaScope")
## Copy the example reads to the temp directory
refPath <- file.path(filter_ref_temp, "subread_target.bam")
file.copy(from = readPath, to = refPath)
data("align_details")
align_details[["type"]] <- "rna"
align_details[["phredOffset"]] <- 10
# Just to get it to align - toy example!
align_details[["maxMismatches"]] <- 10

filtered_map <- filter_host(refPath, lib_dir = filter_ref_temp,
            libs = stringr::str_replace_all(all_species, " ", "_"),
            threads = 1, subread_options = align_details)

stem <- getwd()
file.copy(filtered_map,
          file.path(stem, "..",
                    "extdata", "subread_target.filtered.csv.gz"))

unlink(filter_ref_temp, recursive = TRUE)
unlink(index_temp, recursive = TRUE)
unlink(output_temp, recursive = TRUE)
```

---
title: "Using the SILVA 16S rRNA Database with MetaScope"
author: "Aubrey Odom"
date: "2024-01-01"
output: html_document
---

# The SILVA database

MetaScope is compatible with the SILVA 16S rRNA Database. At the time of writing, the [most recent update](https://www.arb-silva.de/) to SILVA is version 138.1 (released on August 27, 2020). These notes will supplement the main MetaScope pipeline as illustrated in the [vignette](https://wejlab.github.io/metascope-docs/articles/MetaScope_vignette.html).

# SILVA Indices
Currently, MetaScope is only functional with SILVA when using the premade Bowtie2 indices of SILVA 138.1 available [here](https://rutgers.box.com/s/njq1qc9bzawzs3a6513uihxxwgxu9qc5).
# Accessions
Before running any samples, you'll need to run `download_accessions()` with the argument `silva_taxonomy_database = TRUE`.

# Alignment
You can follow the `align_target_bowtie` usage outlined in the MetaScope tutorial, replacing the `libs` parameter with the SILVA indices.

# Identification
When you reach the `metascope_id()` step, you will need to include the parameter `db = "silva"` in the function call to identify down to the genus level using the SILVA taxonomy identifiers. 

# Condense results
All CSV outputs from `metascope_id()` can be compiled into a single MultiAssayExperiment object with the `convert_animalcules_silva()` function as shown here with false filenames:

```{R, eval = FALSE}
out_res <- convert_animalcules_silva(
  meta_counts = all_files,
  annot_path = "animalcules_metadata.csv",
  which_annot_col = "Filename",
  end_string = ".metascope_id.csv",
  qiime_biom_out = FALSE,
  path_to_write = "."
)

saveRDS(out_res, "silva_consolidated_results.RDS")
```

---
title: "Example MetaScope Script"
author: 
- name: Aubrey Odom
  affiliation: 
  - Program in Bioinformatics, Boston University, Boston, MA
  email: aodom@bu.edu
date: '`r format(Sys.Date(), "%B %e, %Y")`'
package: MetaScope
output: 
  BiocStyle::html_document
vignette: >
  %\VignetteIndexEntry{Introdution to MetaScope}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

# Introduction
The following is an example script utilizing the Bowtie 2 sub-workflow. The demultiplex step is omitted.

```{R}
library(MetaScope)
```

# Creation of indices

```{R, eval = FALSE}
nThreads <- 8
loc <- "/restricted/projectnb/pathoscope/reflib/2023_refseq_bowtie"

# Download taxonomy database
  # This should be done ONCE before running any samples
  # The same downloaded database can be used for all samples
download_accessions(file.path(loc, "indices"),
                    NCBI_accessions_database = TRUE,
                    NCBI_accessions_name = "accessionTaxa.sql",
                    silva_taxonomy_database = TRUE,
                    silva_taxonomy_name = "all_silva_headers.rds",
                    blast_16S_database = TRUE,
                    blast_16S_name = "16S_ribosomal_RNA"))

targets <- tolower(c("bacteria"))
filters <- c('homo sapiens', 'mus musculus')
inVec <- c(targets, filters)

# make directories
mk_all_dir <- function(ind) {
    ind_ <- stringr::str_replace_all(ind, c(" " = "_"))
    if (!dir.exists(file.path(loc, ind_))) dir.create(file.path(loc, ind_))
    }
sapply(inVec, mk_all_dir)

# Download reference and representative genomes
sapply(inVec, 
       function(x) download_refseq(x, reference = TRUE,
                                   representative = TRUE,
                                   compress = TRUE,
                                   patho_out = FALSE,
                                   # Same directory as created above
                                   out_dir = file.path(
                                   loc, stringr::str_replace_all(x, c(" " = "_"))),
                                   accession_path = file.path(loc, "indices", "accessionTaxa.sql")))

# Make indices
for (ind in inVec) {
  ind_ <- stringr::str_replace_all(ind, c(" " = "_"))
  # Create the reference library index files in the destination directory
  if (ind == "fungi") {
      mk_bowtie_index(ref_dir = file.path(loc, ind_), lib_dir = file.path(loc, "all_indices"), 
                  lib_name = ind_,
                  bowtie2_build_options = "--large-index",
                  threads = nThreads, overwrite = TRUE)
  } else {
      mk_bowtie_index(ref_dir = file.path(loc, ind_), lib_dir = file.path(loc, "all_indices"), 
                      lib_name = paste0(ind_, "rep"), threads = nThreads, overwrite = TRUE)
  }
}


```

# Command-line pre-processing (BASH script)

Note, the variable `SGE_TASK_ID` refers to an index integer set via a parallel processing job submission system for a computing resource.

```{bash, eval = FALSE}
threads=8

dataDir=/restricted/projectnb/pathoscope/work/aubrey/metascope_benchmarking/miossec_data
dataDir=/restricted/projectnb/pathoscope/work/aubrey/metascope_benchmarking/miossec_data/species_10/abundance_100000/dominance_1

### MAKE WORKING DIR
workingDir=${SGE_TASK_ID}_tmp_miossec
rm -rf $TMPDIR/$workingDir
mkdir $TMPDIR/$workingDir

# Specify input files
input_line=$(sed "${SGE_TASK_ID}q;d" /restricted/projectnb/pathoscope/work/aubrey/metascope_benchmarking/miossec_data/dat_manifest.txt) 
IFS=$'\t'
read -r sampleName readPath1 readPath2 <<< "$input_line"

### TRIM the reads
# CHANGE CROP AND HEADCROP DEPENDING ON MultiQC ANALYSIS
#java -jar ~/pathoscope/code/other/Trimmomatic-0.36/trimmomatic-0.36.jar PE \
#    -phred33 -threads $threads $reads1 $reads2 \
#    $TMPDIR/$workingDir/reads1.paired.fastq.gz \
#    $TMPDIR/$workingDir/reads1.unpaired.fastq.gz \
#    $TMPDIR/$workingDir/reads2.paired.fastq.gz \
#    $TMPDIR/$workingDir/reads2.unpaired.fastq.gz \
#    SLIDINGWINDOW:4:20 LEADING:3 TRAILING:3 MINLEN:36 \
#    CROP:225 HEADCROP:30

### Run MetaScope
indexDir="/restricted/projectnb/pathoscope/reflib/2023_refseq_bowtie/all_indices"
expTag=$sampleName
outDir="/restricted/projectnb/pathoscope/work/aubrey/metascope_benchmarking/refseq2023_output"
tmpDir="$TMPDIR/$workingDir/"

# Choose filter and targets to download
target="bacteriarep,fungi,viruses"
filter="human_T2T,mus_musculus"

# This is where samtools and R are loaded if necessary
module load samtools
module load R

# Change "run_MetaScope.R" to name of R script saved elsewhere
Rscript --vanilla --max-ppsize=500000 run_MetaScope.R \
${readPath1} ${readPath2} ${indexDir} ${expTag} ${outDir} ${tmpDir} ${threads} \
${target} ${filter}

rm -rf $TMPDIR/$workingDir

```

# R script

```{R, eval = FALSE}
# Take in arguments from bash script
args <- commandArgs(trailingOnly = TRUE)

readPath1 <- args[1]
readPath2 <- args[2]
indexDir <- args[3]
expTag <- args[4]
outDir <- args[5]
tmpDir <- args[6]
threads <- args[7]
targets <- stringr::str_split(args[8], ",")[[1]]
filters <- stringr::str_split(args[9], ",")[[1]]

# Time this!
now <- Sys.time()

# Load MetaScope
library(MetaScope)

# Align to targets
do_this <- function(x) stringr::str_replace_all(x, c(" " = "_"))
targets_ <- do_this(targets) 
filters_ <- do_this(filters)

# Identify bt2 params
data(bt2_regular_params)
bt2_params <- bt2_regular_params

target_map <- align_target_bowtie(read1 = readPath1,
                                  read2 = readPath2,
                                  lib_dir = indexDir,
                                  libs =  targets_,
                                  align_dir = tmpDir,
                                  align_file = expTag,
                                  overwrite = TRUE,
                                  threads = threads,
                                  bowtie2_options = paste(bt2_params, "-f"),
                                  quiet = FALSE)

# Align to filters
output <- paste(paste0(tmpDir, expTag), "filtered", sep = ".")
final_map <- filter_host_bowtie(reads_bam = target_map,
                                lib_dir = indexDir,
                                libs = filters_,
                                make_bam = FALSE,
                                output = output,
                                threads = threads,
                                overwrite = TRUE,
                                quiet = FALSE,
                                bowtie2_options = bt2_params)

# MetaScope ID

metascope_id(final_map, input_type = "csv.gz", aligner = "bowtie2",
             accession_path = file.path(loc, "indices", "accessionTaxa.sql"),
             num_species_plot = 15,
             quiet = FALSE,
             out_dir = outDir)

message(capture.output(Sys.time() - now))
```

---
title: "Publicly Hosted MetaScope Databases and Bowtie 2 Indices"
author: "Sean Lu"
date: "2025-01-30"
output: html_document
---

# Introduction

To mitigate the effort of downloading reference genomes, taxonomy identifier databases, and constructing Bowtie 2 indices, we provide pre-compiled versions as described below. Current versions of all of these files can be constructed with MetaScope's functions if users choose not to download the files here.

# Link to [Box Drive](https://rutgers.box.com/s/nar5z3g02r2tz9lr3mdqwq3rcekh4fca)

All databases are hosted centrally on a [Box drive](https://rutgers.box.com/s/nar5z3g02r2tz9lr3mdqwq3rcekh4fca) for download. A brief explanation of available files follows:

# Taxonomy Annotation Database (`2024_Accession_Taxa`)

This file is a zip compressed folder containing a database of taxonomy information prepared with `download_accessions()` as part of MetaScope's MetaRef module.

It is *highly recommended* that all users download this database if they are using NCBI or SILVA databases. Otherwise, users will need to run ` MetaScope::download_accessions()` to obtain this database.

# MetaScope Bowtie2 Indices (`metascope_bowtie2_indices`)

Users of the MetaScope pipeline can easily and efficiently obtain reference genomes of interest using the `download_refseq()` function. However, the additional step of creating Bowtie 2 indices from these genomes to align them with sample data can add hours to these initial steps.

This folder contains various pre-built bowtie2 indices that can be used with the the `align_target_bowtie()` and `filter_host_bowtie` functions. The user's *local* path to these indices should be supplied to the `lib_dir` argument of those functions.

The following indices are available in [Box drive](https://rutgers.box.com/s/nar5z3g02r2tz9lr3mdqwq3rcekh4fca):

1. `2024_ncbi_16S` The NCBI 16S ribosomal RNA database
2. `Greenegenes 13_8` The Greengenes 13.8 16S ribosomal RNA database
3. `MetaScope RefSeq 2023` The entire bacteria, fungi, homo_sapiens, human_T2T, mus_musculus, and viruses refseq nucleotide database downloaded via `download_refseq()` in 2023
4. `MetaScope RefSeq 2022` The entire bacteria, fungi, human, mouse, and viruses refseq nucleotide database downloaded via `download_refseq()` in 2022
5. `MetaScope RefSeq 2020` The entire bacteria, fungi, human/mouse, Phi X 174, and viral refseq nucleotide database downloaded via `download_refseq()` in 2020
6. `PathoScope RefSeq 2018` The entire  bacteria, mouse, and viral refseq nucleotide database downloaded via PathoScope 2.0 software in 2018
7. `PathoScope RefSeq 2015` The entire  bacteria, fungi, human/mouse, phix174, and viral refseq nucleotide database downloaded via PathoScope 2.0 software in 2015
8. `SILVA 138.1` The SILVA 138.1 16S ribosomal RNA database

---
title: "Introduction to MetaScope"
author: 
- name: Aubrey Odom
  affiliation: 
  - Program in Bioinformatics, Boston University, Boston, MA
  email: aodom@bu.edu
- name: Rahul Varki
  affiliation:
  - Research Assistant in Johnson Lab, Boston University School of Medicine, Boston, MA
  email: rvarki@bu.edu
- name: W. Evan Johnson
  affiliation:
  - The Section of Computational Biomedicine, Boston University School of Medicine, Boston, MA
  email: wej@bu.edu
date: '`r format(Sys.Date(), "%B %e, %Y")`'
package: MetaScope
output: 
  BiocStyle::html_document
vignette: >
  %\VignetteIndexEntry{"Introduction to MetaScope"}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

# Important Note

Although it is not a requirement, MetaScope will fun faster and more efficiently for larger samples if the samtools package is present on your device. You can download it at <https://github.com/samtools/samtools>.

# Introduction

MetaScope is a complete metagenomics profiling package that can accurately identify the composition of microbes within a sample at a strain-level resolution. MetaScope can be considered as an updated and expanded R translation of [PathoScope 2.0](https://microbiomejournal.biomedcentral.com/articles/10.1186/2049-2618-2-33), a Python-based metagenomic profiling package created by the Johnson lab. A few improvements made in MetaScope include using the BAM file format instead of the SAM file format for significantly less disk space usage, removing all dependencies to NCBI's now defunct GI sequence annotations, and properly filtering reads that align to filter reference genomes. Functions to analyze host microbiome data are also planned to be added in future updates to the package.

## MetaScope Workflows

The major workflow of MetaScope is delineated below. It is composed of core modules that are formed by groups  of functions.

![Figure 1. The MetaScope workflow and its associated modules with function descriptions. The MetaRef, MetaAlign, MetaFilter, and MetaID modules form the backbone of package operation, whereas the MetaDemultiplex and MetaCombine modules are complementary to the core package functionality.](MetaScope_Workflow.png)

The core modules are as follows:
1. **MetaDemultiplex:** Obtain non-barcoded sequencing reads
2. **MetaRef:** Obtain target and filter genome sequences from NCBI nucleotide database and index using a given aligner
3. **MetaAlign:** Align sequencing reads to indexed target genome sequences
4. **MetaFilter:** Remove reads mapped to indexed host genome sequences
5. **MetaID:** Reassign ambiguously mapped reads to probable source genome
6. **MetaCombine:** Aggregate samples into a MultiAssayExperiment compatible with the `animalcules` R package.

There are two sub-workflows that are included in the package, as seen in Figure 1: the Rbowtie2 and the Rsubread workflow. The major difference is that the functions in the MetaRef, MetaAlign, and MetaFilter modules differ by the aligner utilized. 

The Rbowtie2 functions utilize the [Bowtie2](https://www.nature.com/articles/nmeth.1923) aligner (Langmead 2012) whereas the [Rsubread](https://academic.oup.com/nar/article/47/8/e47/5345150) functions utilize the Rsubread aligner (Liao 2019). The nuances of how to use each function can be found by looking at each function's help manual (R command: `?<name of function>`). For reference, PathoScope 2.0 uses the Bowtie2 aligner in its workflow.

In this vignette, we will analyze the mock data provided in the package via the Rbowtie2 sub-workflow. We will utilize all of the core modules in sequential order. We will make mention of the equivalent Rsubread function whenever an Rbowtie2 function is being used. For the purposes of this example, the MetaCombine module will be omitted.

## Installation

In order to install MetaScope from Bioconductor, run the following code:

```{R show_install, eval = FALSE}
if (!requireNamespace("BiocManager", quietly = TRUE))
  install.packages("BiocManager")
BiocManager::install("MetaScope")
```

```{R load_packages, eval = TRUE}
suppressPackageStartupMessages({
  library(MetaScope)
  library(magrittr)
})
```

# Data

The mock data provided in the package consists of simulated sequencing data generated from the `SAMtools` `wgsim` function (see `extdata_explanations.Rmd` in `inst/script` to see exact commands). The `wgsim` function is a tool which allows for the generation of FASTQ reads from a reference genome (FASTA). The mock data (`reads.fastq`) contains 1500 reads, of which 1000 reads are derived from the *Staphylococcus aureus* RF122 strain and 500 reads are derived from the *Staphylococcus epidermidis* RP62A strain. In this data set, we assume that the *S. aureus* RF122 reads are the reads of interest, and the *S. epidermidis* RP62A reads are known contaminant reads which should be removed during the analysis. Ideally, the microbial composition report (.csv) produced at the end of the analysis should contain only reads assigned to the *S. aureus* RF122 strain.

# MetaDemultiplex: Demultiplexing reads

Sequence runs on NGS instruments are typically carried out with multiple samples pooled together. An index tag (also called a barcode) consisting of a unique sequence of between 6 and 12bp is added to each sample so that the sequence reads from different samples can be identified. For 16s experiments or sequencing conducted using an Illumina machine, the process of demultiplexing (dividing your sequence reads into separate files for each index tag/sample) and generating the FASTQ data files required for downstream analysis can be done using the MetaScope demultiplexing workflow. This consists of the `meta_demultiplex()` function, which takes as arguments a matrix of sample names/barcodes, a FASTQ file of barcodes by sequence header, and a FASTQ file of reads corresponding to the barcodes. Based on the barcodes given, the function extracts all reads for the indexed barcode and writes all the reads from that barcode to separate FASTQ files.

This is an optional step in the analysis process depending on whether your reads are multiplexed. The reads which we are currently trying to analyze are not multiplexed and therefore this step is skipped in our analysis. The example shown below is using different reads that are barcoded in order to show the utility of the function.

```{R meta_demultiplex, message = FALSE}
# Get barcode, index, and read data locations
barcodePath <-
  system.file("extdata", "barcodes.txt", package = "MetaScope")
indexPath <- system.file("extdata", "virus_example_index.fastq",
                         package = "MetaScope")
readPath <-
  system.file("extdata", "virus_example.fastq", package = "MetaScope")

# Get barcode, index, and read data locations
demult <-
  meta_demultiplex(barcodePath,
                   indexPath,
                   readPath,
                   rcBarcodes = FALSE,
                   hammingDist = 2,
                   location = tempfile())
demult
```

# MetaRef: Reference Genome Library

The MetaScope genome library workflow is designed to assist with collection of sequence data from the National Center for Biotechnology Information (NCBI) nucleotide database. Prior to doing so, the potential targets and filters for the analysis should be identified. That is, what "target" species do you expect to find in your metagenomic sample that you would like to identify, and what reads would you like to "filter" out from the data that are not essential to your analysis?

Typically, the targets of the analysis are microbes (that is, viruses, bacteria, and fungi), and we wish to filter out or discard any reads from the host in addition to artificially added sequences, such as Phi X 174. Following identification of the targets and filters, we use a reference genome library to align the vast number of sample reads back to the respective regions of origin in various species.

The `download_refseq()` function automatically extracts custom reference genome libraries in a FASTA file for microbial or host genomes. The user must first indicate a taxon for which to download the genomes, such as 'bacteria' or 'primates'. A table of possible entries can be viewed by accessing the `MetaScope:::taxonomy_table` object. The user may then specify whether they wish to download only the RefSeq reference genomes, or both the reference and representative genomes. The compress option then allows users to specify whether to compress the output FASTA file; this is done by default.

## Creating a Taxonomy Database

A reference of taxonomy reference identifiers (e.g. NCBI accessions) is critical for fully identifying microbes with the MetaScope pipeline. This step should be completed ONCE, PRIOR TO RUNNING ANY AND ALL SAMPLES with the `download_accessions()` function. For an example of how to make a typical database for all host and filter genomes in your sample, please see the [example script](https://wejlab.github.io/metascope-docs/articles/docs/Example_Script.html) on our website.

If using the SILVA 16S database, NCBI nucleotide or RefSeq database, you can use `download_accessions()` with the following arguments:

```{R eval = FALSE}
download_accessions(
  ind_dir = "C:/Users/JohnSmith/ResearchIndices",
  tmp_dir = file_path(ind_dir, "tmp"),
  remove_tmp_dir = TRUE, 
  # For NCBI RefSeq or nucleotide database
  NCBI_accessions_database = TRUE,
  NCBI_accessions_name = "accessionTaxa.sql",
  # For SILVA 16S database
  silva_taxonomy_database = TRUE,
  silva_taxonomy_name = "all_silva_headers.rds"
)
```

However, for this vignette, we know in advance precisely which genomes and strains we will be testing. Therefore, we will generate a small subdatabase of taxonomizr references. We do not need to use `download_accessions()` in this case.

```{R taxonomy_db, eval = TRUE, warning = FALSE, message = FALSE}
tmp_accession <- system.file("extdata","example_accessions.sql", package = "MetaScope")
```

## Downloading target genomes

Even though in this scenario we know exactly from where the reads in the mock data (`reads.fastq`) originate, in most cases we may only have a general idea of read origins. Therefore, in the following code, we will download the genome of the *Staphylococcus aureus* RF122 strain along with the genomes of a few other closely related *Staphylococcus aureus* strains from the NCBI RefSeq database. These genomes together will act as our target genome library.


```{R target_lib, eval = TRUE, warning = FALSE, message = FALSE}
target_ref_temp <- tempfile()
dir.create(target_ref_temp)

all_species <- c("Staphylococcus aureus subsp. aureus Mu50",
                 "Staphylococcus aureus subsp. aureus Mu3",
                 "Staphylococcus aureus subsp. aureus str. Newman",
                 "Staphylococcus aureus subsp. aureus N315",
                 "Staphylococcus aureus RF122", 
                 "Staphylococcus aureus subsp. aureus ST398")
sapply(all_species, download_refseq, 
       reference = FALSE, representative = FALSE, compress = TRUE,
       out_dir = target_ref_temp, caching = TRUE, accession_path = tmp_accession)
```

## Downloading filter genomes

We will also download the reference genome and related sequences* of the *Staphylococcus epidermidis* RP62A strain from the NCBI nucleotide database, in an uncompressed FASTA format. This genome will act as our filter library.

* since we are downloading from the nucleotide database with representative and reference = FALSE, several sequences will be downloaded in addition to the main genome.
```{R filter_lib, warning = FALSE, message = FALSE}
filter_ref_temp <- tempfile()
dir.create(filter_ref_temp)

download_refseq(
  taxon = "Staphylococcus epidermidis RP62A",
  representative = FALSE, reference = FALSE,
  compress = TRUE, out_dir = filter_ref_temp,
  caching = TRUE,
  accession_path = tmp_accession)
```

## Creating indices using a given aligner

We now use `mk_bowtie_index()`, a wrapper for the `Rbowtie2::bowtie2_build` function, to generate Bowtie2 compatible indexes from the reference genomes that were previously downloaded. The target and reference genome files (.fasta or .fasta.gz extension) must be placed into their own separate empty directories prior to using the function. This is due to the fact that the function will attempt to build the indexes from all the files present in the directory. The function will give an error if other files (other than .fasta or .fasta.gz) are present in the directory. Depending on the combined size of the reference genomes, the function will automatically create either small (.bt2) or large (.bt2l) Bowtie2 indexes.

The target and filter reference genomes downloaded in the previous step have been combined and renamed to target.fasta and filter.fasta respectively for convenience. These are the files from which the Bowtie2 indexes will be made from.

```{R make_indexes, eval = TRUE}
# Create temp directory to store the Bowtie2 indices
index_temp <- tempfile()
dir.create(index_temp)

# Create target index
mk_bowtie_index(
  ref_dir = target_ref_temp,
  lib_dir = index_temp,
  lib_name = "target",
  overwrite = TRUE
)

# Create filter index
mk_bowtie_index(
  ref_dir = filter_ref_temp,
  lib_dir = index_temp,
  lib_name = "filter",
  overwrite = TRUE
)
```

# Alignment with Reference Libraries

After acquiring the target and filter genome libraries, we then take the sequencing reads from our sample and map them first to the target library and then to the filter library. MetaScope's Rbowtie2 mapping function utilizes the [Bowtie2](https://www.nature.com/articles/nmeth.1923) aligner (Langmead 2012) which maps reads to a reference genome using a full-text minute index based approach. Essentially, the algorithm extracts substrings which are referred to as "seeds" from the reads and aligns them to the reference genomes with the assistance from the full-text minute index. Seed alignments to the reference genomes are prioritized and then finally extended into full alignments using dynamic programming.

## MetaAlign

Following index creation, we will use the Bowtie2 aligner to map the reads to the target genomes with the `align_target_bowtie()` function (Rsubread equivalent: `align_target()`). The function takes as an input the location of the FASTQ file to align, the directory where the indexes are stored, the names of the indexes to align against, the directory where the BAM file should be written, and the basename of the output BAM file. 

In practice, `align_target_bowtie()` maps reads to each target library separately, removes the unmapped reads from each file, and finally merges and sorts by chromosome the BAM files from each library into a single output file (same with `align_target`). If `SAMtools` is installed on the machine and can be found by the `Sys.which("samtools")` R command, the BAM file will be directly created, otherwise an intermediate SAM file will be created prior to the creation of the BAM file which could potentially create issues if the SAM file is large and there is limited disk space. The default alignment parameters are the same as PathoScope 2.0's default alignment parameters, but users can provide their own Bowtie 2 alignment settings if desired.

We will now align the sample reads (reads.fastq) to the target reference genomes using the Bowtie 2 indexes that we just built.

```{R alignment_align}
# Create a temp directory to store output bam file
output_temp <- tempfile()
dir.create(output_temp)

# Get path to example reads
readPath <-
  system.file("extdata", "reads.fastq", package = "MetaScope")

# Align reads to the target genomes
target_map <-
  align_target_bowtie(
    read1 = readPath,
    lib_dir = index_temp,
    libs = "target",
    align_dir = output_temp,
    align_file = "bowtie_target",
    overwrite = TRUE
  )
```

## MetaFilter
The last step in the mapping workflow is to filter the output BAM file according to the reference genome for the filter/host species. Although we have already filtered out any unmapped reads, which may belong to one or more host species or otherwise, there may still remain some sort of unwelcome contamination in the data from the filter species which we wish to remove. To do this, we employ `filter_host_bowtie()` (Rsubread equivalent: `filter_host()`), which takes as an input the location of the BAM file created from `align_target_bowtie()`, the directory where the indexes are stored, and the names of the filter indexes to align against, to produce a sorted BAM file with any reads that match the filter libraries removed. We will then use this final BAM file downstream for further analysis.

```{R alignment_filter}
final_map <-
  filter_host_bowtie(
    reads_bam = target_map,
    lib_dir = index_temp,
    libs = "filter",
    make_bam = TRUE, # Set to true to create BAM output
    # Default is to create simplified .csv.gz output
    # The .csv.gz output is much quicker to create!
    overwrite = TRUE,
    threads = 1
  )
```

## Evaluating alignments (supplemental)
**Note: the next two code blocks are included for the sake of examining the vignette example, but are not useful for "real life" data. To continue your analysis, head down to the Genome Identification header.**

Prior to the last step in the analysis, we will look at the primary alignments of the mapped reads in the filtered BAM file that we just created using the `filter_host_bowtie()` function. According to the [Bowtie2 manual](http://bowtie-bio.sourceforge.net/bowtie2/manual.shtml), a primary alignment is described as the alignment that received the highest alignment score among all alignments for that read. When looking at the primary alignments of the mapped reads, we can see that the majority of reads have mapped to the correct *Staphylococcus aureus* RF122 strain. However, some residual reads have primary alignments to the other *S. aureus* strains which are incorrect. If we were to stop the analysis at this point, we could potentially be lead to believe that our sample has increased microbial diversity, when it actually does not.

```{R bam_primary_alignment}
bamFile <- Rsamtools::BamFile(final_map)

param <-
  Rsamtools::ScanBamParam(
    flag = Rsamtools::scanBamFlag(isSecondaryAlignment = FALSE),
    what = c("flag", "rname")
  ) #Gets info about primary alignments

aln <- Rsamtools::scanBam(bamFile, param = param)
accession_all <- aln[[1]]$rname
genome_name_all <- accession_all |> 
  taxonomizr::accessionToTaxa(tmp_accession) |>
  taxonomizr::getTaxonomy(sqlFile = tmp_accession, desiredTaxa = "strain")
read_count_table <- sort(table(genome_name_all), decreasing = TRUE)
knitr::kable(
  read_count_table,
  col.names = c("Genome Assigned", "Read Count"))
```

We can also look at the secondary alignments of the mapped reads within our filtered BAM file. A secondary alignment occurs when a read maps to multiple different genomes. We can see that the majority of our secondary alignments are to the other Staphylococcus aureus strains, which makes sense considering that the majority of the primary alignments were to the correct Staphylococcus aureus RF122 strain.

```{R bam_secondary_alignment}
bamFile <- Rsamtools::BamFile(final_map)

param <-
  Rsamtools::ScanBamParam(
    flag = Rsamtools::scanBamFlag(isSecondaryAlignment = TRUE),
    what = c("flag", "rname")
  ) #Gets info about secondary alignments

aln <- Rsamtools::scanBam(bamFile, param = param)
accession_all <- aln[[1]]$rname
genome_name_all <- accession_all |> 
  taxonomizr::accessionToTaxa(tmp_accession) |>
  taxonomizr::getTaxonomy(sqlFile = tmp_accession, desiredTaxa = "strain")
read_count_table <- sort(table(genome_name_all), decreasing = TRUE)
knitr::kable(
  read_count_table,
  col.names = c("Genome Assigned", "Read Count"))
```

# MetaID: Origin Genome Identification

Following the proper alignment of a sample to all target and filter libraries of interest, we may proceed in identifying which genomes are most likely to be represented in the sample. This identification workflow is the core of MetaScope; it features a Bayesian read reassignment model which dramatically improves specificity and sensitivity over other methods (Francis et. al 2013). This is because such a method identifies reads with unique alignments and uses them to guide the reassignment of reads with ambiguous alignments.

The MetaID identification module consists of a single function, `MetaScope_ID()`, which reads in a .bam file, annotates the taxonomy and genome names, reduces the mapping ambiguity using a mixture model, and outputs a .csv file with the results. Currently, it assumes that the genome library/.bam files use NCBI accession names for reference names.

```{R identification, message = FALSE}
output <- metascope_id(
  final_map,
  input_type = "bam",
  # change input_type to "csv.gz" when not creating a BAM
  aligner = "bowtie2",
  num_species_plot = 0,
  accession_path = tmp_accession
)

knitr::kable(output,
             format = "html",
             digits = 2,
             caption = "Table of MetaScope ID results")
```

**Note: the next code block is included for the sake of examining the vignette example, but can be skipped. Your results are in the CSV file produced by the `metascope_id()` function.**

We will now look at the read reassignment results reported in the output CSV file.

```{R CSV_summary}
relevant_col <- dirname(final_map) %>%
  file.path("bowtie_target.metascope_id.csv") %>%
  read.csv() %>% dplyr::select(2:4)

relevant_col |>
  dplyr::mutate(
    Genome = stringr::str_replace_all(Genome, ',.*', ''),
    Genome = stringr::str_replace_all(Genome, "(ST398).*", "\\1"),
    Genome = stringr::str_replace_all(Genome, "(N315).*", "\\1"),
    Genome = stringr::str_replace_all(Genome, "(Newman).*", "\\1"),
    Genome = stringr::str_replace_all(Genome, "(Mu3).*", "\\1"),
    Genome = stringr::str_replace_all(Genome, "(RF122).*", "\\1")
  ) |>
  knitr::kable()
unlink(".bowtie2.cerr.txt")
```

We can see that the read reassignment function has reassigned the majority of the ambiguous alignments back to the *Staphylococcus aureus* RF122 strain, the correct strain of origin.

# Session Info

```{r session_info}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mk_subread_index.R
\name{mk_subread_index}
\alias{mk_subread_index}
\title{Make a Subread index}
\usage{
mk_subread_index(ref_lib, split = 4, mem = 8000, quiet = TRUE)
}
\arguments{
\item{ref_lib}{The name/location of the reference library file, in
(uncompressed) .fasta format.}

\item{split}{The maximum allowed size of the genome file (in GB). If the
\code{ref_lib} file is larger than this, the function will split the
library into multiple parts.}

\item{mem}{The maximum amount of memory (in MB) that can be used by the index
generation process (used by the Rsubread::buildindex function).}

\item{quiet}{Turns off most messages. Default is \code{TRUE}.}
}
\value{
Creates one or more Subread indexes for the supplied reference .fasta
  file. If multiple indexes are created, the libraries will be named the
  \code{ref_lib} basename + "_1", "_2", etc. The function returns the names of the
  folders holding these files.
}
\description{
This function is a wrapper for the \code{Rsubread::buildindex} function. It
will generate one or more Subread indexes from a .fasta file. If the library
is too large (default >4GB) it will automatically be split into multiple
indexes, with _1, _2, etc at the end of the ref_lib basename.
}
\examples{
#### Create a subread index from the example reference library

## Create a temporary directory to store the reference library
ref_temp <- tempfile()
dir.create(ref_temp)
tmp_accession <- system.file("extdata", "example_accessions.sql", package = "MetaScope")

## Download reference genome
out_fasta <- download_refseq('Orthoebolavirus zairense', reference = FALSE,
                             representative = FALSE, out_dir = ref_temp,
                             compress = TRUE, patho_out = FALSE,
                             caching = TRUE, accession_path = tmp_accession)

## Make subread index of reference library
mk_subread_index(out_fasta)
unlink(ref_temp)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/align_target.R
\name{align_target}
\alias{align_target}
\title{Align microbiome reads to a set of reference libraries}
\usage{
align_target(
  read1,
  read2 = NULL,
  lib_dir = NULL,
  libs,
  threads = 1,
  align_file = tools::file_path_sans_ext(read1),
  subread_options = align_details,
  quiet = TRUE
)
}
\arguments{
\item{read1}{Path to the .fastq file to align.}

\item{read2}{Optional: Location of the mate pair .fastq file to align.}

\item{lib_dir}{Path to the index files for all libraries.}

\item{libs}{A vector of character strings giving the basenames of the Subread
index files for alignment. If ALL indices to be used are located in the
current working directory, set \code{lib_dir = NULL}. Default is
\code{lib_dir = NULL}.}

\item{threads}{The number of threads that can be utilized by the function.
Default is 1 thread.}

\item{align_file}{The basename of the output alignment file (without trailing
.bam extension).}

\item{subread_options}{A named \code{list} specifying alignment parameters
for the \code{Rsubread::align()} function, which is called inside
\code{align_target()}. Elements should include type, nthreads,
maxMismatches, nsubreads, phredOffset, unique, and nBestLocations.
Descriptions of these parameters are available under
\code{?Rsubread::align}. Defaults to the global \code{align_details}
object.}

\item{quiet}{Turns off most messages. Default is \code{TRUE}.}
}
\value{
This function writes a merged and sorted .bam file after aligning to
all reference libraries given, along with a summary report file, to the
user's working directory. The function also outputs the new .bam filename.
}
\description{
This is the main MetaScope target library mapping function, using Rsubread
and multiple libraries. Aligns to each library separately, filters unmapped
reads from each file, and then merges and sorts the .bam files from each
library into one output file. If desired, output can be passed to
`filter_host()` to remove reads that also map to filter library genomes.
}
\examples{
#### Align example reads to an example reference library using Rsubread
\donttest{
## Create temporary directory
target_ref_temp <- tempfile()
dir.create(target_ref_temp)

tax <- "Ovine atadenovirus D"

## Create temporary taxonomizr accession
tmp_accession <- system.file("extdata", "example_accessions.sql", package = "MetaScope")

## Download genome
all_ref <- MetaScope::download_refseq(tax,
                                      reference = FALSE,
                                      representative = FALSE,
                                      compress = TRUE,
                                      out_dir = target_ref_temp,
                                      caching = TRUE,
                                      accession_path = tmp_accession)

## Create subread index
ind_out <- mk_subread_index(all_ref)

## Get path to example reads
readPath <- system.file("extdata", "reads.fastq",
                        package = "MetaScope")
## Copy the example reads to the temp directory
refPath <- file.path(target_ref_temp, "reads.fastq")
file.copy(from = readPath, to = refPath)

## Modify alignment parameters object
data("align_details")
align_details[["type"]] <- "rna"
align_details[["phredOffset"]] <- 50
# Just to get it to align - toy example!
align_details[["maxMismatches"]] <- 100

## Run alignment
target_map <- align_target(refPath,
                           libs = stringr::str_replace_all(tax, " ", "_"),
                           lib_dir = target_ref_temp,
                           subread_options = align_details)

## Remove temporary folder
unlink(target_ref_temp, recursive = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_refseq_16S.R
\name{download_refseq_16S}
\alias{download_refseq_16S}
\title{Download RefSeq 16S rRNA Bacterial and Archael libraries}
\usage{
download_refseq_16S(out_dir, combined_name = "refseq_16S.fna")
}
\arguments{
\item{out_dir}{Character string giving the name of the directory to which
libraries should be output. **Required**}

\item{combined_name}{Name of output combined file. Default is
\code{"refseq_16S.fna"}}
}
\value{
Returns a character string of the path to the combined 16S rRNA .fna
   file
}
\description{
This function will automatically download the 16S rRNA RefSeq libraries from
\code{https://ftp.ncbi.nlm.nih.gov/refseq/TargetedLoci/} and combine them
into a single .fna file
}
\note{
This function requires the suggested packages \pkg{RCurl} and
   \pkg{R.utils}. If they are not installed, you will need to install
   them manually.
}
\examples{
\dontrun{
#### Download 16S rRNA Genomes

download_refseq_16S(out_dir = "out_dir",
  combined_name = "refseq_16S.fna")
  }
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metascope_blast.R
\name{blast_reassignment}
\alias{blast_reassignment}
\title{Reassign reads from MetaScope BLASTn alignment}
\usage{
blast_reassignment(
  metascope_blast_path,
  species_threshold,
  num_hits,
  blast_tmp_dir,
  out_dir,
  sample_name,
  reassign_validated = FALSE,
  reassign_to_validated = TRUE
)
}
\arguments{
\item{metascope_blast_path}{Character string. The filepath to a
\code{metascope_blast} CSV output file.}

\item{species_threshold}{Numeric. A number between 0 and 1 indicating
the minimum proportion of reads needed for a taxon to be considered
validated from the BLAST results. Default is 0.2, or 20\%.}

\item{num_hits}{Integer. The number of hits for which to assess validation.
Default is 10, i.e., only the top 10 taxa will be assessed.}

\item{blast_tmp_dir}{Character string. Filepath of the directory where
BLAST results were output from the \code{metascope_blast} function.
Referencing the arguments from \code{metascope_blast}, this would be
\code{file.path(tmp_dir, "blast")}}

\item{out_dir}{Character string, path to output directory.}

\item{sample_name}{Character string, sample name for output files.}

\item{reassign_validated}{Logical. Should reads from validated accessions be
reassigned to other validated accessions. Defaults to \code{FALSE}}

\item{reassign_to_validated}{Logical. Should reads only be re-assigned to
validated accessions or to any accession with more reads than the current
accession. Defaults to \code{TRUE}}
}
\value{
Returns a \code{data.frame} with the reassigned taxa and read counts.
}
\description{
Using the output from \code{metascope_blast()}, the
\code{blast_reassignment()} function takes the results and alters the
original \code{metascope_id()} output to reassign reads that were invalidated
by the BLAST findings. Currently, the implementation of this function only
reassigns reads to a taxon that was already found in the sample at a higher
abundance.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metascope_id.R
\name{locations}
\alias{locations}
\title{Helper Function for MetaScope ID}
\usage{
locations(
  which_taxid,
  which_genome,
  accessions,
  taxids,
  reads,
  out_base,
  out_dir
)
}
\arguments{
\item{which_taxid}{Which taxid to plot}

\item{which_genome}{Which genome to plot}

\item{accessions}{List of accessions from \code{metascope_id()}}

\item{taxids}{List of accessions from \code{metascope_id()}}

\item{reads}{List of reads from input file}

\item{out_base}{The basename of the input file}

\item{out_dir}{The path to the input file}
}
\value{
A plot of the read coverage for a given genome
}
\description{
Used to create plots of genome coverage for any number of accession numbers
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metascope_blast.R
\name{blastn_results}
\alias{blastn_results}
\title{Reformat BLASTn results}
\usage{
blastn_results(
  results_table,
  bam_file,
  num_results = 10,
  num_reads_per_result = 100,
  hit_list = 10,
  num_threads = 1,
  db_path,
  out_path,
  db = NULL,
  sample_name = NULL,
  quiet = quiet,
  accession_path,
  fasta_dir = NULL,
  BPPARAM
)
}
\arguments{
\item{results_table}{data.frame containing the MetaScope results.}

\item{bam_file}{\code{Rsamtools::bamFile} instance for the given sample.}

\item{num_results}{Integer; maximum number of Metascope results to BLAST.
Default is 10.}

\item{num_reads_per_result}{Integer; number of reads to BLAST per result.
Default is 100.}

\item{hit_list}{Integer; how many BLAST results to fetch for each read.
Default is 10.}

\item{num_threads}{Integer; how many threads to use if multithreading.
Default is 1.}

\item{db_path}{Character string; filepath for the location of the
pre-installed BLAST database.}

\item{out_path}{Character string; Output directory to save CSV output files,
including base name of files. For example, given a sample "X78256",
filepath would be \code{file.path(directory_here, "X78256")} with extension
omitted.}

\item{db}{Currently accepts one of \code{c("ncbi", "silva", "other")} Default
is \code{"ncbi"}, appropriate for samples aligned against indices compiled
from NCBI whole genome databases. Alternatively, usage of an alternate
database (like Greengenes2) should be specified with \code{"other"}.}

\item{sample_name}{Character string, sample name for output files.}

\item{quiet}{Logical, whether to print out more informative messages. Default
is FALSE.}

\item{accession_path}{(character) Filepath to NCBI accessions SQL
database. See \code{taxonomzr::prepareDatabase()}.}

\item{fasta_dir}{Character string; Directory where fastas from
\code{metascope_id} are stored.}

\item{BPPARAM}{An optional BiocParallelParam instance determining the
parallel back-end to be used during evaluation.}
}
\value{
Creates and exports num_results number of csv files with blast
  results from local blast
}
\description{
Reformat BLASTn results
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_refseq.R
\name{download_refseq}
\alias{download_refseq}
\title{Download RefSeq genome libraries}
\usage{
download_refseq(
  taxon,
  reference = TRUE,
  representative = FALSE,
  compress = TRUE,
  patho_out = FALSE,
  out_dir = NULL,
  caching = FALSE,
  quiet = TRUE,
  accession_path = NULL
)
}
\arguments{
\item{taxon}{Name of single taxon to download. The taxon name should be a
recognized NCBI scientific or common name, with no grammatical or
capitalization inconsistencies. All available taxonomies are visible by
accessing the \code{MetaScope:::taxonomy_table} object included in the package.}

\item{reference}{Download only RefSeq reference genomes? Defaults to
\code{TRUE}. Automatically set to \code{TRUE} if \code{representative =
TRUE}.}

\item{representative}{Download RefSeq representative and reference genomes?
Defaults to \code{FALSE}. If \code{TRUE}, reference is automatically set at
  \code{TRUE}.}

\item{compress}{Compress the output .fasta file? Defaults to \code{TRUE}.}

\item{patho_out}{Create duplicate outpute files compatible with PathoScope?
Defaults to \code{FALSE}.}

\item{out_dir}{Character string giving the name of the directory to which
libraries should be output. Defaults to creation of a new temporary
directory.}

\item{caching}{Whether to use BiocFileCache when downloading genomes.
Default is \code{FALSE}.}

\item{quiet}{Turns off most messages. Default is \code{TRUE}.}

\item{accession_path}{(character) Filepath to NCBI accessions SQL
database. See \code{taxonomzr::prepareDatabase()}.}
}
\value{
Returns a .fasta or .fasta.gz file of the desired RefSeq genomes.
This file is named after the kingdom selected and saved to the current
directory (e.g. 'bacteria.fasta.gz'). This function also has the option
to return a .fasta file formatted for PathoScope as well
(e.g. bacteria.pathoscope.fasta.gz') if \code{path_out = TRUE}.
}
\description{
This function will automatically download RefSeq genome libraries in a fasta
format from the specified taxon. The function will first download the
summary report at:
\code{ftp://ftp.ncbi.nlm.nih.gov/genomes/refseq/**kingdom**/assembly_summary.txt},
and then use this file to download the genome(s) and combine them in a
single compressed or uncompressed .fasta file.
}
\details{
When selecting the \code{taxon} to be downloaded, if you receive an error
saying \code{Your input is not a valid taxon}, please take a look at the
\code{taxonomy_table} object, which can be accessed with the command
\code{MetaScope:::taxonomy_table)}. Only taxa with exact spelling as they appear
at any level of the table will be acknowledged.
}
\examples{
#### Download RefSeq genomes

## Create temporary taxonomizr accession
tmp_accession <- system.file("extdata", "example_accessions.sql", package = "MetaScope")

## Download all RefSeq reference Bovismacovirus genus genomes
download_refseq('Bovismacovirus', reference = FALSE, representative = FALSE,
                out_dir = NULL, compress = TRUE, patho_out = FALSE,
                caching = TRUE, accession_path = tmp_accession)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_children.R
\name{get_children}
\alias{get_children}
\title{Get child nodes from NCBI taxonomy}
\usage{
get_children(input_taxon, input_rank, tax_dat = NULL)
}
\arguments{
\item{input_taxon}{The parent taxon.}

\item{input_rank}{The taxonomic rank of the input taxon.}

\item{tax_dat}{A dataframe of organism classification information.
At minimum, should have a column indicating "strain", and and all others
should be taxonomic ranks. Each row should be a taxonomic relationship.
This defaults to \code{NULL}, which calls the `taxonomy_table` object.}
}
\value{
Returns a vector of all the child species and/or strains of the
input taxon.
}
\description{
This function will utilize an organism classification table to obtain all
children species and/or strains with available NCBI reference sequences given
a parent taxon and its rank.
}
\examples{
## Get all child species and strains in bacteria superkingdom
get_children('Bacteria','superkingdom')

## Get all child species and strains in fungi kingdom
get_children('Fungi', 'kingdom')

## Get all child species in primate order
get_children('Primates', 'order')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MetaScope-package.R
\docType{package}
\name{MetaScope-package}
\alias{MetaScope}
\alias{MetaScope-package}
\title{MetaScope: Tools and functions for preprocessing 16S and metagenomic sequencing microbiome data}
\description{
This package contains tools and methods for preprocessing microbiome data. Functionality includes library generation, demultiplexing, alignment, and microbe identification. It is in part an R translation of the PathoScope 2.0 pipeline.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/wejlab/metascope https://wejlab.github.io/metascope-docs/}
  \item Report bugs at \url{https://github.com/wejlab/MetaScope/issues}
}

}
\author{
\strong{Maintainer}: Sean Lu \email{seanlu96@gmail.com} (\href{https://orcid.org/0009-0007-8005-6125}{ORCID})

Authors:
\itemize{
  \item Aubrey Odom \email{aodom@bu.edu} (\href{https://orcid.org/0000-0001-7113-7598}{ORCID})
  \item Rahul Varki \email{rvarki@bu.edu} (\href{https://orcid.org/0009-0003-5721-9484}{ORCID})
  \item W. Evan Johnson \email{wej@bu.edu} (\href{https://orcid.org/0000-0002-6247-6595}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_accessions.R
\name{download_accessions}
\alias{download_accessions}
\title{Download indexes required for MetaScope ID and MetaBlast modules}
\usage{
download_accessions(
  ind_dir,
  tmp_dir = file.path(ind_dir, "tmp"),
  remove_tmp_dir = TRUE,
  NCBI_accessions_database = TRUE,
  NCBI_accessions_name = "accessionTaxa",
  silva_taxonomy_database = TRUE,
  silva_taxonomy_name = "all_silva_headers",
  blast_16S_database = TRUE,
  blast_16S_name = "16S_ribosomal_RNA"
)
}
\arguments{
\item{ind_dir}{Character string. Directory filepath where indices should be
saved. Required.}

\item{tmp_dir}{Character path to directory for storing temp files. (Useful
to avoid redownloading) Defaults to \code{file.path(ind_dir, "tmp")}}

\item{remove_tmp_dir}{Delete tmp_dir after downloads are complete? Defaults
to \code{TRUE}}

\item{NCBI_accessions_database}{Logical. Download taxonomizr NCBI accessions
database? Defaults to \code{TRUE}.}

\item{NCBI_accessions_name}{Character string. Filename (with or without
extension) to save taxonomizr NCBI accessions database. Defaults to
\code{"accessionTaxa.sql"}.}

\item{silva_taxonomy_database}{Logical. Download SILVA taxonomy database?
Defaults to \code{TRUE}.}

\item{silva_taxonomy_name}{Character string. Filename (with or without
extension) to save SILVA taxonomy database.
Defaults to the file supplied with the package,
\code{"all_silva_headers.rds"}.}

\item{blast_16S_database}{Logical. Download BLAST 16S database?
Defaults to \code{TRUE}.}

\item{blast_16S_name}{Character string. Filename (without extension) to save \
BLAST 16S database.
Defaults to the file supplied with the package,
\code{"16S_ribosomal_RNA"}.}
}
\value{
Exports database(s) with names and to location specified by the user.
}
\description{
This is a necessary step for all samples utilizing NCBI and SILVA databases
in the MetaScope pipeline. As specified by the user,
\code{download_accessions} will automatically download the NCBI accessions
database, the SILVA taxonomy database, and or the NCBI Blast 16S database and
prepare consolidated databases for downstream use with the MetaID and
MetaBLAST modules. This package relies on the \code{taxonomizr} package.
}
\examples{
\dontrun{
  download_accessions(
    ind_dir = "C:/Users/JohnSmith/Research",
    tmp_dir = file.path(ind_dir, "tmp"),
    remove_tmp_dir = TRUE,
    NCBI_accessions_database = TRUE,
    NCBI_accessions_name = "accessionTaxa.sql",
    silva_taxonomy_database = TRUE,
    silva_taxonomy_name = "all_silva_headers.rds")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metascope_blast.R
\name{blastn_single_result}
\alias{blastn_single_result}
\title{blastn_single_result}
\usage{
blastn_single_result(
  results_table,
  bam_file,
  which_result,
  num_reads = 100,
  hit_list = 10,
  num_threads,
  db_path,
  quiet,
  accession_path,
  bam_seqs,
  out_path,
  sample_name,
  fasta_dir = NULL
)
}
\arguments{
\item{results_table}{A dataframe of the Metascope results}

\item{bam_file}{A sorted bam file and index file, loaded with
Rsamtools::bamFile}

\item{which_result}{Index in results_table for which result to Blast search}

\item{num_reads}{Number of reads to blast per result}

\item{hit_list}{Number of how many blast results to fetch per read}

\item{num_threads}{Number of threads if multithreading}

\item{db_path}{Blast database path}

\item{quiet}{Logical, whether to print out more informative messages. Default
is FALSE.}

\item{accession_path}{(character) Filepath to NCBI accessions SQL
database. See \code{taxonomzr::prepareDatabase()}.}

\item{bam_seqs}{A list of the sequence IDs from the bam file}

\item{out_path}{Path to output results.}

\item{sample_name}{Character string, sample name for output files.}

\item{fasta_dir}{Path to where fasta files are stored.}
}
\value{
Returns a dataframe of blast results for a metascope result
}
\description{
blastn_single_result
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mk_bowtie_index.R
\name{mk_bowtie_index}
\alias{mk_bowtie_index}
\title{Make a Bowtie2 index}
\usage{
mk_bowtie_index(
  ref_dir,
  lib_dir,
  lib_name,
  bowtie2_build_options,
  threads = 1,
  overwrite = FALSE
)
}
\arguments{
\item{ref_dir}{The path to the directory that contains the reference files
either uncompressed or compressed (.gz). NOTE: This directory should
contain only the reference fasta files to be indexed.}

\item{lib_dir}{The path to the directory where Bowtie2 index files should
be created.}

\item{lib_name}{The basename of the index file to be created (without the
.bt2 or .bt2l extension)}

\item{bowtie2_build_options}{Optional: Options that can be passed to the
mk_bowtie_index() function. All options should be passed as one string. To
see all the available options that can be passed to the function use
Rbowtie2::bowtie2_build_usage(). NOTE: Do not specify threads here.}

\item{threads}{The number of threads available to the function. Default is 1
thread.}

\item{overwrite}{Whether existing files should be overwritten. Default is
FALSE.}
}
\value{
Creates the Bowtie2 indexes of the supplied reference .fasta files.
  Returns the path to the directory containing these files.
}
\description{
This function is a wrapper for the \code{Rbowtie2::bowtie2_build} function.
It will create either small (.bt2) or large Bowtie2 indexes (.bt2l)
depending on the combined size of the reference fasta files.
}
\examples{
#### Create a bowtie index from the example reference library

## Create a temporary directory to store the reference library
ref_temp <- tempfile()
dir.create(ref_temp)

tmp_accession <- system.file("extdata", "example_accessions.sql", package = "MetaScope")

## Download reference genome
download_refseq('Bovismacovirus', reference = FALSE, representative = FALSE,
                out_dir = ref_temp, compress = TRUE, patho_out = FALSE,
                caching = TRUE, accession_path = tmp_accession)

## Create the reference library index files in the current directory
mk_bowtie_index(ref_dir = ref_temp, lib_dir = ref_temp,
                lib_name = "target", threads = 1, overwrite = FALSE)

## Remove temporary directory
unlink(ref_temp, recursive = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter_host.R
\name{filter_host}
\alias{filter_host}
\title{Use Rsubread to align reads against one or more filter libraries and
subsequently remove mapped reads}
\usage{
filter_host(
  reads_bam,
  lib_dir = NULL,
  libs,
  make_bam = FALSE,
  output = paste(tools::file_path_sans_ext(reads_bam), "filtered", sep = "."),
  subread_options = align_details,
  YS = 1e+05,
  threads = 1,
  quiet = TRUE
)
}
\arguments{
\item{reads_bam}{The name of a merged, sorted .bam file that has previously
been aligned to a reference library. Likely, the output from running an
instance of \code{align_target()}.}

\item{lib_dir}{Path to the directory that contains the filter Subread index
files.}

\item{libs}{The basename of the filter libraries (without index
extension).}

\item{make_bam}{Logical, whether to also output a bam file with host reads
filtered out. A .csv.gz file will be created instead if \code{FALSE}.
Creating a bam file is costly on resources over creating a compressed
csv file with only relevant information, so default is \code{FALSE}.}

\item{output}{The desired name of the output .bam or .csv.gz file. Extension is
automatically defined by whether \code{make_bam = TRUE.} Default is the
basename of \code{unfiltered_bam} + \code{.filtered} + extension.}

\item{subread_options}{A named \code{list} specifying alignment parameters
for the \code{Rsubread::align()} function, which is called inside
\code{align_target()}. Elements should include type, nthreads,
maxMismatches, nsubreads, phredOffset, unique, and nBestLocations.
Descriptions of these parameters are available under
\code{?Rsubread::align}. Defaults to the global \code{align_details}
object.}

\item{YS}{yieldSize, an integer. The number of alignments to be read in from
the bam file at once for chunked functions. Default is 100000.}

\item{threads}{The amount of threads available for the function. Default is 1
thread.}

\item{quiet}{Turns off most messages. Default is \code{TRUE}.}
}
\value{
The name of a filtered, sorted .bam file written to the user's
  current working directory. Or, if \code{make_bam = FALSE}, a .csv.gz file
  containing a data frame of only requisite information to run
  \code{metascope_id()}.
}
\description{
After aligning your sample to a target library with \code{align_target()},
use \code{filter_host()} to remove unwelcome host contamination using filter
reference libraries. This function takes as input the name of the .bam file
produced via \code{align_target()}, and produces a sorted .bam file with any
reads that match the filter libraries removed. This resulting .bam file may
be used upstream for further analysis. This function uses Rsubread. For the
Rbowtie2 equivalent of this function, see \code{filter_host_bowtie}.
}
\details{
A compressed .csv can be created to produce a smaller output file that is
created more efficiently and is still compatible with \code{metascope_id()}.
}
\examples{
#### Filter reads from bam file that align to any of the filter libraries

## Assuming a bam file has been created previously with align_target()
\donttest{
## Create temporary directory
filter_ref_temp <- tempfile()
dir.create(filter_ref_temp)

## Create temporary taxonomizr accession
tmp_accession <- system.file("extdata", "example_accessions.sql", package = "MetaScope")


## Download filter genome
all_species <- c("Staphylococcus aureus subsp. aureus str. Newman")
all_ref <- vapply(all_species, MetaScope::download_refseq,
                  reference = FALSE, representative = FALSE, compress = TRUE,
                  out_dir = filter_ref_temp, caching = FALSE,
                  accession_path = tmp_accession,
                  FUN.VALUE = character(1))
ind_out <- vapply(all_ref, mk_subread_index, FUN.VALUE = character(1))

## Get path to example reads
readPath <- system.file("extdata", "subread_target.bam",
                        package = "MetaScope")

## Copy the example reads to the temp directory
refPath <- file.path(filter_ref_temp, "subread_target.bam")
file.copy(from = readPath, to = refPath)
utils::data("align_details")
align_details[["type"]] <- "rna"
align_details[["phredOffset"]] <- 10
# Just to get it to align - toy example!
align_details[["maxMismatches"]] <- 10

## Align and filter reads
filtered_map <- filter_host(
  refPath, lib_dir = filter_ref_temp,
  libs = stringr::str_replace_all(all_species, " ", "_"),
  threads = 1, subread_options = align_details)

## Remove temporary directory
unlink(filter_ref_temp, recursive = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metascope_blast.R
\name{find_taxids}
\alias{find_taxids}
\title{Gets Taxonomy IDS from accessions names
Returns a dataframe with the original accession, the associated taxonomy ID
and associated taxonomy species}
\usage{
find_taxids(all_fastas, accession_path)
}
\arguments{
\item{all_fastas}{Biostrings loaded object of all fastas}

\item{accession_path}{Path to taxonomizr accessions database}
}
\value{
Data.frame of taxonomy Ids and Species name
}
\description{
Gets Taxonomy IDS from accessions names
Returns a dataframe with the original accession, the associated taxonomy ID
and associated taxonomy species
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metascope_blast.R
\name{get_multi_seqs}
\alias{get_multi_seqs}
\title{Gets multiple sequences from different accessions in a bam file}
\usage{
get_multi_seqs(ids_n, bam_file, seq_info_df, metascope_id_tax, sorted_bam_file)
}
\arguments{
\item{ids_n}{List of vectors with Taxonomy IDs and the number of sequences to get from each}

\item{bam_file}{A sorted bam file and index file, loaded with
Rsamtools::bamFile}

\item{seq_info_df}{Dataframe of sequence information from metascope_blast()}

\item{metascope_id_tax}{Data.frame of taxonomy information}

\item{sorted_bam_file}{Filepath to sorted bam file}
}
\value{
Biostrings format sequences
}
\description{
Returns fasta sequences from a bam file with given taxonomy IDs
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metascope_blast.R
\name{metascope_blast}
\alias{metascope_blast}
\title{Blast reads from MetaScope aligned files}
\usage{
metascope_blast(
  metascope_id_path,
  bam_file_path = list.files(tmp_dir, ".updated.bam$", full.names = TRUE)[1],
  tmp_dir,
  out_dir,
  sample_name,
  fasta_dir = NULL,
  num_results = 10,
  num_reads = 100,
  hit_list = 10,
  num_threads = 1,
  db_path,
  quiet = FALSE,
  db = NULL,
  accession_path = NULL
)
}
\arguments{
\item{metascope_id_path}{Character string; path to a csv file output by
`metascope_id()`.}

\item{bam_file_path}{Character string; full path to bam file for the same
sample processed by `metascope_id`. Note that the `filter_bam` function
must have output a bam file, and not a .csv.gz file. See
`?filter_bam_bowtie` for more details. Defaults to
\code{list.files(file_temp, ".updated.bam$")[1]}.}

\item{tmp_dir}{Character string, a temporary directory in which to host
files.}

\item{out_dir}{Character string, path to output directory.}

\item{sample_name}{Character string, sample name for output files.}

\item{fasta_dir}{Directory where fasta files for blast will be stored.}

\item{num_results}{Integer, the maximum number of taxa from the metascope_id
output to check reads. Takes the top n taxa, i.e. those with largest
abundance. Default is 10.}

\item{num_reads}{Integer, the maximum number of reads to blast per microbe.
If the true number of reads assigned to a given taxon is smaller, then the
smaller number will be chosen. Default is 100. Too many reads will involve
more processing time.}

\item{hit_list}{Integer, number of blast hit results to keep. Default is 10.}

\item{num_threads}{Integer, number of threads if running in parallel
(recommended). Default is 1.}

\item{db_path}{Character string. The database file to be searched (including
basename, but without file extension). For example, if the nt database is
in the nt folder, this would be /filepath/nt/nt assuming that the database
files have the nt basename. Check this path if you get an error message
stating "No alias or index file found".}

\item{quiet}{Logical, whether to print out more informative messages. Default
is FALSE.}

\item{db}{Currently accepts one of \code{c("ncbi", "silva", "other")} Default
is \code{"ncbi"}, appropriate for samples aligned against indices compiled
from NCBI whole genome databases. Alternatively, usage of an alternate
database (like Greengenes2) should be specified with \code{"other"}.}

\item{accession_path}{(character) Filepath to NCBI accessions SQL
database. See \code{taxonomzr::prepareDatabase()}.}
}
\value{
This function writes an updated csv file with metrics.
}
\description{
This function allows the user to check a subset of identified reads against
NCBI BLAST and the nucleotide database to confirm or contradict results
provided by MetaScope. It aligns the top `metascope_id()` results to NCBI
BLAST database. It REQUIRES that command-line BLAST and a separate nucleotide
database have already been installed on the host machine. It returns a csv
file updated with BLAST result metrics.
}
\details{
This function assumes that you used the NCBI nucleotide database to process
samples, with a download date of 2021 or later. This is necessary for
compatibility with the bam file headers.

This is highly computationally intensive and should be ran with multiple
cores, submitted as a multi-threaded computing job if possible.

Note, if best_hit_strain is FALSE, then no strain was observed more often
among the BLAST results.
}
\examples{
\dontrun{
### Create temporary directory
file_temp <- tempfile()
dir.create(file_temp)

bamPath <- system.file("extdata", "bowtie_target.bam",
                       package = "MetaScope")
file.copy(bamPath, file_temp)

metascope_id(file.path(file_temp, "bowtie_target.bam"), aligner = "bowtie2",
             input_type = "bam", out_dir = file_temp, num_species_plot = 0,
             update_bam = TRUE)

## Run metascope blast
### Get export name and metascope id results
out_base <- bamPath |> basename() |> tools::file_path_sans_ext() |>
  tools::file_path_sans_ext()
metascope_id_path <- file.path(file_temp, paste0(out_base, ".metascope_id.csv"))

# NOTE: change db_path to the location where your BLAST database is stored!
db <- "/restricted/projectnb/pathoscope/data/blastdb/nt/nt"

tmp_accession <- system.file("extdata", "example_accessions.sql", package = "MetaScope"

metascope_blast(metascope_id_path,
                bam_file_path = file.path(file_temp, "bowtie_target.bam"),
                tmp_dir = file_temp,
                out_dir = file_temp,
                sample_name = out_base,
                db_path = db,
                num_results = 10,
                num_reads = 5,
                hit_list = 10,
                num_threads = 3,
                db = "ncbi",
                quiet = FALSE,
                fasta_dir = NULL,
                accession_path = tmp_accession)

## Remove temporary directory
unlink(file_temp, recursive = TRUE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{bt2_16S_params}
\alias{bt2_16S_params}
\title{A universal parameter object for Bowtie 2 16S alignment}
\format{
list
}
\usage{
data(bt2_16S_params)
}
\description{
This character string provides several Bowtie 2 options to provide an
optimized alignment specifically optimized for 16S amplicon sequencing data.
This object can be used with functions that use the Bowtie 2 aligner through
the \code{Rbowtie2} package, namely \code{align_target_bowtie()} and
\code{filter_host_bowtie}. These settings can be substituted for default
settings by passing \code{} to the \code{bowtie2_options} argument.
}
\details{
The default parameters listed in this object are
"--local -R 2 -N 0 -L 25 -i S,1,0.75 -k 5 --score-min L,0,1.88"

Note that k is actually 10 and is doubled internally from 5.
The \code{score-min} function was chosen such that the minimum alignment score
allowed requires 98% identity.

Further delineation of Bowtie 2 parameters is provided in the
\href{https://bowtie-bio.sourceforge.net/bowtie2/manual.shtml}{Bowtie 2
manual}.
}
\examples{
data("bt2_16S_params")

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metascope_inspect.R
\name{metascope_inspect}
\alias{metascope_inspect}
\title{Inspect a Bowtie index}
\usage{
metascope_inspect(bt2_base, ...)
}
\arguments{
\item{bt2_base}{The path to the Bowtie2 index files minus minus trailing
.1.bt2/.2.bt2 extension names.}

\item{...}{Required additional arguments to be passed on to the \code{bowtie2-inspect}
binaries. See below for details.}
}
\description{
This function can be use to call the \code{bowtie2-inspect} wrapper which
wraps the \code{bowtie2-inspect-s} and \code{bowtie2-inspect-l} binaries.
}
\details{
All additional arguments in ... are intepreted as
parameters to be passed on to \code{bowtie2-inspect} wrapper. All of them
\code{Character} or \code{Numeric} scalar. You can put all additional
arguments in one \code{Character} (e.g. "--across 60 --names") with white
space separation, or put them in different \code{Character}
(e.g. "--across","60","--names"). See the output of
\code{metascope_inspect_usage()} for details about available parameters.
}
\examples{
\dontrun{
metascope_inspect(bt2_base)
}
}
\references{
{
Langmead B, Wilks C, Antonescu V, Charles R. Scaling
read aligners to hundreds of threads on general-purpose processors.
Bioinformatics. 2018 Jul 18. doi: 10.1093/bioinformatics/bty648.

Langmead B, Salzberg SL. Fast gapped-read alignment with
Bowtie 2. Nature Methods. 2012 Mar 4;9(4):357-9. doi: 10.1038/nmeth.1923.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/demultiplex.R
\name{extract_reads}
\alias{extract_reads}
\title{Helper function for demultiplexing}
\usage{
extract_reads(
  barcodeIndex,
  barcodes,
  sampleNames,
  index,
  reads,
  location = "./demultiplex_fastq",
  rcBarcodes = TRUE,
  hDist = 0,
  quiet = TRUE
)
}
\arguments{
\item{barcodeIndex}{Which barcode (integer number or index) in the barcodes
or sample name to use for read extraction.}

\item{barcodes}{A list of all barcodes in the sequencing dataset. Correlates
and in same order as \code{sampleNames}.}

\item{sampleNames}{A list of sample names or identifiers associated with each
barcode in the barcodes list.}

\item{index}{A \code{Biostrings::DNAStringSet} that contains the read headers
and barcode sequence for each header in the sequence slot.}

\item{reads}{A \code{Biostrings::QualityScaledXStringSet} that has the same
headers and order as the index file, but contains the read sequences and
their quality scores.}

\item{location}{A directory location to store the demultiplexed read files.
Defaults to generate a new subdirectory at './demultiplex_fastq'}

\item{rcBarcodes}{Should the barcode indices in the barcodes list be reverse
complemented to match the sequences in the index DNAStringSet? Defaults to
\code{TRUE}.}

\item{hDist}{Uses a Hamming Distance or number of base differences to allow
for inexact matches for the barcodes/indexes. Defaults to 0. Warning: if
the Hamming Distance is >=1 and this leads to inexact index matches to more
than one barcode, that read will be written to more than one demultiplexed
read files.}

\item{quiet}{Turns off most messages. Default is \code{TRUE}.}
}
\value{
Writes a single .fastq file that contains all reads whose index
  matches the barcode specified. This file will be written to the location
  directory, and will be named based on the specified sampleName and barcode,
  e.g. './demultiplex_fastq/SampleName1_GGAATTATCGGT.fastq.gz'
}
\description{
Helper function for demultiplexing sequencing reads, designed in a way to
allow for parallelization across barcodes (parallel extraction of reads by
barcode). This function takes a specific barcode (numeric index) from lists
of sample names/barcodes, a \code{Biostrings::DNAStringSet} of barcodes by
sequence header, and a \code{Biostrings::QualityScaledXStringSet} of reads
corresponding to the barcodes. Based on the barcode index given, it extracts
all reads for the indexed barcode and writes all the reads from that barcode
to a separate .fastq file.
}
\examples{

## Create temporary directory
ref_temp <- tempfile()
dir.create(ref_temp)

## Load example barcode, index, and read data into R session
barcodePath <- system.file("extdata", "barcodes.txt", package = "MetaScope")
bcFile <- read.table(barcodePath, sep = "\t", header = TRUE)

indexPath <- system.file("extdata", "virus_example_index.fastq",
package = "MetaScope")
inds <- Biostrings::readDNAStringSet(indexPath, format = "fastq")

readPath <- system.file("extdata", "virus_example.fastq",
                        package = "MetaScope")
reads <- Biostrings::readQualityScaledDNAStringSet(readPath)

## Extract reads from the first barcode
results <- extract_reads(1, bcFile[, 2], bcFile[, 1], inds, reads,
                        rcBarcodes = FALSE, location = ref_temp)

## Extract reads from multiple barcodes
more_results <- lapply(1:6, extract_reads, bcFile[, 2], bcFile[, 1], inds,
                       reads, rcBarcodes = FALSE, location = ref_temp)

## Remove temporary directory
unlink(ref_temp, recursive = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metascope_blast.R
\name{add_in_taxa_ncbi}
\alias{add_in_taxa_ncbi}
\title{Adds in taxa if input used NCBI database}
\usage{
add_in_taxa_ncbi(metascope_id_in, accession, BPPARAM)
}
\arguments{
\item{metascope_id_in}{MetaScope ID file with NCBI taxa qnames}

\item{BPPARAM}{An optional BiocParallelParam instance determining the
parallel back-end to be used during evaluation.}
}
\value{
data.frame or tibble of taxonomy information
}
\description{
Returns MetaScope Table with NCBI taxa in separate columns
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metascope_id.R
\name{metascope_id}
\alias{metascope_id}
\title{Identify which genomes are represented in a processed sample}
\usage{
metascope_id(
  input_file,
  input_type = "csv.gz",
  aligner = "bowtie2",
  db = "ncbi",
  db_feature_table = NULL,
  accession_path = NULL,
  priors_df = NULL,
  tmp_dir = dirname(input_file),
  out_dir = dirname(input_file),
  convEM = 1/10000,
  maxitsEM = 25,
  update_bam = FALSE,
  num_species_plot = NULL,
  group_by_taxa = "species",
  quiet = TRUE
)
}
\arguments{
\item{input_file}{The .bam or .csv.gz file of sample reads to be identified.}

\item{input_type}{Extension of file input. Should be either "bam" or
"csv.gz". Default is "csv.gz".}

\item{aligner}{The aligner which was used to create the bam file. Default is
"bowtie2" but can also be set to "subread" or "other".}

\item{db}{Currently accepts one of \code{c("ncbi", "silva", "other")} Default
is \code{"ncbi"}, appropriate for samples aligned against indices compiled
from NCBI whole genome databases. Alternatively, usage of an alternate
database (like Greengenes2) should be specified with \code{"other"}.}

\item{db_feature_table}{If \code{db = "other"}, a data.frame must be supplied
with two columns, "Feature ID" matching the names of the alignment indices,
and a second \code{character} column supplying the taxon identifying
information.}

\item{accession_path}{(character) Filepath to NCBI accessions SQL database.
See \code{taxonomzr::prepareDatabase()}.}

\item{priors_df}{data.frame containing priors data. The data.frame consists
of two columns, 'species' containing species name, and 'prior_weights'
containing the prior weights (as a percent; integer).}

\item{tmp_dir}{Path to a directory to which bam and updated bam files can be
saved. Required.}

\item{out_dir}{The directory to which the .csv output file will be output.
Defaults to \code{dirname(input_file)}.}

\item{convEM}{The convergence parameter of the EM algorithm. Default set at
\code{1/10000}.}

\item{maxitsEM}{The maximum number of EM iterations, regardless of whether
the convEM is below the threshhold. Default set at \code{50}. If set at
\code{0}, the algorithm skips the EM step and summarizes the .bam file 'as
is'.}

\item{update_bam}{Whether to update BAM file with new read assignments.
Default is \code{FALSE}. If \code{TRUE}, requires \code{input_type = "bam"}
such that a BAM file is the input to the function.}

\item{num_species_plot}{The number of genome coverage plots to be saved.
Default is \code{NULL}, which saves coverage plots for the ten most highly
abundant species.}

\item{group_by_taxa}{Character. Taxonomy level at which accessions should be
grouped. Defaults to \code{"species"}}

\item{quiet}{Turns off most messages. Default is \code{TRUE}.}
}
\value{
This function exports a .csv file with annotated read counts to
  genomes with mapped reads to the location returned by the function.
  Depending on the parameters specified, can also output an updated BAM
  file, and fasta files for additional analysis downstream.
}
\description{
This function will read in a .bam or .csv.gz file, annotate the taxonomy and
genome names, reduce the mapping ambiguity using a mixture model, and output
a .csv file with the results. Currently, it assumes that the genome
library/.bam files use NCBI accession names for reference names (rnames in
.bam file).
}
\examples{
#### Align reads to reference library and then apply metascope_id()
## Assuming filtered bam files already exist

## Create temporary directory
file_temp <- tempfile()
dir.create(file_temp)

## Get temporary accessions database
tmp_accession <- system.file("extdata", "example_accessions.sql", package = "MetaScope")

#### Subread aligned bam file

## Create object with path to filtered subread csv.gz file
filt_file <- "subread_target.filtered.csv.gz"
bamPath <- system.file("extdata", filt_file, package = "MetaScope")
file.copy(bamPath, file_temp)

## Run metascope id with the aligner option set to subread
metascope_id(input_file = file.path(file_temp, filt_file),
             aligner = "subread", num_species_plot = 0,
             input_type = "csv.gz", accession_path = tmp_accession)

#### Bowtie 2 aligned .csv.gz file

## Create object with path to filtered bowtie2 bam file
bowtie_file <- "bowtie_target.filtered.csv.gz"
bamPath <- system.file("extdata", bowtie_file, package = "MetaScope")
file.copy(bamPath, file_temp)

## Run metascope id with the aligner option set to bowtie2
metascope_id(file.path(file_temp, bowtie_file), aligner = "bowtie2",
             num_species_plot = 0, input_type = "csv.gz",
             accession_path = tmp_accession)

## Remove temporary directory
unlink(file_temp, recursive = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metascope_id.R
\name{count_matches}
\alias{count_matches}
\title{Count the number of base lengths in a CIGAR string for a given operation}
\usage{
count_matches(x, char = "M")
}
\arguments{
\item{x}{Character. A CIGAR string for a read to be parsed. Examples of
possible operators include "M", "D", "I", "S", "H", "=", "P", and "X".}

\item{char}{A single letter representing the operation to total for the
given string.}
}
\value{
an integer number representing the total number of alignment
operations for the read that was summarized by the input CIGAR string.
}
\description{
The 'CIGAR' (Compact Idiosyncratic Gapped Alignment Report) string is how the
SAM/BAM format represents spliced alignments. This function will accept a
CIGAR string for a single read and a single character indicating the
operation to be parsed in the string. An operation is a type of column that
appears in the alignment, e.g. a match or gap. The integer following the
operator specifies a number of consecutive operations. The
\code{count_matches()} function will identify all occurrences of the operator
in the string input, add them, and return an integer number representing the
total number of operations for the read that was summarized by the input
CIGAR string.
}
\details{
This function is best used on a vector of CIGAR strings using an apply
function (see examples).
}
\examples{
# A single cigar string: 3M + 3M + 5M
cigar1 <- "3M1I3M1D5M"
count_matches(cigar1, char = "M")

# Parse with operator "P": 2P
cigar2 <- "4M1I2P9M"
count_matches(cigar2, char = "P")

# Apply to multiple strings: 1I + 1I + 5I
cigar3 <- c("3M1I3M1D5M", "4M1I1P9M", "76M13M5I")
vapply(cigar3, count_matches, char = "I",
       FUN.VALUE = numeric(1))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metascope_blast.R
\name{blast_result_metrics}
\alias{blast_result_metrics}
\title{Calculates result metrics from a blast results table}
\usage{
blast_result_metrics(blast_results_table_path, accession_path, db = NULL)
}
\arguments{
\item{blast_results_table_path}{path for blast results csv file}

\item{accession_path}{(character) Filepath to NCBI accessions SQL
database. See \code{taxonomzr::prepareDatabase()}.}

\item{db}{Currently accepts one of \code{c("ncbi", "silva", "other")} Default
is \code{"ncbi"}, appropriate for samples aligned against indices compiled
from NCBI whole genome databases. Alternatively, usage of an alternate
database (like Greengenes2) should be specified with \code{"other"}.}
}
\value{
a vector with best_hit, uniqueness_score, species_percentage_hit
  genus_percentage_hit, species_contaminant_score, and
  genus_contaminant_score
}
\description{
This function calculates the best hit (genome with most blast read hits),
uniqueness score (total number of genomes hit), species percentage hit
(percentage of reads where MetaScope species also matched the blast hit
species), genus percentage hit (percentage of reads where blast genus matched
MetaScope aligned genus) and species contaminant score (percentage of reads
that blasted to other species genomes) and genus contaminant score
(percentage of reads that blasted to other genus genomes)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metascope_inspect.R
\name{metascope_inspect_usage}
\alias{metascope_inspect_usage}
\title{Print available arguments that can be passed to metascope_inspect()}
\usage{
metascope_inspect_usage()
}
\value{
Information about available arguments that can be passed to
metascope_inspect()
}
\description{
Calling metascope_inspect_usage() prints the available arguments
that can be passed to the ... argument of the metascope_inspect() function of
the package. Note that some arguments are invalid if they are already
handled as explicit function arguments.
}
\examples{
metascope_inspect_usage()
}
\references{
{
Langmead B, Wilks C, Antonescu V, Charles R. Scaling
read aligners to hundreds of threads on general-purpose processors.
Bioinformatics. 2018 Jul 18. doi: 10.1093/bioinformatics/bty648.

Langmead B, Salzberg SL. Fast gapped-read alignment with
Bowtie 2. Nature Methods. 2012 Mar 4;9(4):357-9. doi: 10.1038/nmeth.1923.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_animalcules_patho.R
\name{convert_animalcules_patho}
\alias{convert_animalcules_patho}
\title{Create a multi-assay experiment from PathoScope 2.0 output for usage with
animalcules}
\usage{
convert_animalcules_patho(
  patho_counts,
  annot_path,
  which_annot_col,
  end_string = "-sam-report.tsv"
)
}
\arguments{
\item{patho_counts}{Character string, a directory filepath to the counts ID CSVs output by
\code{metascope_id()}.}

\item{annot_path}{The filepath to the CSV annotation file for the samples.}

\item{which_annot_col}{The name of the column of the annotation file
containing the sample IDs. These should be the same as the
\code{meta_counts} root filenames.}

\item{end_string}{The end string used at the end of the metascope_id files.
Default is "-sam-report.tsv".}
}
\value{
Returns a MultiAssay Experiment file of combined sample counts data.
The MultiAssay Experiment will have a counts assay ("MGX").
}
\description{
This function serves as a legacy integration method for usage with
PathoScope 2.0 outputs. Upon completion of the PathoScope 2.0 pipeline, users can analyze and visualize
abundances in their samples using the animalcules package. After running this function, the user should save the returned MAE
to an RDS file using a function like \code{saveRDS} to upload the output into
the \code{animalcules} package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter_host.R
\name{filter_host_bowtie}
\alias{filter_host_bowtie}
\title{Use Rbowtie2 to align reads against one or more filter libraries and
subsequently remove mapped reads}
\usage{
filter_host_bowtie(
  reads_bam,
  lib_dir,
  libs,
  make_bam = FALSE,
  output = paste(tools::file_path_sans_ext(reads_bam), "filtered", sep = "."),
  bowtie2_options = NULL,
  YS = 1e+05,
  threads = 1,
  overwrite = FALSE,
  quiet = TRUE
)
}
\arguments{
\item{reads_bam}{The name of a merged, sorted .bam file that has previously
been aligned to a reference library. Likely, the output from running an
instance of \code{align_target_bowtie()}.}

\item{lib_dir}{Path to the directory that contains the filter Bowtie2 index
files.}

\item{libs}{The basename of the filter libraries (without .bt2 or .bt2l
extension).}

\item{make_bam}{Logical, whether to also output a bam file with host reads
filtered out. A .csv.gz file will be created instead if \code{FALSE}.
Creating a bam file is costly on resources over creating a compressed
csv file with only relevant information, so default is \code{FALSE}.}

\item{output}{The desired name of the output .bam or .csv.gz file. Extension is
automatically defined by whether \code{make_bam = TRUE.} Default is the
basename of \code{unfiltered_bam} + \code{.filtered} + extension.}

\item{bowtie2_options}{Optional: Additional parameters that can be passed to
the filter_host_bowtie() function. To see all the available parameters use
Rbowtie2::bowtie2_usage(). See Details for default parameters.
NOTE: Users should pass all their parameters as
one string and if optional parameters are given then the user is
responsible for entering all the parameters to be used by Bowtie2.
The only parameters that should NOT be specified here is the threads.}

\item{YS}{yieldSize, an integer. The number of alignments to be read in from
the bam file at once for chunked functions. Default is 100000.}

\item{threads}{The amount of threads available for the function. Default is 1
thread.}

\item{overwrite}{Whether existing files should be overwritten. Default is
FALSE.}

\item{quiet}{Turns off most messages. Default is \code{TRUE}.}
}
\value{
The name of a filtered, sorted .bam file written to the user's
  current working directory. Or, if \code{make_bam = FALSE}, a .csv.gz file
  containing a data frame of only requisite information to run
  \code{metascope_id()}.
}
\description{
After a sample is aligned to a target library with
\code{align_target_bowtie()}, we may use \code{filter_host_bowtie()} to
remove unwelcome host contamination using filter reference libraries. This
function takes as input the name of the .bam file produced via
\code{align_target_bowtie()}, and produces a sorted .bam or .csv.gz file with
any reads that match the filter libraries removed. This resulting .bam file
may be used downstream for further analysis. This function uses Rbowtie2 For
the Rsubread equivalent of this function, see \code{filter_host}.
}
\details{
A compressed .csv can be created to produce a smaller output file that is
created more efficiently and is still compatible with \code{metascope_id()}.

The default parameters are the same that PathoScope 2.0 uses.
"--very-sensitive-local -k 100 --score-min L,20,1.0"
}
\examples{
#### Filter reads from bam file that align to any of the filter libraries

## Assuming a bam file has already been created with align_target_bowtie()
# Create temporary filter library
filter_ref_temp <- tempfile()
dir.create(filter_ref_temp)

## Create temporary taxonomizr accession
tmp_accession <- system.file("extdata", "example_accessions.sql", package = "MetaScope")

## Download reference genome
MetaScope::download_refseq("Orthoebolavirus zairense",
                           reference = FALSE,
                           representative = FALSE,
                           compress = TRUE,
                           out_dir = filter_ref_temp,
                           caching = TRUE,
                           accession_path = tmp_accession)

## Create temp directory to store the indices
index_temp <- tempfile()
dir.create(index_temp)

## Create filter index
MetaScope::mk_bowtie_index(
  ref_dir = filter_ref_temp,
  lib_dir = index_temp,
  lib_name = "filter",
  overwrite = TRUE
)

## Create temporary folder to hold final output file
output_temp <- tempfile()
dir.create(output_temp)

## Get path to example bam
bamPath <- system.file("extdata", "bowtie_target.bam",
                       package = "MetaScope")
target_copied <- file.path(output_temp, "bowtie_target.bam")
file.copy(bamPath, target_copied)

## Align and filter reads
filter_out <-
  filter_host_bowtie(
    reads_bam = target_copied,
    lib_dir = index_temp,
    libs = "filter",
    threads = 1
  )

## Remove temporary directories
unlink(filter_ref_temp, recursive = TRUE)
unlink(index_temp, recursive = TRUE)
unlink(output_temp, recursive = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metascope_blast.R
\name{get_seqs}
\alias{get_seqs}
\title{Gets sequences from bam file}
\usage{
get_seqs(id, bam_file, n = 10, bam_seqs)
}
\arguments{
\item{id}{Taxonomy ID of genome to get sequences from}

\item{bam_file}{A sorted bam file and index file, loaded with
Rsamtools::bamFile}

\item{n}{Number of sequences to retrieve}

\item{bam_seqs}{A list of the sequence IDs from the bam file}
}
\value{
Biostrings format sequences
}
\description{
Returns fasta sequences from a bam file with a given taxonomy ID
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/align_target.R
\name{align_target_bowtie}
\alias{align_target_bowtie}
\title{Align microbiome reads to set of indexed Bowtie2 libraries}
\usage{
align_target_bowtie(
  read1,
  read2 = NULL,
  lib_dir,
  libs,
  align_dir,
  align_file,
  bowtie2_options = NULL,
  threads = 1,
  overwrite = FALSE,
  quiet = TRUE
)
}
\arguments{
\item{read1}{Path to the .fastq file to align.}

\item{read2}{Optional: Location of the mate pair .fastq file to align.}

\item{lib_dir}{Path to the directory that contains the Bowtie2 indexes.}

\item{libs}{The basename of the Bowtie2 indexes to align against (without
trailing .bt2 or .bt2l extensions).}

\item{align_dir}{Path to the directory where the output alignment file should
be created.}

\item{align_file}{The basename of the output alignment file (without trailing
.bam extension).}

\item{bowtie2_options}{Optional: Additional parameters that can be passed to
the align_target_bowtie() function. To see all the available parameters use
Rbowtie2::bowtie2_usage(). See Details for default parameters. NOTE: Users
should pass all their parameters as one string and if optional parameters
are given then the user is responsible for entering all the parameters to
be used by Bowtie2. The only parameter that should NOT be specified here is
the number of threads.}

\item{threads}{The number of threads that can be utilized by the function.
Default is 1 thread.}

\item{overwrite}{Whether existing files should be overwritten. Default is
FALSE.}

\item{quiet}{Turns off most messages. Default is \code{TRUE}.}
}
\value{
Returns the path to where the output alignment file is stored.
}
\description{
This is the main MetaScope target library mapping function, using Rbowtie2
and multiple libraries. Aligns to each library separately, filters unmapped
reads from each file, and then merges and sorts the .bam files from each
library into one output file. If desired, output can be passed to
`filter_host_bowtie()` to remove reads that also map to filter library
genomes.
}
\details{
The default parameters are the same that PathoScope 2.0 uses.
"--very-sensitive-local -k 100 --score-min L,20,1.0"

If you experience any issues with reading the input files, make sure that
the file(s) are not located in a read-only folder. This can be circumvented
by copying files to a new location before running the function.
}
\examples{
#### Align example reads to an example reference library using Rbowtie2
\donttest{
## Create temporary directory to store file
target_ref_temp <- tempfile()
dir.create(target_ref_temp)

tmp_accession <- system.file("extdata", "example_accessions.sql", package = "MetaScope")

## Dowload reference genome
MetaScope::download_refseq("Morbillivirus hominis",
                           reference = FALSE,
                           representative = FALSE,
                           compress = TRUE,
                           out_dir = target_ref_temp,
                           caching = TRUE,
                           accession_path = tmp_accession
)

## Create temporary directory to store the indices
index_temp <- tempfile()
dir.create(index_temp)

## Create bowtie2 index
MetaScope::mk_bowtie_index(
  ref_dir = target_ref_temp,
  lib_dir = index_temp,
  lib_name = "target",
  overwrite = TRUE
)

## Create temporary directory for final file
output_temp <- tempfile()
dir.create(output_temp)

## Get path to example reads
readPath <- system.file("extdata", "virus_example.fastq",
                        package = "MetaScope")

## Align to target genomes
target_map <-
  MetaScope::align_target_bowtie(
    read1 = readPath,
    lib_dir = index_temp,
    libs = "target",
    align_dir = output_temp,
    align_file = "bowtie_target",
    overwrite = TRUE,
    bowtie2_options = "--very-sensitive-local"
  )

## Remove extra folders
unlink(target_ref_temp, recursive = TRUE)
unlink(index_temp, recursive = TRUE)
unlink(output_temp, recursive = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{bt2_regular_params}
\alias{bt2_regular_params}
\title{A universal parameter object for Bowtie 2 metagenomic or non-16S alignment}
\format{
list
}
\usage{
data(bt2_regular_params)
}
\description{
This character string provides several Bowtie 2 options to provide a 95% identity
alignment useful for metagenomes. This object can be used with functions that
use the Bowtie 2 aligner through the \code{Rbowtie2} package, namely
\code{align_target_bowtie()} and \code{filter_host_bowtie}. These settings
can be substituted for default settings by passing \code{} to the
\code{bowtie2_options} argument.
}
\details{
The default parameters listed in this object are
"--local -R 2 -N 0 -L 25 -i S,1,0.75 -k 5 --score-min L,0,1.7".

Further delineation of Bowtie 2 parameters is provided in the
\href{https://bowtie-bio.sourceforge.net/bowtie2/manual.shtml}{Bowtie 2
manual}.
}
\examples{
data("bt2_regular_params")

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metascope_blast.R
\name{taxid_to_name}
\alias{taxid_to_name}
\title{Converts NCBI taxonomy ID to scientific name}
\usage{
taxid_to_name(taxids, accession_path)
}
\arguments{
\item{taxids}{List of NCBI taxids to convert to scientific name}

\item{accession_path}{(character) Filepath to NCBI accessions SQL
database. See \code{taxonomzr::prepareDatabase()}.}
}
\value{
Returns a dataframe of blast results for a metascope result
}
\description{
Converts NCBI taxonomy ID to scientific name
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_animalcules_silva.R
\name{convert_animalcules_silva}
\alias{convert_animalcules_silva}
\title{Create a multi-assay experiment from MetaScope output for usage with
animalcules with the SILVA 13_8 database}
\usage{
convert_animalcules_silva(
  meta_counts,
  annot_path,
  which_annot_col,
  end_string = ".metascope_id.csv",
  qiime_biom_out = FALSE,
  path_to_write = ".",
  caching = TRUE
)
}
\arguments{
\item{meta_counts}{A vector of filepaths to the counts ID CSVs output by
\code{metascope_id()} created with the SILVA database.}

\item{annot_path}{The filepath to the CSV annotation file for the samples.
This CSV metadata/annotation file should contain at least two columns,
one with names of all samples WITHOUT the extension listed in \code{end_string},
e.g. for output file "sample_x76.metascope_id.csv", the column specified in
\code{which_annot_col} should contain the entry "sample_x76". Sample names
containing characters "_", "-", and "." are fine, however sample names
beginning with numbers should be renamed to have a prefix, e.g. "777897sample"
should be renamed to "X777897sample" for both the output file name and the
annotation name.}

\item{which_annot_col}{The name of the column of the annotation file
containing the sample IDs. These should be the same as the
\code{meta_counts} root filenames.}

\item{end_string}{The end string used at the end of the metascope_id files.
Default is ".metascope_id.csv".}

\item{qiime_biom_out}{Would you also like a qiime-compatible biom file
output? If yes, two files will be saved: one is a biom file of the counts
table, and the other is a specifically formatted mapping file of metadata
information. Default is \code{FALSE}.}

\item{path_to_write}{If \code{qiime_biom_out = TRUE}, where should output QIIME
files be written? Should be a character string of the folder path. Default is
'.', i.e. the current working directory.}

\item{caching}{Whether to use BiocFileCache when downloading genomes.
Default is \code{FALSE}.}
}
\value{
Returns a MultiAssay Experiment file of combined sample counts data
  and/or saved biom file and mapping file for analysis with QIIME. The 
  MultiAssayExperiment will have a counts assay ("MGX").
}
\description{
Upon completion of the MetaScope pipeline, users can analyze and visualize
abundances in their samples using the animalcules package. This function
allows interoperability of \code{metascope_id} output with both animalcules
and QIIME. After running this function, the user should save the returned MAE
to an RDS file using a function like \code{saveRDS} to upload the output into
the \code{animalcules} package. NOTE: This function is for outputs that were
generated with the SILVA 13_8 database.
}
\examples{
tempfolder <- tempfile()
dir.create(tempfolder)

# Create three different samples
samp_names <- c("X123", "X456", "X789")
all_files <- file.path(tempfolder,
                       paste0(samp_names, ".csv"))

create_IDcsv <- function (out_file) {
  final_taxids <- c("AY846380.1.2583", "AY909584.1.2313", "HG531388.1.1375")
  final_genomes <- rep("Genome name", 3)
  best_hit <- sample(seq(100, 1050), 3)
  proportion <- best_hit/sum(best_hit) |> round(2)
  EMreads <- best_hit + round(runif(3), 1)
  EMprop <- proportion + 0.003
  dplyr::tibble("TaxonomyID" = final_taxids,
                "Genome" = final_genomes,
                "read_count" = best_hit, "Proportion" = proportion,
                "EMreads" = EMreads, "EMProportion" = EMprop) |>
    dplyr::arrange(dplyr::desc(.data$read_count)) |>
    utils::write.csv(file = out_file, row.names = FALSE)
  message("Done!")
  return(out_file)
}
out_files <- vapply(all_files, create_IDcsv, FUN.VALUE = character(1))

# Create annotation data for samples
annot_dat <- file.path(tempfolder, "annot.csv")
dplyr::tibble(Sample = samp_names, RSV = c("pos", "neg", "pos"),
              month = c("March", "July", "Aug"),
              yrsold = c(0.5, 0.6, 0.2)) |>
  utils::write.csv(file = annot_dat,
                   row.names = FALSE)

# Convert samples to MAE
outMAE <- convert_animalcules_silva(meta_counts = out_files,
                                    annot_path = annot_dat,
                                    which_annot_col = "Sample",
                                    end_string = ".metascope_id.csv",
                                    qiime_biom_out = FALSE,
                                    caching = TRUE)

unlink(tempfolder, recursive = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/align_target.R
\name{combined_header}
\alias{combined_header}
\title{Create a combined .bam header}
\usage{
combined_header(bam_files, header_file = "header_tmp.sam")
}
\arguments{
\item{bam_files}{A character vector of the locations/file names of .bam files
from which to combine the headers.}

\item{header_file}{A file name and location for the output file for the
combined header. This will be a .sam format file without any reads.
Defaults to 'header_tmp.sam'.}
}
\value{
This function will return a combined header from all the supplied
.bam files.
}
\description{
This function generates a combined header from multiple .bam files from
different reference libraries (e.g. a split bacterial library). It is not
intended for use by users.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metascope_blast.R
\name{add_in_taxa}
\alias{add_in_taxa}
\title{Adds in taxa if silva database
Returns MetaScope Table with silva taxa in separate columns}
\usage{
add_in_taxa(metascope_id_in, caching, path_to_write)
}
\arguments{
\item{metascope_id_in}{MetaScope ID file with silva taxa}

\item{caching}{Boolean for if all_silva_headers.rds is already downloaded}

\item{path_to_write}{Path to save all_silva_headers.rds}
}
\value{
Data.frame of taxonomy information
}
\description{
Adds in taxa if silva database
Returns MetaScope Table with silva taxa in separate columns
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter_host.R
\name{remove_matches}
\alias{remove_matches}
\title{Helper function to remove reads matched to filter libraries}
\usage{
remove_matches(
  reads_bam,
  read_names,
  output,
  YS,
  threads,
  aligner,
  make_bam,
  quiet
)
}
\arguments{
\item{reads_bam}{The name of a merged, sorted .bam file that has previously
been aligned to a reference library. Likely, the output from running an
instance of \code{align_target()}.}

\item{read_names}{A \code{list} of target query names from \code{reads_bam}
that have also aligned to a filter reference library. Each \code{list}
element should be a vector of read names.}

\item{output}{The name of the .bam or .csv.gz file that to which the filtered
alignments will be written.}

\item{YS}{yieldSize, an integer. The number of alignments to be read in from
the bam file at once for chunked functions. Default is 100000.}

\item{threads}{The number of threads to be used in filtering the bam file.
Default is 1.}

\item{aligner}{The aligner which was used to create the bam file.}

\item{make_bam}{Logical, whether to also output a bam file with host reads
filtered out. A .csv.gz file will be created instead if \code{FALSE}.
Creating a bam file is costly on resources over creating a compressed
csv file with only relevant information, so default is \code{FALSE}.}

\item{quiet}{Turns off most messages. Default is \code{TRUE}.}
}
\value{
Depending on input \code{make_bam}, either the name of a filtered,
  sorted .bam file written to the user's current working directory, or an RDS
  file containing a data frame of only requisite information to run
  \code{metascope_id()}.
}
\description{
Using the \code{filter_host()} function, we align our sequencing sample to
all filter libraries of interest. The \code{remove_matches()} function allows
for removal of any target reads that are also aligned to filter libraries.
}
\details{
This function is not intended for direct use.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/align_target.R
\name{bam_reheader_R}
\alias{bam_reheader_R}
\title{Replace the header from a .bam file}
\usage{
bam_reheader_R(
  head,
  old_bam,
  new_bam = paste(tools::file_path_sans_ext(old_bam), "h.bam", sep = "")
)
}
\arguments{
\item{head}{A file name and location for the .sam file with the new header.}

\item{old_bam}{A file name and location for the .bam file which you would}

\item{new_bam}{A file name for the new .bam file with a replaced header.
Defaults to the same base filename plus 'h.bam'. For example, 'example.bam'
will be written as 'exampleh.bam'.}
}
\value{
This function will return a new .bam file with a replaced header. The
function also outputs the new .bam filename.
}
\description{
This function replaces the header from one .bam file with a header from a
different .sam file. This function mimics the function of the 'reheader'
function in samtools. It is not intended for use by users.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/align_target.R
\name{filter_unmapped_reads}
\alias{filter_unmapped_reads}
\title{Filter unmapped reads}
\usage{
filter_unmapped_reads(bamfile)
}
\arguments{
\item{bamfile}{Location for the .bam file to filter & remove all unmapped
reads}
}
\value{
This function will overwrite the existing .bam file with a new .bam
file in the same location that has only mapped lines. The function itself
returns the output .bam file name.
}
\description{
This function will remove all unmapped reads or lines in a .bam file
(warning: overwrites the original file!). This function is needed because
combining multiple .bam files from different microbial libraries may lead to
some reads that mapped to one library and have unmapped entries from another
library. This will remove any unmapped entries and leave all reference mapped
lines in the .bam file.
}
\details{
It is not intended for direct use.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metascope_blast.R
\name{check_blastn_exists}
\alias{check_blastn_exists}
\title{Check if blastn exists on the system}
\usage{
check_blastn_exists()
}
\value{
Returns TRUE if blastn exists on the system, else FALSE.
}
\description{
This is an internal function that is not meant to be used
  outside of the package. It checks whether blastn exists on the system.
}
\details{
Checks if blastn is installed
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{align_details}
\alias{align_details}
\title{A universal parameter settings object for Rsubread alignment}
\format{
list
}
\usage{
data(align_details)
}
\description{
This object is a named vector of multiple options that can be chosen for
functions that involve alignment with Rsubread, namely \code{align_target()}
and \code{filter_host()}. Both functions take an object for the parameter
\code{settings}, which are provided by \code{align_details} by default, or
may be given by a user-created object containing the same information.
}
\details{
The default options included in \code{align_details} are \code{type = "dna"},
\code{maxMismatches = 3}, \code{nsubreads = 10}, \code{phredOffset = 33},
\code{unique = FALSE}, and \code{nBestLocations = 16}. Full descriptions of
these parameters can be read by accessing \code{?Rsubread::align}.
}
\examples{
data("align_details")

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/align_target.R
\name{merge_bam_files}
\alias{merge_bam_files}
\title{Merge multiple .bam files}
\usage{
merge_bam_files(
  bam_files,
  destination,
  head_file = paste(destination, "_header.sam", sep = ""),
  quiet = TRUE
)
}
\arguments{
\item{bam_files}{A list of file names for the .bam files to be merged.}

\item{destination}{A file name and location for the merged .bam file.}

\item{head_file}{A file name and location for the combined header file.
Defaults to the destination. For example, 'example.bam' will be written as
'example.bam'.}

\item{quiet}{Turns off most messages. Default is \code{TRUE}.}
}
\value{
This function merges .bam files and combines them into a single file.
The function also outputs the new .bam filename.
}
\description{
This function merges .bam files. It first used the combined_header function
to generate a combined header for all the files, reheaders the files, and
then merges and sorts the .bam files. It is similar to the 'samtools merge'
function, but it allows the .bam files to have different headers. It is not
intended for direct use.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/demultiplex.R
\name{meta_demultiplex}
\alias{meta_demultiplex}
\title{Demultiplexing sequencing reads}
\usage{
meta_demultiplex(
  barcodeFile,
  indexFile,
  readFile,
  rcBarcodes = TRUE,
  location = NULL,
  threads = 1,
  hammingDist = 0,
  quiet = TRUE
)
}
\arguments{
\item{barcodeFile}{Path to a file containing a .tsv matrix with a
header row, and then sample names (column 1) and barcodes (column 2).}

\item{indexFile}{Path to a .fastq file that contains the barcodes for
each read. The headers should be the same (and in the same order) as
\code{readFile}, and the sequence in the \code{indexFile} should be the
corresponding barcode for each read. Quality scores are not considered.}

\item{readFile}{Path to the sequencing read .fastq file that corresponds
to the \code{indexFile}.}

\item{rcBarcodes}{Should the barcode indexes in the barcodeFile be reverse
complemented to match the sequences in the \code{indexFile}? Defaults to
\code{TRUE}.}

\item{location}{A directory location to store the demultiplexed read files.
Defaults to generate a new temporary directory.}

\item{threads}{The number of threads to use for parallelization
(BiocParallel). This function will parallelize over the barcodes and
extract reads for each barcode separately and write them to separate
demultiplexed files.}

\item{hammingDist}{Uses a Hamming Distance or number of base differences to
allow for inexact matches for the barcodes/indexes. Defaults to \code{0}.
Warning: if the Hamming Distance is \code{>=1} and this leads to inexact
index matches to more than one barcode, that read will be written to more
than one demultiplexed read files.}

\item{quiet}{Turns off most messages. Default is \code{TRUE}.}
}
\value{
Returns multiple .fastq files that contain all reads whose index
  matches the barcodes given. These files will be written to the location
  directory, and will be named based on the given sampleNames and barcodes,
  e.g. './demultiplex_fastq/SampleName1_GGAATTATCGGT.fastq.gz'
}
\description{
Function for demultiplexing sequencing reads arranged in a common format
provided by sequencers (such as Illumina) generally for 16S data. This
function takes a matrix of sample names/barcodes, a .fastq file of barcodes
by sequence header, and a .fastq file of reads corresponding to the barcodes.
Based on the barcodes given, the function extracts all reads for the indexed
barcode and writes all the reads from that barcode to separate .fastq files.
}
\examples{

## Get barcode, index, and read data locations
barcodePath <- system.file("extdata", "barcodes.txt", package = "MetaScope")
indexPath <- system.file("extdata", "virus_example_index.fastq",
                         package = "MetaScope")
readPath <- system.file("extdata", "virus_example.fastq",
                         package = "MetaScope")

## Demultiplex
demult <- meta_demultiplex(barcodePath, indexPath, readPath, rcBarcodes = FALSE,
                      hammingDist = 2)
demult

}
